/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  asm_opt.P
  Author(s)		:  Kostis F. Sagonas
  Last modification	:  June 14, 1993
========================================================================*/


/*======================================================================*/
/*  asm_opt(+Code, +LabelTab, +Index, +Sym, +IsModule, -OptCode)	*/
/*======================================================================*/

asm_opt(Code, LabelTab, Index, Sym, IsModule, OptCode) :-
	asm_transform(Code, LabelTab, Index, Code1),
	( option(optimize) -> 
		asm_optimize(Code1, Sym, IsModule, OptCode)
	; OptCode = Code1 
	).


/*======================================================================*/
/*  asm_transform(+Code, +LabelTab, +Index, -TransformedCode)		*/
/*======================================================================*/

asm_transform([], _, _, []).
asm_transform([Inst|Rest], LabelTab, Index, [NInst|NRest]) :-
	asm_pre(Inst, LabelTab, Index, NInst),
	asm_transform(Rest, LabelTab, Index, NRest).

asm_pre(call(Sym,Npars), LabelTab, _Index, NInst) :-
	( Sym = disj(_Label),
	  find_label(((Sym,_),EP), LabelTab),
	  NInst = calld(EP,Npars)
	; /* Sym = sym(...) */
	  ( sym_prop(internal, Sym),
		sym_prop(ep(EP), Sym),
		NInst = calld(EP,Npars)
	  ; NInst = call(Sym,Npars)
	  )
	).
asm_pre(execute(Sym), LabelTab, _Index, NInst) :-
	( Sym = disj(_Label),
	  find_label(((Sym,_),EP), LabelTab),
	  NInst = executed(EP)
	; /* Sym = sym(...) */
	  ( sym_prop(internal, Sym),
		sym_prop(ep(EP), Sym),
		NInst = executed(EP)
	  ; NInst = execute(Sym)
	  )
	).
asm_pre(switchonbound(Narg,Lab,NCls), _, Index, switchonbound(Narg,Ino,NCls)) :-
	member(cases(Lab,_,_,Ino), Index).
asm_pre(Inst, _, _, Inst).


/*======================================================================*/
/*  asm_optimize(+Code, +Sym, +IsModule, -OptCode)			*/
/*	Produces OptCode from the Code for predicate Sym.  At this	*/
/*	point the only optimization done is changing calls and executes	*/
/*	to direct calls and executes.  Other optimizations may be done	*/
/*	later (subject to performing a global compilation).		*/
/*======================================================================*/

asm_optimize([], _, _, []).
asm_optimize([Inst|Rest], Sym, IsModule, [NInst|NRest]) :-
	asm_direct(Inst, Sym, IsModule, NInst),
	asm_optimize(Rest, Sym, IsModule, NRest).


/*======================================================================*/
/*  asm_direct(+Inst, +PredSym, +IsModule, -NInst)			*/
/*	Changes calls and executes to direct calls (calld) and direct	*/
/*	executes (executed).  This optimization is possible under the	*/
/*	following conditions:						*/
/*	  - calls:							*/
/*		1. When the call is to the symbol for which we are	*/
/*		   generating code but it is not a tail-recursive	*/
/*		   call to the predicate (ex. nrev/2 or qsort/2).	*/
/*		2. Call within a module ..... (not implemented).	*/
/*	  - executes:							*/
/*		1. When the execute is to the symbol for which we are	*/
/*		   generating code (i.e. tail-recursive call).		*/
/*		2. When the execute is to a symbol which is `defined'	*/
/*		   within the module that we are compiling.  In this	*/
/*		   case the file MUST be a module (to avoid possible	*/
/*		   problems with overriding the definition of		*/
/*		   predicates by (re)loading a file/module).		*/
/*======================================================================*/

asm_direct(call(Sym,Npars), PredSym, _IsModule, NInst) :-
	( Sym = PredSym -> sym_prop(ep(EP), Sym), NInst = calld(EP,Npars)
	/* Add here appropriate code for calling within a module. */
	; NInst = call(Sym,Npars)
	).
asm_direct(execute(Sym), PredSym, _IsModule, NInst) :-
	( Sym = PredSym -> 
		sym_prop(ep(EP), Sym), NInst = executed(EP)
	/*  UNFORTUNATELY THE FOLLOWING CODE CANNOT BE INCLUDED
	 *	; IsModule =:= 1 -> 
	 *		( sym_prop(defined, Sym) -> 
	 *			sym_prop(ep(EP), Sym), NInst = executed(EP) 
	 *		; NInst = execute(Sym)
	 *		)
 	 */
	; NInst = execute(Sym)
	).
asm_direct(Inst, _, _, Inst).


/*======================================================================*/
/* --- other utilities ------------------------------------------------ */
/*======================================================================*/
 
find_label((abs(Value),Value), _).
find_label(Elem, label_tab(_,LabelTabList,_)) :-
	membercheck(Elem, LabelTabList).
find_label((Label,-1), _) :-
	error(('Label ', Label, ' is not found during assembling')).
 
force_asm_opt_loaded.

