/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  aux.P
  Author(s)		:  Saumya Debray, Kostis F. Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  June 3, 1993
========================================================================*/


/*======================================================================*/
/* member1: This is basically membercheck/2.				*/
/*======================================================================*/

member1(X, [X|_]) :- !.
member1(X, [_|L]) :- member1(X, L).

/*======================================================================*/
/* logical or								*/
/*======================================================================*/

logical_or(X,Y,Z) :-
	((X =:= 1, Z = 1) ;
	 (X =\= 1,
	  ((Y =:= 1, Z = 1) ;
	   (Y =\= 1, Z = 0)
	  )
	 )
	).

/*======================================================================*/
/*  messages and such							*/
/*======================================================================*/

message(Message) :- ttywritenl(Message).

warning(Message) :- message(('++Warning: ', Message)).

error(Message) :- 
	( conget('error occurred', 1) -> true
	; conset('error occurred', 1)
	),
	message(('++Error: ', Message)).

/*======================================================================*/
/*  setoptions(+Options)						*/
/*	Compiler option setting.					*/
/*======================================================================*/

setoptions(_) :- current_option(Option), conset(Option, 0), fail.
setoptions(Options) :- setoptions1(Options).

setoptions1([]).
setoptions1([Opt|Rest]) :- 
	( current_option(Opt) -> conset(Opt, 1)
	; warning(('Unknown compiler option: ', Opt))
	),
	setoptions1(Rest).

/* need to test out optimizations -- tls */
current_option(no_check).
current_option(nocompile).
current_option(verbo).
current_option(dumpasm).
current_option(sysmod).
current_option(noindex).
current_option(noinitvar).
current_option(ccp).
current_option(profile).
current_option(table_all).
current_option(optimize).

/* more options for type inference -- taken out by Kostis (10/16/92)

current_option(modeinfer).
current_option(typeinfer).
current_option(funcinfer).
current_option(shallow).

current_option(show_et).
current_option(print_local).
current_option(write_prog).
current_option(fix).
current_option(least).
current_option(d1).
current_option(step).
 */

/*======================================================================*/
/*  option(+Option)							*/
/*	Succeeds if Option is one of the compilation options.		*/
/*======================================================================*/

option(X) :- conget(X, 1).

/*======================================================================*/
/*  time_message(+StartTime, +Message)					*/
/*	Display the Message and the elapsed cpu time.			*/
/*======================================================================*/

time_message(StartTime, Message) :- 
	cputime(Time),
	Laps is Time - StartTime,
	message(('[', Message, ', cpu time used: ', Laps, ' seconds]')).

verbo_time_message(StartTime, Message) :- 
	( option(verbo) -> time_message(StartTime, Message) ; true ).

/*======================================================================*/
/*  subsumes(+X, +Y):							*/
/*	X subsumes Y, i.e. X is an instance of Y			*/
/*	The procedure is side-effect free.				*/
/*======================================================================*/

subsumes(X,Y) :- \+ X=Y,!,fail.
subsumes(X,Y) :- numbervars(Y,0,_), \+ X=Y,!,fail.
subsumes(_,_).

/*======================================================================*/
/*  subsumed_member(+Elem, +List):					*/
/*	check if Elem is subsumed by a one of elements in List		*/
/*======================================================================*/

subsumed_member(X, L) :- member(Y, L), subsumes(Y, X).

/*======================================================================*/
/* dict_lookup(+Key,-+Value,+-Dict)					*/
/*	find the Value associated with Key in the dictionary;		*/
/*	if no entry for it, insert it.					*/
/*		Key can be a variable also.				*/
/*======================================================================*/

dict_lookup(Key, Value, Dict) :- 
	var(Dict), !, 
	Dict = ['@'(Key, Value) | _].
dict_lookup(Key, Value, ['@'(Y, Value) | _]) :- 
	Key == Y, !.
dict_lookup(Key, Value, [_|Dict]) :- dict_lookup(Key, Value, Dict).

/*======================================================================*/
/* debug_writenl(A) : 							*/
/*======================================================================*/

debug_writenl(A) :- 
	( option(d1) ->  ttywritenl(('       ', A)) ; true ).

/*----freeze------------------------------------------------------------*/

freeze(X) :- numbervars(X, 0, _).

/*----melt--------------------------------------------------------------*/
/* melt(+Term, -NewTerm):	melt a frozen term to a new copy	*/
/*----------------------------------------------------------------------*/

melt(X, Y) :- melt(X, Y, _).

melt(X, X, _) :- atomic(X), !.
melt('$VAR'(I), V, FrozenVarDict) :- !, dict_lookup(I, V, FrozenVarDict).
melt(X, Y, FrozenVarDict) :-
	functor(X, F, N),
	functor(Y, F, N),
	meltargs(N, X, Y, FrozenVarDict).

meltargs(0, _, _, _) :- !.
meltargs(I, X, Y, FrozenVarDict) :-
	arg(I, X, X1),
	arg(I, Y, Y1),
	melt(X1, Y1, FrozenVarDict),
	I1 is I - 1,
	meltargs(I1, X, Y, FrozenVarDict).

/*----writetop------------------------------------------------------------*/

writetop(Term, Level) :-
	telling(F), tell(user), writetop0(Term, Level), nl, tell(F).

writetop0(Term, _Level) :- atomic(Term), !, write(Term).
writetop0(Term, _Level) :- var(Term), !, write(Term).
writetop0(Term, Level) :- 
	Level1 is Level - 1,
	Term =.. [F|Args],
	write(F),write('('),
	( Level1 < 0 -> write('_') ; writetop_args(Args,Level1), ! ),
	write(')'),
	!.

writetop_args([X], Level) :- writetop0(X, Level), !.
writetop_args([X|R], Level) :-
	writetop0(X, Level), write(','), writetop_args(R, Level).
