/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  build_graph.P
  Author(s)		:  Kate Dvortsova
  Last modification	:  October 30, 1992
========================================================================*/

build_call_graph([]):-!.
build_call_graph([Pred|Tail]):-
          pred_build(Tail,_,Pred,Ans),!,
            Pred=pred(Sym,_,_),
          ( Ans==r -> sym_propin(td(r),Sym) ; true),
          build_call_graph(Tail). 

/*  Is predicate table depended or not?  */


pred_build(Tail,Tail,pred(Sym,_,_),t):-
          sym_prop(tabled(_,_,_,_),Sym),!.
pred_build(Tail,Tail,pred(Sym,_,_),Ans):-
          sym_prop(td(Ans),Sym),!.
pred_build(Tail,Outtail,pred(Sym,Clist,_),Ans):-
         check_clist(Tail,Outtail,Clist,Ans),
         ( Ans==t ->
          sym_propin(td(t),Sym) ; true ).

/* check the body of the predicate */

check_clist(Tail,Tail,[],r):-!.

check_clist(Tail,Y,[clause(_,Goal,_)|L],Ans):-
         check_goal(Tail,Outtail,Goal,Ans1),
         (Ans1==t -> (Ans=t,Y=Outtail);
                    (!, check_clist(Outtail,Y,L,Ans))).

check_goal(Tail,Tail,inlinegoal(_,_,_),r):-!.
check_goal(Tail,Outtail,Goal,Ans):-
          meta_pred(Goal,Args),!,
          args(Tail,Outtail,Args,Ans1),
          (Ans1=t(_,_) -> Ans=t ; Ans=r).
check_goal(Tail,Tail,goal(sym(Name,Arity,_),_),r):-
           standard_symbol(Name,Arity,_),!.
check_goal(Tail,Tail,goal(Sym,_),t):-
           sym_prop(tabled(_,_,_,_),Sym),!.
check_goal(Tail,Tail,goal(Sym,_),Ans):-
           sym_prop(td(Ans),Sym),!.


check_goal(Tail,Outtail,goal(Sym,_),Ans):-
           restmember(Tail,pred(Sym,Clist,_),Tail1),!,
           pred_build(Tail1,Outtail,pred(Sym,Clist,_),Ans).

check_goal(Tail,Tail,goal(_,_),r):-!.
check_goal(Tail,Outtail,and(Goal1,Goal2),Ans):-
            check_goal(Tail,Tail1,Goal1,Ans1),
            (Ans1==t -> (Outtail=Tail1,Ans=t) ;         
             (!,check_goal(Tail1,Outtail,Goal2,Ans))).

check_goal(Tail,Outtail,or(Goal1,Goal2),Ans):-
            check_goal(Tail,Tail1,Goal1,Ans1),
            (Ans1==t -> (Outtail=Tail1,Ans=t) ;
             (!,check_goal(Tail1,Outtail,Goal2,Ans))).


check_goal(Tail,Outtail,if(Goal1,Goal2),Ans):-
            check_goal(Tail,Tail1,Goal1,Ans1),
            (Ans1==t -> (Outtail=Tail1,Ans=t) ;         
             (!,check_goal(Tail1,Outtail,Goal2,Ans))).

check_goal(Tail,Outtail,not(Goal1),Ans):-
            check_goal(Tail,Outtail,Goal1,Ans).
 
check_goal(_,_,_,r).
restmember([],_,[]):-!,fail.
restmember([El|T],El,T):-!.
restmember([X|T],El,[X|T1]):-
          restmember(T,El,T1).

            /* Meta-predicates */
meta_pred(goal(sym(call,1,_,_),Args),Args):-!.
meta_pred(goal(sym(Name,3,_,_),Args),Args):-
         Name=='findall' ; Name=='setof' ; Name=='bagof',!.
meta_pred(goal(sym(not,1,_,_),Args),Args):-!.

args(Tail,Tail,[],r).
args(Tail,Outtail,[Arg|Args],Ans):-
         argument(Tail,Outtail1,Arg,Ans1),
         (Ans1=t(P,A) -> (Ans=t(P,A),Outtail=Outtail1) ;
          args(Outtail1,Outtail,Args,Ans) ).
argument(Tail,Outtail,constant(Sym),Ans):-
         check_goal(Tail,Outtail,goal(Sym,_),Ans1),!,
         (Ans1==t -> (Sym=sym(P,A,_,_),Ans=t(P,A)) ; Ans=r).
argument(Tail,Outtail,structure(Sym,Args),Ans):-
         check_goal(Tail,Outtail1,goal(Sym,_),Ans1),
         Sym=sym(P,A,_,_),
         (Ans1==t -> (Ans=t(P,A),Outtail=Outtail1) ;
            args(Tail,Outtail,Args,Ans)).
argument(Tail,Tail,_,r). 

          /******************/   

/*-------------- Build_graph/1 -----------------*/

build_graph(module(_,_,_,_,List,Query)):-
      build_call_graph(List),
      cut_check(List),
      cut_check(Query).


/*------------------- not & cut checking--------------------------------*/

cut_check([]).
cut_check([pred(Sym,Clist,_)|List]):-
          Sym=sym(C,Ar,_,_),
          pred_check(Clist,c(C,Ar)),
          cut_check(List).

pred_check([],_):-!.
pred_check([clause(_,Goal,_)|L],Cl):-
          goal_cut(Goal,_,Cl),!, 
          pred_check(L,Cl).

          /* not checking */

goal_cut(not(inlinegoal(_,_,_)),_,_):-!.
goal_cut(not(Goal),T,Cl):-
        meta_pred(Goal,Args),!,
        args([],_,Args,Ans1),!,
        (Ans1=t(P,A) -> (error1(P,A,Cl),
             (var(T) -> T=t(P,A) ; true)) 
              ; true), 
             str_check(Args,Cl).
goal_cut(not(goal(sym(Name,Arity,_,_))),_,_):-
          standard_symbol(Name,Arity,_),!.
goal_cut(not(goal(Sym,_)),T,Cl):-!,
         ( ( sym_prop(td(t),Sym) ; sym_prop(tabled(_,_,_,_),Sym)) ->  /*IF*/
             (   Sym=sym(P,A,_,_),                                    /*THEN*/
                 error1(P,A,Cl),
              (var(T) -> T=t(P,A) ; true)  ) ;
             true ).                                                  /*ELSE*/
goal_cut(not(and(Goal1,Goal2)),T,Cl):-!,
          goal_cut(not(Goal1),T,Cl),!,
          goal_cut(not(Goal2),T,Cl).
goal_cut(not(or(Goal1,Goal2)),T,Cl):-!,
          goal_cut(not(Goal1),T,Cl),!,
          goal_cut(not(Goal2),T,Cl).
goal_cut(not(not(Goal)),T,Cl):-!,
          goal_cut(not(Goal),T,Cl).
goal_cut(not(if(Goal1,Goal2)),T,Cl):-
          goal_cut(Goal1,T,Cl),
          goal_cut(not(Goal2),T,Cl).

       /*  cut checking */

goal_cut(inlinegoal('!',0,_),T,Cl):-!,
         (var(T) -> true ;
         ( T=t(P,A), error2(P,A,Cl)) ).
goal_cut(inlinegoal(_,_,_),_,_):-!.
goal_cut(Goal,T,Cl):-
         meta_pred(Goal,Args),!,
         args([],_,Args,Ans1),!,
         (Ans1=t(_,_) -> 
          ( (var(T) -> T=Ans1 ; true),
           str_check(Args,Cl)) ; true).
       
goal_cut(goal(sym(Name,Arity,_,_),_),_,_):-
         standard_symbol(Name,Arity,_),!.
goal_cut(goal(Sym,_),T,_):-
       (  (sym_prop(td(t),Sym) ; sym_prop(tabled(_,_,_,_),Sym)) ->  /* IF */
          (Sym=sym(P,A,_,_),(var(T) -> T=t(P,A) ; true)             /*THEN*/
           ) ;             
           true ), !.                                               /* ELSE */
         
goal_cut(and(Goal1,Goal2),T,Cl):-
         goal_cut(Goal1,T,Cl),
         goal_cut(Goal2,T,Cl). 
goal_cut(or(Goal1,Goal2),T,Cl):-
         Goal2=inlinegoal('!',0,_),!,
         goal_cut(Goal2,T,Cl),
         goal_cut(Goal1,T,Cl).
goal_cut(or(Goal1,Goal2),T,Cl):-
         goal_cut(Goal1,T,Cl),
         goal_cut(Goal2,T,Cl).
goal_cut(if(Goal1,Goal2),T,Cl):-
         goal_cut(Goal1,T,Cl),
         goal_cut(Goal2,T,Cl).

      /* structure_checking */

str_check([],_).
str_check([Arg|Args],Cl):-
          el_check(Arg,Cl),!,
          str_check(Args,Cl).

el_check(structure(Sym,Args),Cl):-
         (Sym=sym('\+',1,_,_) ; Sym=sym('fail_if',1,_,_)),
          args([],_,Args,t(P,A)),!,
          error1(P,A,Cl).

el_check(structure(_,Args),Cl):-!,
          str_check(Args,Cl).

el_check(_,_).

error1(P,A,c(C,Ac)):-
        K=P/A,
        L=C/Ac,
        error(('''/+'' or ''fail_if'' of table depended symbol ',K, ' in a clause of ',L)).

error2(P,A,c(C,Ac)):-
        K=P/A,
        L=C/Ac,
        error(('''!'' after table depended symbol ',K, ' in a clause of ', L)).
 

/* ------------------ end of file build_graph.P ----------------------- */


