/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  compile.P
  Author(s)		:  Saumya Debray, Kostis Sagonas, Terrance Swift
  Modified by		:  Kostis Sagonas, Terrance Swift
  Last modification	:  June 12, 1993
========================================================================*/

mc(Modname, Options) :- compile(Modname, Modname, Options, _).

/*======================================================================*/
/* compile(+ModName, +BaseName, +CompilerOptions, -ExportedPredicates)	*/
/*	Compiles a module with the compiler options specified and	*/
/*	returns the list of exported predicates.			*/
/*======================================================================*/

compile(Modname, Basename, Options, ExportList) :- 
	file_names(Basename, PFile, HFile, OFile, AFile),
	setoptions(Options), initialize,
	message(('[Compiling ', Basename, ']')), cputime(TimeStart),
	PIFile = PFile,
	parse(Modname, PIFile, HFile, Module0),
    ( option(profile) -> 
		time_message(TimeStart, 'Parse finito'), cputime(TimeParse)
    ; TimeParse = TimeParse ),	% To avoid a silly lamp variable message!
	abort_if_syntax_errors(Modname), !,
	( option(table_all) -> define_table(Module0) ; true ),
	( conget('table #', 0) -> Module1 = Module0 
	; cp_opt(Module0, Module1), 
          (\+option(no_check) -> build_graph(Module1) ; true)
	), 
	flatten(Module1, Module),
    ( option(profile) -> 
		time_message(TimeParse, 'Flatten finito'), cputime(TimeFlatten)
    ; TimeFlatten = TimeFlatten ),% To avoid a silly lamp variable message!
	get_symtab(Module, SymTab),
	sym_complete(SymTab),
	useinfer(Module, IsModule, ExportList), !,
    ( option(profile) -> time_message(TimeFlatten, 'Useinfer finito') ; true ),
	hack_self_compile,
	compile1(Modname, Module, SymTab, AFile, OFile, IsModule),
	( option(nocompile) -> print_result(Module) ; true ),
	( conget('error occurred', 1) ->
		str_cat('/bin/rm ', OFile, Command), system(Command), 
		message(('The file ', OFile, ' is removed')), abort
	; true
	), !,
	( IsModule =:= 0 -> IsModuleMsg = '' ; IsModuleMsg = 'Module ' ),
	time_message(TimeStart, (IsModuleMsg, Modname, ' compiled')), !.


file_names(Basename, PFile, HFile, OFile, AFile) :-
	str_cat(Basename, '.P', PFile),
	str_cat(Basename, '.H', HFile),
	str_cat(Basename, '.O', OFile),
	str_cat(Basename, '.A', AFile).

initialize :-
	conset('table #', 0),
	conset('syntax errors', 0),
	conset('error occurred', 0),
	prepare.   /* uses globalset here! */

abort_if_syntax_errors(Modname) :-
	( conget('syntax errors', N), N =\= 0, Modname \== user -> 
		% Forget about syntax errors when compiling module user
	 	message(('Aborting compilation of module ', 
			 Modname, ' due to ', N, ' syntax error(s).')), 
		abort
	; true
	).


compile1(Modname, Module, SymTab, AFile, OFile, IsModule) :-
      ( option(nocompile) ->		%  generate *.A file
	  write_asmheader(AFile, Modname, SymTab, IsModule), myclose(AFile)
      ; conget('error occurred', 1) -> message('Abort due to errors'), abort
      ; option(dumpasm) -> 
	  write_asmheader(AFile,Modname,SymTab, IsModule), % generate .A header
	  asm_genheader(OFile, Modname, SymTab, IsModule), % generate .O header
	  compile0(Module, genasm, AFile, OFile, IsModule),
	  myclose(AFile), myclose(OFile)
      ; asm_genheader(OFile, Modname, SymTab, IsModule),   % generate .O header
	compile0(Module, noasm, AFile, OFile, IsModule),
	myclose(OFile)
      ).

/*======================================================================*/
/* compile0(+Module, +(genasm|noasm), +AFile, +OFile, +IsModule)	*/
/*	Compiles a module in the internal form.				*/
/*======================================================================*/

compile0(module(SymTab,_,_,_,ClList,[MainPred]), GenAsm, AFile, OFile, M) :-
	member(PredBlk, [MainPred|ClList]),
	compile_predicate(PredBlk, GenAsm, genout, AFile, OFile, M, SymTab),
	fail.	/* just reclaim space */
compile0(_, _, _, _, _).


/*======================================================================*/
/* compile_predicate(+pred(Sym,Clauses,_), +(genasm|noasm), 		*/
/*		     +(genout|noout), +AFile, +OFile,			*/
/*		     +IsModule, +SymTab)				*/
/*	Compile the clauses of a single predicate into an assembly	*/
/*	(*.A) file and/or an object (*.O) file.				*/
/*======================================================================*/

compile_predicate(PredDef, GenAsm, GenOut, AFile, OFile, IsModule, SymTab) :-
	( PredDef = pred(Sym,_,_) ; PredDef = fun(Sym,_,_) ),
	sym_name(Sym, P, N),
	( option(verbo) -> message(('        ....compiling ', P/N)) ; true ),
	( option(sysmod) -> true
	; standard_symbol(P, N, _) ->
		warning(('Redefining the builtin predicate/function: ', P/N))
	; true
	),
	( sym_prop(im(Mod), Sym) ->
		error(('Cannot redefine ', P/N,
		' (already declared imported from ', Mod, '!)'))
	; true
	),
	preprocess(PredDef, PredBlks, SymTab),
	compile_pred_block(PredBlks, PilCode-[], 0, NC),
	compile_peephole_opt(PilCode, PilCode1),
	( GenAsm = genasm -> tell(AFile), write_asm_instructions(PilCode1)
	; true
	),
	( GenOut = genout -> tell(OFile), asm_PIL(Sym, PilCode1, NC, IsModule)
	; true ),
	!.

compile_pred_block([], PilCode-PilCode, NCin, NCin).
compile_pred_block([PredDef|PRest], PilCode-PilCodeR, NCin, NCout) :-
	compile_predicate2(PredDef, PilCode-PilCodeM, NClauses),
	NCmid is NCin + NClauses,
	!,
	compile_pred_block(PRest, PilCodeM-PilCodeR, NCmid, NCout).

/*======================================================================*/
/* compile_predicate2(+pred(Sym,Clauses,Prag), -PilCode-PilCodeR, -NC)	*/
/*	Compiles the clauses of a single predicate after preprocessing.	*/
/*	A single predicate in the source level can be transformed by	*/
/*	preprocess/3 into several predicates.  compile_predicate2/2	*/
/*	does the compilation for one of them.  The result is a list of	*/
/*	PIL code in assembler format.  The predicate also returns the	*/
/*	number of clauses NC (length of the Clauses list).		*/
/*======================================================================*/

compile_predicate2(pred(Sym,Clauses,Prag), PilCode-PilCodeR, NC) :-
	gennum(L),
	PilCode = [label(pred,(Sym,L))|PilCode1],
	find_if_tabled(Sym,Tabled,Tabind),
	makeindex(Sym, Clauses, PilCode1, PilCodeM, Prag, Tabled, Tabind),
	need_neck(Clauses, Neck),
%    ( option(profile) -> cputime(TimeStart) ; TimeStart = TimeStart ),
	compile_clauses(Clauses, Sym, PilCodeM, PilCodeR, Neck, 0, NC).
%    ( option(profile) -> 
%		time_message(TimeStart, 'Compile_clauses finito') 
%    ; true).
compile_predicate2(fun(Sym,Clauses,Prag), PilCode-PilCodeR, NC) :-
	gennum(L),
	PilCode = [label(pred,(Sym,L))|PilCode1],
	makeindex(Sym, Clauses, PilCode1, PilCodeM, Prag, 0, _), % non-tabled
	need_neck(Clauses, Neck),
	compile_clauses(Clauses, Sym, PilCodeM, PilCodeR, Neck, 0, NC).

find_if_tabled(Sym, 1, tabled(Tabind,Chs,Rhs,Ar)):-
	sym_type(Sym, tabled(Tabind,Chs,Rhs,Ar)),!. 
find_if_tabled(_, 0, _). 

/* need a neck instruction (shallow backtracking) when there are more than */
/* one clause and the options is turned on */

need_neck([_,_|_], 1) :- option(shallow), !.
need_neck(_, 0).

compile_clauses([], _, PilCode, PilCode, _, Nin, Nin).
compile_clauses([Clause|Clauses], Sym, PilCode, PilCodeR, Neck, Nin, Nout) :-
	compile_clause(Sym, Clause, PilCode, PilCodeM, Neck),
	Nmid is Nin + 1,  
	compile_clauses(Clauses, Sym, PilCodeM, PilCodeR, Neck, Nmid, Nout).


/*======================================================================*/
/* hack_self_compile/0							*/
/*	Forces all the modules that are used by the XSB compiler to	*/
/*	be loaded into the system before overwriting the .O file when	*/
/*	self-compiling the module (or the whole system).		*/
/*======================================================================*/

hack_self_compile :-
        force_tprog_loaded,
        force_inprog_loaded,
        force_tp_var_loaded,
        force_varproc_loaded,
        force_tp_eval_loaded,
        force_tp_goal_loaded,
	force_asm_opt_loaded,
        force_tp_index_loaded,
        force_peephole_loaded,
	force_asm_inst_loaded,
        force_asm_pass2_loaded,
        force_preprocess_loaded,
        force_tp_geninline_loaded.

