/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  listutil.P
  Author(s)		:  Saumya Debray, Kostis F. Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  September 21, 1992
========================================================================*/


/* --- some Prolog list utilities used in cmplib ---------------------- */

attach(X, Y) :- (var(Y), Y = [X|_]) ;
		(nonvar(Y), Y = [_|T], attach(X, T)).


merge([], L1, L1).
merge([H|T], L1, L3) :- (member(H, L1), merge(T, L1, L3)) ;
			(L3 = [H | L4], merge(T, L1, L4)).


%   absmember(+Element, +List)
%   Predicate absmember/2 is similar to the good old member/2, except 
%   that it checks for identity rather than unifiability.

absmember(X, [H|T]) :- '=='(X,H) ; absmember(X, T).


%   absmerge(+List1, +List2, ?Merged_List)
%   Predicate absmerge/3 is similar to merge/3, except that it uses 
%   absmember/2 rather than member/2.

absmerge([], L1, L1).
absmerge([H|T], L1, L3) :- (absmember(H, L1), absmerge(T, L1, L3)) ;
			   (L3 = [H | L4], absmerge(T, L1, L4)).


%   member2(?Element, +List)
%   Predicate member2/2 is used just to check for membership in a list 
%   whose tail is a variable. It fails if the element is not found or
%   if the list is an improper list (including variable).

member2(X,L) :- nonvar(L), L = [Y|Z], (X = Y ; member2(X,Z)).


%   closetail(?List)
%   Predicate closetail/1 closes the tail of an open-ended list.

closetail([]).
closetail([_|L]) :- closetail(L).


%   length(+List, ?Length)
%   Predicate length/2 returns the length of list List. This predicate
%   should be used with the first argument instantiated. For a predicate
%   that works both ways use the length/2 in basics.

length([], 0).
length([_|R], N) :- length(R, N1), N is N1 + 1.


insert(Elem, [Elem|_]) :- !.
insert(Elem, [_|Rest]) :- insert(Elem, Rest).

membercheck(Elem, Table) :- nonvar(Table), Table = [Head|Rest],
			    (Elem = Head ; membercheck(Elem, Rest)).

nthmember(X, [X|_], 1).
nthmember(X, [_|Rest], N) :- nthmember(X,Rest,N1), nonvar(N1), N is N1 + 1.
