/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  parse.P
  Author(s)		:  Saumya Debray, Kostis Sagonas, Terrance Swift
  Modified by		:  Jiyang Xu, David S. Warren
  Last modification	:  June 1, 1993
========================================================================*/


/*======================================================================*/
/* parse(+Modname, +PFile, +HFile,					*/
/*	-module(-SymTab,-TyList,-DcList,-FnList,-ClList, -QrList))	*/
/*======================================================================*/

parse(Modname, PFile, HFile, Module) :-
	(Modname = user
	 ->	see(user), parse_clauses(user, Module0)
	 ; file_exists(PFile)
	 ->	(file_exists(HFile)
		 ->	parse1(Modname,HFile,Module0)
		 ;	Module0 = Module0	% init (lamp variable)
		),
		parse1(Modname,PFile,Module0)
	 ;	error(('Cannot find the file ', PFile)), fail, Module0=Module0
	),
	module_close(Module0,Module), !.

parse1(Modname, File, Module) :-
	see(File),
	parse_clauses(Modname, Module),
	seen.


/*======================================================================*/
/* parse_clauses(+ModName, #module(#SymTab, #TyList, #DcList,		*/
/*				   #FnList, #ClList, #QrList))		*/
/*======================================================================*/

parse_clauses(ModName, Module) :-
	my_read(Clause, VariableList),
	expand_term(Clause, Expanded_Clause),
	( Expanded_Clause == end_of_file, !
	; otherwise, !,
		parse_clause(Expanded_Clause, VariableList, ModName, Module),
		parse_clauses(ModName, Module)
	).
parse_clauses(M, _) :-	% Under normal conditions this clause is unreachable!
	error(('Clauses of ', M, ' cannot be parsed.')).

my_read(Term, Vars) :- stat_flag(10, File), file_read(File, Term, Vars).

/*======================================================================*/
/* parse_clause(+Clause, +VariableList, +ModName, #module(#SymTab,	*/
/*			#TyList, #DcList, #FnList, #ClList, #QrList))	*/
/*======================================================================*/

parse_clause([], _, _, _) :- !.
parse_clause([Cl|Cls], VariableList, Modname, Module) :- !,
	parse_clause(Cl, VariableList, Modname, Module),
	parse_clause(Cls, VariableList, Modname, Module).
parse_clause(Cl, VariableList, _Modname,
	     module(SymTab,_TyList,_DcList,_FnList,ClList,QrList)) :-
%	( Cl = (':-'(FH=FB, C)) ->	parse_fundef(FH, FB, C, ClList,SymTab)
	( Cl = (':-'(H, B)) -> 		singleton_check((H:-B), VariableList),
					parse_preddef(H, B, ClList, SymTab)
	; Cl = (':-'(Directive)) ->	( parse_directive(Directive, SymTab), !
					; parse_query(Directive, QrList,SymTab)
					)
	; Cl = ('?-'(Q)) -> 		parse_query(Q,QrList,SymTab)
%	; Cl = (FH=FB) ->		parse_fundef(FH,FB,true,ClList,SymTab)
	; singleton_check((Cl:-true), VariableList),
	  parse_preddef(Cl, true, ClList, SymTab)
	).


/*======================================================================*/
/* parse_directive(+Directive, #Symbol_Table)				*/
/*	Processes the given directive and updates the symbol table	*/
/*	accordingly.  However, since the symbol table uses buffers to	*/
/*	record the number of symbols that have been found, we should	*/
/*	be certain that Directive can be parsed as a directive before	*/
/*	inserting any symbols to the symbol table (non-backtrackable	*/
/*	insert).							*/
/*======================================================================*/

parse_directive(export(X), SymTab) :- 	
	parse_list(X, SymTab, [ex], (export)).
parse_directive(import(from(X, Mod)), SymTab) :-
	parse_list(X, SymTab, [im(Mod)], (import)),
	sym_insert(Mod, 0, [module], SymTab, _). 
%  parse_directive(type(X), SymTab) :-
%	parse_list(X, SymTab, [typepred], (type)).
%  parse_directive(hide(X), SymTab) :-
%	parse_list(X, SymTab, [hide], (hide)).
parse_directive(index(X), SymTab) :-
	parse_index(X, SymTab).
parse_directive(index(Ps,Arg,Size), SymTab) :-
	Ps = Pname/Arity,
	sym_insert(Pname,Arity,[index(Arg)],SymTab,_),
	index(Ps,Arg,Size), 
	fail.
parse_directive(index(Ps,Arg), SymTab) :-
	Ps = Pname/Arity,
	sym_insert(Pname,Arity,[index(Arg)],SymTab,_),
	index(Ps,Arg,0),
	fail.
parse_directive(mode(X), SymTab) :- 
	parse_mode(X, SymTab).
parse_directive(parallel(X), SymTab) :-
	parse_list(X, SymTab, [(parallel)], (parallel)).
parse_directive(fun(X), SymTab) :-
	parse_list(X, SymTab, [(fun)], (fun)).
%  parse_directive(file(X), SymTab) :-
%	parse_list(X, SymTab, [file], (file)).
parse_directive(table(X), SymTab) :-
	parse_list(X, SymTab, [tabled(_,_)], (table)).
parse_directive(table(Pred,Hashsize), SymTab) :- 
	parse_list(Pred, SymTab, [tabled(_,Hashsize)], (table)).
parse_directive(table(Pred,Chs,Rhs), SymTab) :- 
	parse_list(Pred, SymTab, [tabled(Chs,Rhs)], (table)).
parse_directive(table_all, _SymTab) :-	% This is quite dirty!
	conset(table_all, 1).
parse_directive(op(P,T,S), _SymTab) :-
	op(P,T,S), 
	fail.
parse_directive(hilog(X), _SymTab) :-
	add_hilog_symbol(X), 
	fail.


/*======================================================================*/
/* process regular clauses --- collect them in List			*/
/*	ListType : tylist, dclist, cllist				*/
/* This code is also used in cp_opt.P, so make sure any changes here    */
/* dont break cp_opt/2							*/
/*======================================================================*/

parse_preddef(H, B, List, SymTab) :-
	functor(H, F, A),
	sym_insert(F, A, [defined], SymTab, PredSym),
	member(pred(PredSym,PCs,_), List),
	trans_head(H, goal(_, ArgList), SymTab, 0, VidMed),
	trans_goals(B, Body, SymTab, VidMed, _VidOut),
	add_new_clause(clause(ArgList, Body, _), PCs),
	sym_propin(pred, PredSym).


/*======================================================================*/
/* process queries --- same as process regular clauses, but adding	*/
/* a predicate '_$main'/0.						*/
/*	ListType : tylist, dclist, cllist				*/
/*======================================================================*/

parse_query(B, List, SymTab) :-
	parse_preddef('_$main', (B->fail), List, SymTab).


/*======================================================================*/
/* process function definitions --- collect them in ClList		*/
/*======================================================================*/
/* ------------ Commented out for now (Kostis 5/23/1993) ---------------
parse_fundef(FH,FB,FC,ClList, SymTab) :-
	var(FH), !, 		%  handle X = X, etc.
	parse_preddef((FH=FB), FC, ClList,SymTab).
parse_fundef(FH,FB,FC,ClList, SymTab) :-
	functor(FH, F, A),
	sym_insert(F, A, [(fun),defined], SymTab, FunSym),
	member(fun(FunSym, PCs, _), ClList),
	trans_head(FH, goal(_, ArgList), SymTab, 0, VidMed),
	trans_arg(FB, Body, SymTab, VidMed, VidMed2),
	trans_goals(FC, Constraint, SymTab, VidMed2, _VidOut),
	add_new_clause(clause(ArgList,funbody(Body,Constraint), _), PCs).
------------------------------------------------------------------------*/

/*======================================================================*/
/* parse_list(+PredList, #SymTab, +SymbolProp, +Directive) 		*/
/*	Processes export/import/type lists (fills in the symbol table).	*/
/*	It also does the same for hide/parallel/fun/table lists.	*/
/*======================================================================*/

parse_list(P/A, SymTab, SymbolProp, _) :- 
	nonvar(P), integer(A), 
	!,
	sym_insert(P, A, SymbolProp, SymTab, _).
parse_list((Pred, Preds), SymTab, SymbolProp, Directive) :- 
	!,
	parse_list(Pred, SymTab, SymbolProp, Directive),
	parse_list(Preds, SymTab, SymbolProp, Directive).
parse_list(X, _, _, Directive) :-
	warning(('Non predicate specification "', X, 
		 '" found in ', Directive, ' directive')),
	fail.

/*======================================================================*/
/*	Closes the lists in the completed module description.		*/
/*======================================================================*/

module_close(module(SymTab,TyList0,DcList0,FnList0,ClList0,QrList0),
	     module(SymTab,TyList,DcList,FnList,ClList,QrList)) :-
	clauselist_straight(TyList0,TyList),
	clauselist_straight(DcList0,DcList), 
	clauselist_straight(FnList0,FnList), 
	clauselist_straight(ClList0,ClList), 
	parse_preddef('_$main', true, QrList0, SymTab),
	clauselist_straight(QrList0,QrList), 
	!.

clauselist_straight([],[]).
clauselist_straight([pred(Sym,Cls,P)|Rest0],[pred(Sym,Lcls,P)|Rest]) :-
	clause_listify(Cls,Lcls),
	clauselist_straight(Rest0,Rest).
clauselist_straight([fun(Sym,Cls,P)|Rest0],[fun(Sym,Lcls,P)|Rest]) :-
	clause_listify(Cls,Lcls),
	clauselist_straight(Rest0,Rest).

/*======================================================================*/
/* auxiliary routines							*/
/*======================================================================*/

get_symtab(module(SymTab,_,_,_,_,_), SymTab).

/*======================================================================*/
/*======================================================================*/

parse_mode((Pred, Preds), SymTab) :- !,
	parse_mode(Pred, SymTab),
	parse_mode(Preds, SymTab).
parse_mode(Pred, SymTab) :- 
	functor(Pred, Name, Arity),
	sym_insert(Name, Arity, [mode(Pred)], SymTab, _Sym).

/*======================================================================*/
/*======================================================================*/

parse_index((Pred, Preds), SymTab) :- !,
	parse_index(Pred, SymTab),
	parse_index(Preds, SymTab).
parse_index(Pname/Arity-Arg, SymTab) :- !,
	index(Pname/Arity, Arg, 0),
	sym_insert(Pname, Arity, [index(Arg)], SymTab, _).
parse_index(Pname/Arity, SymTab) :- 
	index(Pname/Arity, 1, 0),
	sym_insert(Pname, Arity, [index(1)], SymTab, _).

/*======================================================================*/
/* trans_goals(+Term, -Goal, #SymTab, +VidIn, -VidOut)			*/
/*======================================================================*/

trans_goals('$$var'(Vid), goal(Sym,[varocc(Vid)]), SymTab, VidIn, VidOut) :-
	sym_insert(call, 1, [], SymTab, Sym),
	( var(Vid) -> VidOut is VidIn + 1, Vid = VidOut
			% The input term is the 1st occurance of a variable
	; VidOut = VidIn
	).
trans_goals((A, B), and(Goal1, Goal2), SymTab, VidIn, VidOut) :- 
	trans_goals(A, Goal1, SymTab, VidIn, VidMed),
	trans_goals(B, Goal2, SymTab, VidMed, VidOut).
trans_goals((A; B), or(Goal1, Goal2), SymTab, VidIn, VidOut) :- 
	trans_goals(A, Goal1, SymTab, VidIn, VidMed),
	trans_goals(B, Goal2, SymTab, VidMed, VidOut).
trans_goals((A->B), if(Goal1, Goal2), SymTab, VidIn, VidOut) :- 
	trans_goals(A, Goal1, SymTab, VidIn, VidMed),
	trans_goals(B, Goal2, SymTab, VidMed, VidOut).
trans_goals('\+'(A), not(Goal1), SymTab, VidIn, VidOut) :- 
	trans_goals(A, Goal1, SymTab, VidIn, VidOut).
trans_goals(fail_if(A), not(Goal1), SymTab, VidIn, VidOut) :- 
	trans_goals(A, Goal1, SymTab, VidIn, VidOut).
trans_goals('_$complete'(A), '_$complete'(Goal1), SymTab, VidIn, VidOut) :- 
	trans_goal(A, Goal1, SymTab, VidIn, VidOut).
trans_goals(T, Goal, SymTab, VidIn, VidOut) :-
	trans_goal(T, Goal, SymTab, VidIn, VidOut).

trans_goal(Term, Goal, SymTab, VidIn, VidOut) :-
	functor(Term, P, A),
	( inline_predicate(P, A), Goal = inlinegoal(P, A, ArgList)
	; Goal = goal(Sym, ArgList), sym_insert(P, A, _Prop, SymTab, Sym)
	),
	Term =.. [_|RawArgs],
	trans_args(RawArgs, ArgList, SymTab, VidIn, VidOut).

trans_head(Term, goal(_Sym, ArgList), SymTab, VidIn, VidOut) :-
	Term =.. [_|RawArgs],
        trans_args(RawArgs, ArgList, SymTab, VidIn, VidOut).

/*======================================================================*/
/* trans_args(+RawArgList, -ArgList, #SymTab, +VidIn, -VidOut)		*/
/*======================================================================*/

trans_args([], [], _SymTab, VidIn, VidIn).
trans_args([RawArg|RawArgs], [Arg|ArgList], SymTab, VidIn, VidOut) :-
	trans_arg(RawArg, Arg, SymTab, VidIn, VidMed),
	trans_args(RawArgs, ArgList, SymTab, VidMed, VidOut).

trans_arg(RawArg, Arg, _SymTab, VidIn, VidOut) :- var(RawArg),
	VidOut is VidIn + 1,
	RawArg = '$$var'(VidOut),
	Arg = varocc(VidOut).
trans_arg('$$var'(ID), varocc(ID), _SymTab, VidIn, VidIn) :- integer(ID).
	% the test is needed for self-compiling! Trick.
	% when compile this module, '$$var' may mean a structure!
trans_arg(RawArg, integer(RawArg), _SymTab, VidIn, VidIn) :- integer(RawArg).
trans_arg(RawArg, real(RawArg), _SymTab, VidIn, VidIn) :- real(RawArg).
trans_arg(RawArg, constant(Sym), SymTab, VidIn, VidIn) :- atom(RawArg),
	sym_insert(RawArg, 0, _Prop, SymTab, Sym).
trans_arg(string(S), string(S), _SymTab, VidIn, VidIn) :- atom(S).
	% again, the test is needed for compile this module
trans_arg(RawArg, structure(Sym, ArgList), SymTab, VidIn, VidOut) :-
	functor(RawArg, N, A),
	RawArg =.. [_|RawArgs],
	sym_insert(N, A, _Prop, SymTab, Sym),
	trans_args(RawArgs, ArgList, SymTab, VidIn, VidOut).

