/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  peephole.P
  Author(s)		:  Saumya Debray, Kostis F. Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  May 24, 1993
========================================================================*/


/************************************************************************
*									*
*		some code for peephole optimization			*
*								 	*
*************************************************************************/

/*  "peephole_opt" is the top-level optimizer, it calls various others.	*/

compile_peephole_opt(Pil, OptPil) :-
	compile_popt1(Pil, Pil1),
	compile_popt2(Pil1, Pil2),
	compile_popt3(Pil2,Pil3),
	compile_popt4(Pil3,_,OptPil),
	!.


compile_popt1([], []).
compile_popt1([Inst|Rest], Pil1) :- compile_popt11(Inst, Rest, Pil1).

compile_popt11(puttvar(T,R), [getstr(R,S)|PilRest], [putstr(T,S)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [getstr(R,S)|PilRest], [getstr(T,S)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [puttbreg(R)|PilRest], [puttbreg(T)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [addreg(R,S)|PilRest], [addreg(T,S)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [subreg(R,S)|PilRest], [subreg(T,S)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [mulreg(R,S)|PilRest], [mulreg(T,S)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(movreg(T,R), [divreg(R,S)|PilRest], [divreg(T,S)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(putpvar(V,R), [getpval(V,R)|PilRest],
			     [putpvar(V,R)|OptPilRest]) :-
	compile_popt1(PilRest, OptPilRest).
compile_popt11(putpval(V,R), [getstr(R,Str)|PilRest],
			     [getstrv(V,Str)|OptPilRest]) :-
	\+ sym_name(Str,'.',2),		/* to enable list opt */
	compile_popt1(PilRest, OptPilRest).
compile_popt11(putpvar(V,R), [getstr(R,Str)|PilRest],
			     [putstrv(V,Str)|OptPilRest]) :-
	\+ sym_name(Str,'.',2),		/* to enable list opt */
	compile_popt1a(PilRest, OptPilRest).
compile_popt11(gettval(R,R), PRest, OptPRest) :-
	compile_popt1(PRest, OptPRest).
compile_popt11(movreg(R,R), PRest, OptPRest) :- 
	compile_popt1(PRest, OptPRest).
compile_popt11(jump(L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(jump(Addr), [jump(_)|PRest], [jump(Addr)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(jumpz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(jumpnz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(jumplt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest, OptPRest).
compile_popt11(jumple(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest, OptPRest).
compile_popt11(jumpgt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(jumpge(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1(PRest,OptPRest).
compile_popt11(Inst, PilRest, [Inst|OptPilRest]) :- 
	compile_popt1(PilRest, OptPilRest).


compile_popt1a([], []).
compile_popt1a([Inst|PilRest], Pil1) :- 
	compile_popt1a1(Inst, PilRest, Pil1).

compile_popt1a1(unipvar(X), PilRest, [bldpvar(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unipval(X), PilRest, [bldpval(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unitvar(X), PilRest, [bldtvar(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unitval(X), PilRest, [bldtval(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unicon(X), PilRest, [bldcon(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(uninumcon(X), PilRest, [bldnumcon(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unifloat(X), PilRest, [bldfloat(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(unistring(X), PilRest, [bldstring(X)|OptPilRest]) :-
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(gettval(R,R), PilRest, OptPilRest) :- 
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(movreg(R,R), PilRest, OptPilRest) :- 
	compile_popt1a(PilRest, OptPilRest).
compile_popt1a1(jump(L), [label(_,L)|PRest], [jump(_Addr)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(jump(Addr), [jump(_)|PRest], [jump(Addr)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(jumpz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(jumpnz(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(jumplt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest, OptPRest).
compile_popt1a1(jumple(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest, OptPRest).
compile_popt1a1(jumpgt(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(jumpge(_,L), [label(T,L)|PRest], [label(T,L)|OptPRest]) :-
	compile_popt1a(PRest,OptPRest).
compile_popt1a1(Inst, PilRest, OptPilRest) :- 
	compile_popt1([Inst|PilRest], OptPilRest).


/*  "compile_popt2" optimizes list instructions: (put|get)str '.' are 
    replaced by (put|get)list, and (get|put|uni|bld)con '[]' are replaced
    by (get|put|uni|bld)nil, respectively.				*/

compile_popt2([], []).
compile_popt2([Inst|PilRest], Pil1) :- compile_popt21(Inst, PilRest, Pil1).

compile_popt21(getstr(R,Str), PilRest, [getlist(R)|OptPilRest]) :-
	sym_name(Str,'.',2),
	compile_popt2(PilRest, OptPilRest).
compile_popt21(putstr(R, Str), PilRest, [putlist(R)|OptPilRest]) :-
	sym_name(Str,'.',2),
	compile_popt2(PilRest, OptPilRest).

compile_popt21(getcon(R,Sym), PilRest, [getnil(R)|OptPilRest]) :-
	sym_name(Sym,[],0),
	compile_popt2(PilRest, OptPilRest).
compile_popt21(putcon(R,Sym), PilRest, [putnil(R)|OptPilRest]) :-
	sym_name(Sym,[],0),
	compile_popt2(PilRest, OptPilRest).
compile_popt21(unicon(Sym), PilRest, [uninil|OptPilRest]) :-
	sym_name(Sym,[],0),
	compile_popt2(PilRest, OptPilRest).
compile_popt21(bldcon(Sym), PilRest, [bldnil|OptPilRest]) :-
	sym_name(Sym,[],0),
	compile_popt2(PilRest, OptPilRest).
compile_popt21(Inst, PilRest, [Inst|OptPilRest]) :-
	compile_popt2(PilRest, OptPilRest).


compile_popt3([],[]).
compile_popt3([Inst|Rest], Pil) :- compile_popt31(Inst, Rest, Pil).

compile_popt31(getlist(R0), [unitvar(R1),unitvar(R2)|Rest],
      			    [getlist_tvar_tvar(R0,R1,R2)|OptRest]) :-
	compile_popt3(Rest, OptRest).
compile_popt31(getcomma(R0), [unitvar(R1),unitvar(R2)|Rest],
      			     [getcomma_tvar_tvar(R0,R1,R2)|OptRest]) :-
	compile_popt3(Rest, OptRest).
compile_popt31(Inst, Rest, [Inst|OptRest]) :- 
	compile_popt3(Rest, OptRest).

/*  "popt4" eliminates some redundant "execute"s, mainly those arising from
    the generation of new predicates for handling cuts and indexing.   */

compile_popt4([],_,[]).
compile_popt4([label(T,(Sym,K))|Rest],Seen,[label(T,(Sym,K))|ORest]) :-
	!,
	( Sym = disj(_)
	; member1(Sym,Seen)
	),
	compile_popt4(Rest,Seen,ORest).
compile_popt4([execute(Sym),label(T,(Sym,K))|Rest],Seen,OList) :-
	!,
	( Sym = disj, OList = [execute(Sym),label(T,(Sym,K))|ORest]
	; /* Sym = sym(...) */
	  popt_chkmember(Sym,Seen,SFlag),
	  ( SFlag =:= 1,	/* already seen, "execute" can't be dropped */
	    OList = [execute(Sym),label(T,(Sym,K))|ORest]
	  ; SFlag =\= 1,	/* not seen, "execute" can be dropped */
	    OList = [label(T,(Sym,K))|ORest]
	  )
	),
	compile_popt4(Rest,Seen,ORest).
compile_popt4([Inst|Rest],Seen,[Inst|ORest]) :-
	compile_popt4(Rest,Seen,ORest).

popt_chkmember(P,L,Flag) :-
	( var(L) -> L = [P|_], Flag = 0
	; L = [P1|L1], ( P = P1 -> Flag = 1 ; popt_chkmember(P,L1,Flag) )
	).


force_peephole_loaded.
