/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  printresult.P
  Author(s)		:  Saumya Debray, Terrance Swift, Kostis Sagonas
  Modified by		:  Jiyang Xu
  Last modification	:  May 23, 1993
========================================================================*/


print_result(_Module) :-
	option(show_et),
	show_et, 
	fail.
print_result(Module) :-
	ttywritenl('***** Exported symbols ***************'),
	get_symtab(Module, SymTab),
	print_result0(SymTab, ex),
	fail.
print_result(Module) :-
	option(print_local),
	ttywritenl('***** Local symbols ***************'),
	get_symtab(Module, SymTab),
	print_result0(SymTab, (local)),
	fail.
print_result(_Module).

print_result0(SymTab, _Domain) :- option(typeinfer), !,
	print_result1(SymTab, _Env).
print_result0(SymTab, Domain) :-
	extract_symlist(Domain, SymList, SymTab),
	write_symlist(SymList).

write_symlist([]) :- nl.
write_symlist([Sym|Rest]) :- 
	sym_name(Sym, N, A),
	write(N/A), write(' '),
	write_symlist(Rest).
	
print_result1(SymTab, Env) :-
	sym_scan(Sym, SymTab, RTab),
	sym_name(Sym, Name, Arity),
	!,
	( sym_prop(Env, Sym) ->
		( sym_prop(type(Type), Sym) ->
			print_typing(Name, Arity, Type)
		; ttywritenl(Name/Arity)
		)
	; true
	),
	print_result1(RTab, Env).
print_result1(_SymTab, _Env).
/*======================================================================*/
/* print_typing(+Name, +Arity, +Type).					*/
/* print the types of preds which are inferred by the system		*/
/*======================================================================*/
print_typing(_P, _A, []) :- !.
print_typing(P, A, [Typing|Typings]) :- !,
	freeze(Typing),				% To print in a better looking
	write('  '), write(P/A), write(' ---  ['),
	( Typing = arrow(ArgType, ResType) ->
		write_typing(0, A, ArgType),
		write(' --> '), write(ResType)
	; write_typing(0, A, Typing)
	),
	write(' ]'), nl,
	print_typing(P, A, Typings), !.
print_typing(P, A, arrow(ArgType, ResType)) :- !,
	freeze(arrow(ArgType, ResType)),	% To print in a better looking
	write('  '), write(P/A), write(' ---  ['),
	write_typing(0, A, ArgType), write(' --> '), 
	write(ResType),
	write(' ]'), nl, !.
print_typing(P, A, im(Mod)) :- !,
	( Mod = system -> true
	; ttywritenl(('  ', P/A, ' ---     imported from ', Mod))
	).
print_typing(P, A, Type) :- ttywritenl(('  ', P/A, ' ---        ', Type)).

write_typing(A, A, _) :- !.
write_typing(I, A, Typing) :-
	I1 is I + 1,
	( I1 > 1 -> write(' X ')
	; tab(1)
	),
	arg(I1, Typing, One),
	write(One),
	write_typing(I1, A, Typing).

/*======================================================================*/
/* show_et								*/
/*======================================================================*/
show_et :-
	ttywritenl('/** Extension Table **/'),
	recorded(et_key(_), Answer, _),
	functor(Answer, P, A),
	ttywritenl((' -- ', P/A, '   ...  ', Answer)),
	fail.
show_et :- ttywritenl('/** End Extension Table **/').

