/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  singleton.P
  Author(s)		:  Kostis F. Sagonas
  Last modification	:  July 15, 1992
========================================================================*/


/*======================================================================*/
/*  singleton_check(+Clause, +VariableList).				*/
/*======================================================================*/

singleton_check(_, []) :- 
	!.	% This captures clauses with no variables to be checked, 
		% (for example ground clauses, or clauses that contain
		% only underscores as variables), where VariableList is
		% variable and no singleton test is needed.
singleton_check(Clause, VariableList) :-
	singleton_warnings(Clause, VariableList),
	fail.	% Undo numbervars' side-effects and reclaim all space used.
singleton_check(_, _).


singleton_warnings(Clause, VariableList) :-
	numbervars(Clause, 1, NumberOfVars),
	N is NumberOfVars - 1,
	functor(Occurences, occ, N),
	calculate_occurences(Clause, Occurences),
	Clause = (Head :- _),
	functor(Head, P, A),
	write_singleton_warnings(VariableList, Occurences, P/A).


calculate_occurences(Term, Occurences) :-
	functor(Term, _, Arity),
	'calculate occurences'(Term, 0, Arity, Occurences).


'calculate occurences'(_, N, N, _) :- 
	!.
'calculate occurences'(Term, N, Arity, Occurences) :-
	NewN is N + 1,
	arg(NewN, Term, Arg),	% We cannot use ll_arg/3 here (Term may be a list)
	( Arg = '$VAR'(Number) ->
		( integer(Number) ->
			ll_arg(Number, Occurences, OccCount),
			( var(OccCount) -> OccCount = found(_)
			; OccCount = found(Times) -> 
				( var(Times) -> Times = 'at least twice'
				; true
				)
			)
		; functor(Arg, _, ArgArity),
		  'calculate occurences'(Arg, 0, ArgArity, Occurences)
		)
	; atomic(Arg) -> 
		true
	; functor(Arg, _, ArgArity),
	  'calculate occurences'(Arg, 0, ArgArity, Occurences)
	),
	'calculate occurences'(Term, NewN, Arity, Occurences).


write_singleton_warnings([], _, _) :- 
	!.	% This captures the open-end of VariableList.
write_singleton_warnings([vv(VarName,'$VAR'(N))|Rest], Occurences, Pred) :-
	name(VarName, CharList),
	( CharList = [95|_] ->	% variable Var begins with an "_"
		true		% (underscore), so its OK, ignore it!.
	; ll_arg(N, Occurences, OccCount),
	  ( OccCount == found('at least twice') -> true
	  ; warning(('Singleton variable ', VarName, ' in a clause of ', Pred))
	  )
	),
	write_singleton_warnings(Rest, Occurences, Pred). 

