/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tabdef.P
  Author(s)		:  Kate Dvortsova
  Last modification	:  December,7, 1992
  ======================================================================*/

/* true, if Pred call Sym */

direct_call(Pred,Sym):-
          Pred=pred(_,Clist,_),
          call_clist(Sym,Clist).

call_clist(_,[]):-!,fail.
call_clist(Sym,[clause(_,Goal,_)|L]):-
           ( call_goal(Sym,Goal) -> true ;
             call_clist(Sym,L)).

call_goal(Sym,goal(Sym,_)):-!.
call_goal(Sym,and(Goal1,Goal2)):-
          ( call_goal(Sym,Goal1) -> true ;
            call_goal(Sym,Goal2) ).
call_goal(Sym,or(Goal1,Goal2)):-
          ( call_goal(Sym,Goal1) -> true ;
            call_goal(Sym,Goal2) ).
call_goal(Sym,if(Goal1,Goal2)):-
          ( call_goal(Sym,Goal1) -> true ;
            call_goal(Sym,Goal2) ).
call_goal(Sym,not(Goal)):-
            call_goal(Sym,Goal).
call_goal(Sym,Goal):-
            meta_pred(Goal,Args),
            call_args(Args,Sym,_).
/**********************************************************/

/* true, if there is a path from Sym1 to Sym2 */

path(Sym1,_,_):-
            sym_prop(tabled(_,_,_,_),Sym1),!,
            fail.                            /* this loop has already broken */
path(Sym1,Sym2,Rest):-
            member(pred(Sym1,Clist,_),Rest),
            exclude(Sym1,Rest,Rest1),
            clist_path(Sym1,Clist,Sym2,Rest1).

/* 1-parameter Sym1 is needed to exclude direct recursion */

clist_path(_,[],_,_):-!,fail.
clist_path(Sym1,[clause(_,Goal,_)|L],Sym,Rest):-
             ( goal_path(Sym1,Goal,Sym,Rest) -> true ;
               clist_path(Sym1,L,Sym,Rest) ).

goal_path(Sym,goal(Sym,_),Sym,_):-!,fail.
goal_path(_,goal(Sym,_),Sym,_):-!.
goal_path(_,inlinegoal(_,_,_),_,_):-!,fail.

goal_path(Sym1,Goal,Sym,Rest):-                       /* meta-predicates */
               meta_pred(Goal,Args),
               args(Sym1,Args,Sym,Rest),!.
	   
goal_path(_,goal(Sym1,_),_,_):-
               Sym1=sym(Name,Arity,_,_),
               standard_symbol(Name,Arity,_),!,fail.

goal_path(_,goal(Sym1,_),Sym2,Rest):-!,
               path(Sym1,Sym2,Rest).

goal_path(Sym1,and(Goal1,Goal2),Sym,Rest):-
               ( goal_path(Sym1,Goal1,Sym,Rest) -> true ;
                 goal_path(Sym1,Goal2,Sym,Rest)).

goal_path(Sym1,or(Goal1,Goal2),Sym,Rest):-
               ( goal_path(Sym1,Goal1,Sym,Rest) -> true ;
                 goal_path(Sym1,Goal2,Sym,Rest)).
         
goal_path(Sym1,if(Goal1,Goal2),Sym,Rest):-
               ( goal_path(Sym1,Goal1,Sym,Rest) -> true ;
                 goal_path(Sym1,Goal2,Sym,Rest)).

goal_path(Sym1,not(Goal),Sym,Rest):-
               goal_path(Sym1,Goal,Sym,Rest).

/******************************************************************/
exclude(_,[],[]).
exclude(Sym,[pred(Sym,_,_)|L],L):-!.
exclude(Sym,[X|L],[X|L1]):-
       exclude(Sym,L,L1).

member(_,[]):-!,fail.
member(El,[El|_]):-!.
member(El,[_|L]):-
       member(El,L).
/****************************************************************/
/* How many loops ? */

score([],[],_).
score([Pred|L],[(s(Sym,N))|L1],Rest):-
        Pred=pred(Sym,_,_),
        score_sym(Pred,Rest,0,N,Rest),!,
        score(L,L1,Rest).

score_sym(_,[],N,N,_):-!.
score_sym(Pred,[Pred|L],N1,N2,Rest):-!,  /* recursion has already removed */  
          score_sym(Pred,L,N1,N2,Rest).
score_sym(Pred,[pred(Sym1,_,_)|L],N,N2,Rest):-
          Pred=pred(Sym,_,_),
          ( direct_call(Pred,Sym1)  -> (how_many(Sym,Sym1,Rest,0,K,Rest),
             N1 is N+K) ; N1 is N),!,
            score_sym(Pred,L,N1,N2,Rest).

how_many(_,_,[],K,K,_).
how_many(Sym,Sym1,[Pred|L],K,N,Rest):-
          Pred=pred(Sym2,_,_),
          sym_name(Sym1,Name1,Arity1),
          sym_name(Sym2,Name2,Arity2), 
          ((direct_call(Pred,Sym),
           (path(Sym1,Sym2,Rest) ;(Name1=Name2,Arity1=Arity2)) ) ->
           K1 is K+1 ; K1 is K),!,
           how_many(Sym,Sym1,L,K1,N,Rest).

choice([],s(Sym,_),Sym).
choice([T|L],T1,X):-
           T=s(sym(_,A,_,_),N),
           T1=s(sym(_,A1,_,_),N1),
          ( N>N1 -> choice(L,T,X) ;
            ((N=N1,A<A1) -> choice(L,T,X) ;
                    choice(L,T1,X))).
/**************************************************************/
direct_rec([],[]).
direct_rec([pred(Sym,Clist,_)|L],L1):-
           need_tab(Sym,Clist,Clist,_),
           sym_propin(tabled(_,_),Sym),
	   sym_name(Sym, Name, Arity),
           message(('% Compiling predicate ', 
		    Name, '/', Arity, ' as tabled predicate')),
           direct_rec(L,L1).
direct_rec([Pred|L],[Pred|L1]):-
           direct_rec(L,L1).

define_table(Module):-
           Module=module(_,_,_,_,List,_),
           direct_rec(List,Rest),
         %  write('direct_rec is ok'),nl,
           reduce_loop(Rest).

reduce_loop([]).
reduce_loop(Rest):-
           call_itself(Rest,Rest1,Rest),
           ( Rest1=[] -> true ;
           (score(Rest1,Srest1,Rest1),
           Srest1=[S|_],
           choice(Srest1,S,Sym),
           sym_propin(tabled(_,_),Sym),
	   sym_name(Sym, Name, Arity),
           message(('% Compiling predicate ', 
		    Name, '/', Arity, ' as tabled predicate')),
           exclude(Sym,Rest1,Rest2),
           reduce_loop(Rest2))).

call_itself([],[],_).
call_itself([Pred|L],[Pred|L1],Rest):-
           Pred=pred(Sym,_,_),
           path(Sym,Sym,Rest),!,
           call_itself(L,L1,Rest).
call_itself([_|L],L1,Rest):-
           call_itself(L,L1,Rest).


/*********************************************************************/

            /* Meta-predicates */
meta_pred(goal(sym(call,1,_,_),Args),Args):-!.
meta_pred(goal(sym(Name,3,_,_),Args),Args):-
         Name=='findall' ; Name=='setof' ; Name=='bagof',!.
meta_pred(goal(sym(not,1,_,_),Args),Args):-!.

args(_,[],_,_):-!,fail.
args(Sym0,[Arg|Args],Sym,Rest):-
         (argument(Sym0,Arg,Sym,Rest) 
                               -> true ;
           args(Sym0,Args,Sym,Rest) ).

argument(Sym0,constant(Sym1),Sym,Rest):-
         goal_path(Sym0,goal(Sym1,_),Sym,Rest),!.

argument(Sym0,structure(Sym1,Args),Sym,Rest):-
         ( goal_path(Sym0,goal(Sym1,_),Sym,Rest) -> true ;
            args(Sym0,Args,Sym,Rest) ).

call_args([],_,_):-!,fail.                   /* direct call */
call_args([Arg|Args],Sym,Sarg):-
          ( carg(Arg,Sym,Sarg) -> true ;
            call_args(Args,Sym,Sarg) ).
carg(constant(Sym),Sym,_):-!.
carg(structure(Sym,Args),Sym1,Sarg):-
           ( Sym==Sym1 -> (true, Sarg=Args) ;
             call_args(Args,Sym1,Sarg) ).
/************************************************************************/
/* true, if tabling is needed for recursive procedure */

need_tab(_Sym,[],_Clist,_Term):-!,fail.    % all clauses have been checked
need_tab(Sym,[Clause|L],Clist,Term):-
        Clause=clause(Args,Goal,_),
	( ok_clause(Sym,Args,Goal,R) ->
		(var(R) -> need_tab(Sym,L,Clist,Term) ;   %no recursion
		   (var(Term) ->        
		(terminate(Clause,Clist) -> need_tab(Sym,L,Clist,t) ; true)) ;
                    need_tab(Sym,L,Clist,Term))           %T=t
        ; true) .

/* true, if the clause without recursion or recursion is with diff. args */
/* 4-argument=r, if it is recursion, otherwise it's variable */
ok_clause(_Sym,_Arg,inlinegoal(_,_,_),_):-!.
ok_clause(Sym,Arg1,Goal,R):-
	meta_pred(Goal,Args),
        (call_args(Args,Sym,Sarg) ->
        (!,R=r,ok_arg(Arg1,Sarg)) ; true).
ok_clause(Sym1,_,goal(Sym2,_),_):-
	Sym1\=Sym2,!.
ok_clause(Sym,Arg1,goal(Sym,Arg2),r):-
	sym_name(Sym,_,Arity),
	(Arity=0 -> fail ; ok_arg(Arg1,Arg2)).
ok_clause(Sym,Arg,not(Goal),R):-
	ok_clause(Sym,Arg,Goal,R).

ok_clause(Sym,Arg,and(Goal1,Goal2),R):-
	ok_clause(Sym,Arg,Goal1,R),
	ok_clause(Sym,Arg,Goal2,R).

ok_clause(Sym,Arg,or(Goal1,Goal2),R):-
	ok_clause(Sym,Arg,Goal1,R),
	ok_clause(Sym,Arg,Goal2,R).

ok_clause(Sym,Arg,if(Goal1,Goal2),R):-
	ok_clause(Sym,Arg,Goal1,R),
	ok_clause(Sym,Arg,Goal2,R).

/* true, if arguments of the call are different */
ok_arg([],[]):-!,fail.
ok_arg([Arg1|L1],[Arg2|L2]):-
	(diff(Arg1,Arg2) -> true ; ok_arg(L1,L2)).

diff(structure(_,Args),varocc(X)):-
	check_str(X,Args),!.
diff(varocc(X),structure(_,Args)):-
	check_str(X,Args).
/* looking for variable in the structure */
check_str(_,[]):-!,fail.
check_str(X,[varocc(X)|_]):-!.
check_str(X,[structure(_,Args)|L]):-
	(check_str(X,Args) -> true ;
	  check_str(X,L)).
check_str(X,[_|L]):-
	check_str(X,L).

/* looking for teminate condition */
/* true if there is a clause before Clause, which can terminate recursion*/
terminate(Clause,[Clause|_]):-!,fail.
terminate(Clause,[clause(_,Goal,_)|L]):-
	(term_clause(Goal) -> true ;
         terminate(Clause,L)).
term_clause(inlinegoal(_,_,_)):-!.
term_clause(goal(Sym,Args)):-
	meta_pred(goal(Sym,Args),_),!,fail.
term_clause(goal(Sym,_)):-
	sym_name(Sym,Name,Arity),
	standard_symbol(Name,Arity,_),!.
term_clause(goal(Sym,_)):-
	sym_prop(tabled(_,_,_,_),Sym).	
term_clause(and(Goal1,Goal2)):-
	term_clause(Goal1),
	term_clause(Goal2).

term_clause(or(Goal1,Goal2)):-
	term_clause(Goal1),
	term_clause(Goal2).

term_clause(if(Goal1,Goal2)):-
	term_clause(Goal1),
	term_clause(Goal2).
/*********************************************************************/

