/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tp_cond.P
  Author(s)		:  Saumya Debray, Terrance Swift, Kostis Sagonas
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  September 19, 1992
========================================================================*/



/*======================================================================*/
/* tvar_hold(+TestGoal,	+HoldRegsIn, -HoldRegsOut)			*/
/*	It collects all the temp vars used in the test and append	*/
/*	their id in HoldRegs, since these temp vars should not be	*/
/*	released in the "then" part until the next chunk.		*/
/*======================================================================*/

tvar_hold(and(Goal,Goals,_),HoldIn,HoldOut) :-
    tvar_hold(Goal,HoldIn,HoldMed),
    tvar_hold(Goals,HoldMed,HoldOut).
tvar_hold(or(Goal,Goals,_),HoldIn,HoldOut) :-
    tvar_hold(Goal,HoldIn,HoldMed),
    tvar_hold(Goals,HoldMed,HoldOut).
tvar_hold(inlinegoal(_,_,Args,_),HoldIn,HoldOut) :-
    tvar_hold_args(Args,HoldIn,HoldOut).

tvar_hold_args([],HoldIn,HoldIn).
tvar_hold_args([Arg|Args],HoldIn,HoldOut) :-
	tvar_hold_arg(Arg,HoldIn,HoldMed),
	tvar_hold_args(Args,HoldMed,HoldOut).

tvar_hold_arg(integer(_),  HoldIn,HoldIn).
tvar_hold_arg(constant(_), HoldIn,HoldIn).
tvar_hold_arg(string(_),   HoldIn,HoldIn).
tvar_hold_arg(structure(_,Args),HoldIn,HoldOut) :- 
	tvar_hold_args(Args,HoldIn,HoldOut).
tvar_hold_arg(varocc(Vid,Prag),HoldIn,HoldOut) :-
	( vprag_type(Prag, t) -> HoldOut=[Vid|HoldIn] ; HoldOut=HoldIn ).

/*======================================================================*/
/* tcond(+TestGoal, +SuccessLabel, +FailLabel, 				*/
/*	 -Pil-Pilr, +Tin, -Tout, +HoldRegs)				*/
/* 	Generates code for the test part of an 'if_then_else'.		*/
/*======================================================================*/

tcond(and(Goal,Goals,_),SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold) :-
    gennum(LabId), IntSuccLab = label(_,(disj,LabId)),
    tcond(Goal,IntSuccLab,FailLab,Pil,[IntSuccLab|Pilm],Tin,Tmid,Hold),
    tcond(Goals,SuccLab,FailLab,Pilm,Pilr,Tmid,Tout,Hold).
tcond(or(Goal,Goals,_),SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold) :-
    gennum(LabId), IntFailLab = label(_,(disj,LabId)),
    tcond(Goal,SuccLab,IntFailLab,Pil,
    			[IntFailLab|Pilm],Tin,Tmid,Hold),
    tcond(Goals,SuccLab,FailLab,Pilm,Pilr,Tmid,Tout,Hold).
tcond(inlinegoal(Op,_,Args,_),SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold) :-
    ( Args = [Arg1,Arg2], arithrelop(Op),
	tcond_relop(Op,Arg1,Arg2,SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold)
    ; Args = [Arg],		% Op = var/nonvar (only for the moment)
	tcond_sot(Op,Arg,SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold)
    ; Args = [], Op = true, Pil = Pilr, Tin = Tout
    ).


tcond_relop(Op,Arg1,Arg2,SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold) :-
    SuccLab = label(_,SL0), FailLab = label(_,FL0),
    eval_exp(Arg1,R1,Pil,Pilm,Hold,Tin,Tmid),
    eval_exp(Arg2,R2,Pilm,
    		[subreg(R2,R1),Jumpinst,jump(SL0)|Pilr],Hold,Tmid,Tout1),
    eval_arithreloptab(Op,R1,FL0,Jumpinst),
    release_if_done0(Arg1,R1,Hold,Tout1,Tout2),
    release_if_done0(Arg2,R2,Hold,Tout2,Tout).

tcond_sot(Op,varocc(Vid,Prag),SuccLab,FailLab,Pil,Pilr,Tin,Tout,Hold) :-
    alloc_reg(Prag, Tin, Tmid),
    ( vprag_type(Prag,t), vprag_loc(Prag,L), R = L, Tmid = Tmid1
    ; getreg(Tmid,R),hold(R,Tmid,Tmid1)
    ),
    putvarinst(Prag,R,Pil,Pilm,Tmid1),
    SuccLab = label(_,SL0),
    FailLab = label(_,FL0),
    ( Op == var -> Pilm = [switchonterm(R,FL0,FL0)|Pilr]
    ; /* nonvar */ Pilm = [switchonterm(R,SL0,SL0),jump(FL0)|Pilr]
    ),
    release_if_done(Vid,R,Prag,Hold,Tmid1,Tout).
tcond_sot(Op,_Arg,_SuccLab,_FailLab,Pil,Pilr,_Tin,_Tout,_Hold) :-
     warning(('argument to ', Op, '/1 is not a variable')),
     ( Op == var -> Pil = [fail|Pilr] ; /* nonvar */ Pil = Pilr ).

