/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tp_geninline.P 
  Author(s)		:  Saumya Debray, Kostis Sagonas, Terrance Swift
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  May 21, 1993
========================================================================*/

/*======================================================================*/
/* geninline(+Pname, +Args, HoldRegs, -Pil, -PilRest, +Tin, -Tout)	*/
/*	Generate code for inline predicates.				*/
/*		+ Args: in intermediate program format.			*/
/*		- Pil-PilRest: pil code generated, as a differential	*/
/*			       list.					*/
/*======================================================================*/

geninline(true, [], _, Pil, Pil, T, T).
geninline(fail, [], _, [fail|Pil], Pil, T, T).
geninline(halt, [], _, [halt|Pilr], Pilr, T, T).

geninline(var, [Arg], HoldR, Pil, Pilr, Tin, Tout) :-
	geninl_load_lhs(Arg, R, Pil, 
			[switchonterm(R,abs(-1),abs(-1))|Pilr], Tin, Tmid, _),
	geninl_unload_lhs(Arg, R, HoldR, Tmid, Tout).
geninline(nonvar, [Arg], HoldR, Pil, Pilr, Tin, Tout) :-
	geninl_load_lhs(Arg, R, Pil, Pil1, Tin, Tmid, _),
	gennum(LabId), Label = (disj(LabId),LabId),
	Pil1 = [switchonterm(R,Label,Label),fail,label(swit,Label)|Pilr],
	geninl_unload_lhs(Arg, R, HoldR, Tmid, Tout).
geninline('_$builtin', [integer(N)], _, [builtin(N)|Pilr], Pilr, T, T).
geninline('_$savecp', [Arg], _, Pil, Pilr, Tin, Tout) :-
	geninline_savecp(Arg, Pil, Pilr, Tin, Tout).
geninline('_$cutto', [Arg], _, Pil, Pilr, Tin, Tout) :-
	geninline_cutto(Arg, Pil, Pilr, Tin, Tout).
geninline('_$return', [Arg], _, Pil, Pilr, Tin, Tout) :-
   	tputpar(Arg, 0, Pil, Pilr, Tin, Tout).
geninline('_$getreturn', [Arg], _, Pil, Pilr, Tin, Tout) :-
   	tgetpar(Arg, 0, Pil, Pilr, Tin, Tout).

geninline(is, [Arg1,Arg2], HoldR, Pil, Pilr, Tin, Tout) :-
	geninline_is(Arg1, Arg2, HoldR, Pil, Pilr, Tin, Tout).
geninline(=, [Lhs,Rhs], HoldR, Pil, Pilr, Tin, Tout) :-
	geninl_load_lhs(Lhs, R, Pil, Pil1, Tin, Tmid0, SaveReg),
	release_if_tvar_if_done(Rhs, R, HoldR, Tmid0, Tmid1),
	tgetpar(Rhs, R, Pil1, Pilr, Tmid1, Tmid2),
	append(SaveReg, Tmid2, Tmid3),
	geninl_unload_lhs(Lhs, R, HoldR, Tmid3, Tout).
geninline(Arithrelop, [Arg1,Arg2], HoldR, Pil, Pilr, Tin, Tout) :-
	arithrelop(Arithrelop),
	eval_exp(Arg1, R1, Pil, Pilm, HoldR, Tin, Tmid),
	eval_exp(Arg2, R2, Pilm,[subreg(R2,R1),JumpInst|Pilr],HoldR,Tmid,Tout1),
	eval_arithreloptab(Arithrelop, R1, abs(-1), JumpInst),
	(Arg1 = varocc(Vid1,Prag1) ; (Vid1 = nonvbl, Prag1 = [])),
	(Arg2 = varocc(Vid2,Prag2) ; (Vid2 = nonvbl, Prag2 = [])),
	release_if_done(Vid1, R1, Prag1, HoldR, Tout1, Tout2),
	release_if_done(Vid2, R2, Prag2, HoldR, Tout2, Tout).

/*======================================================================*/
/*======================================================================*/

geninl_load_lhs(Term,R,Pil,Pilr,Tin,Tout,SaveR) :-
	( Term = varocc(_Vid,Prag) ->
		alloc_reg(Prag, Tin, Tmid), vprag_type(Prag, T),
		vprag_occ(Prag, Oc), vprag_loc(Prag, Loc),
		( T = t -> R = Loc, Tout = Tmid,
		  ( (Oc=l; Oc=v), SaveR = [] ; SaveR = [R] )
		; getreg(Tmid, R), hold(R, Tmid, Tout), SaveR = []
		),
		putvarinst(Prag, R, Pil, Pilr, Tout)
	; getreg(Tin, R), hold(R, Tin, Tmid), SaveR = [],
	  tputpar(Term, R, Pil, Pilr, Tmid, Tout)
	).
%   telling(X),tell(user),
%   write('called with term: '),write(Term),
%   write('  Tin= '),write(Tin),write('  Tout= '),write(Tout),nl,
%   write('  Pil= '),write(Pil), write(' Pilr= '),write(Pilr),nl,told,tell(X).
%
%:- import tell/1, telling/1, told/0, write/1, nl/0 from standard.


geninl_unload_lhs(Term, R, HoldR, Tin, Tout) :-
	( Term = varocc(Vid,Prag) -> 
		release_if_done(Vid, R, Prag, HoldR, Tin, Tout)
	; release(R, Tin, Tout)
	).

/*======================================================================*/
/*======================================================================*/

geninline_is(LHS,Exp,HoldR,Pil,Pilr,Tin,Tout) :-
	( LHS = varocc(_Vid,_Prag) ->
		eval_exp(Exp,R,Pil,Pilm,HoldR,Tin,Tout1),
		release_if_tvar_if_done(LHS,R,HoldR,Tout1,Tout2),
		tgetpar(LHS,R,Pilm,Pilr,Tout2,Tout)
	; LHS = integer(Int) ->
		getreg(Tin, R1), hold(R1, Tin, Tmid),
		Pil = [putnumcon(R1,Int)|Pil1],
		eval_exp(Exp,R,Pil1,[subreg(R,R1),jumpnz(R1,abs(-1))|Pilr],
			 HoldR,Tmid,Tout1),
		release(R1,Tout1,Tout2), release(R,Tout2,Tout)
	; LHS = real(Real) ->
		getreg(Tin, R1), hold(R1, Tin, Tmid),
		Pil = [putfloat(R1,Real)|Pil1],
		eval_exp(Exp,R,Pil1,[subreg(R,R1),jumpnz(R1,abs(-1))|Pilr],
			 HoldR,Tmid,Tout1),
		release(R1,Tout1,Tout2), release(R,Tout2,Tout)
	; error('variable or numerical constant expected as 1st arg of is/2'),
	  Pil = [fail | Pilr], Tin = Tout
	). 

/*======================================================================*/
/*======================================================================*/

geninline_savecp(varocc(_,Prag), Pil, Pilr, Tin, Tout) :-
	( vprag_type(Prag, t) ->
		Pil = [gettbreg(R)|Pilr], alloc_reg(Prag, Tin, Tout)
	; Pil = [getpbreg(R)|Pilr], Tin = Tout
	),
	vprag_loc(Prag, R).

geninline_cutto(varocc(_,Prag), Pil, Pilr, Tin, Tout) :-
	( vprag_type(Prag, t) ->
		vprag_loc(Prag, R), vprag_occ(Prag, Oc),
		alloc_reg(Prag, Tin, Tmid), Pil = [puttbreg(R)|Pilr],
		( (Oc=l; Oc=v), release(R, Tmid, Tout) ; Tmid = Tout )
	; Pil = [putpbreg(R)|Pilr], Tin = Tout, vprag_loc(Prag, R)
	).

/*======================================================================*/
/*======================================================================*/

release_if_tvar_if_done(varocc(Vid,Prag),N,HoldR,Tin,Tout) :- !,
	release_if_done(Vid,N,Prag,HoldR,Tin,Tout).
release_if_tvar_if_done(_,_,_HoldR,Tin,Tin).


/*======================================================================*/
/*======================================================================*/

force_tp_geninline_loaded.

