/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tp_goal.P
  Author(s)		:  Saumya Debray, Kostis Sagonas, Terrance Swift
  Modified by		:  Jiyang Xu, David S. Warren
  Last modification	:  July 11, 1992
========================================================================*/


/*======================================================================*/
/* tpar/7 generate PIL code (gets, puts, blds, unis) for a parameter	*/
/* of a goal. tgoalargs/6 loops through args and calls tpar/7.		*/
/*									*/
/* tgoalargs(+Args, +N, -Pil, -Pilr, +Tin, -Tout)			*/
/* tWpar(+Arg, +N, -Pil, -Pilr, +Tin, -Tout)				*/
/*	+Args and +Arg: Args in intermediate format			*/
/*	+N: the Nth parameter of the goal				*/
/*	W: is get or put						*/
/*======================================================================*/
tgoalargs([],_,Pil,Pil,T,T).
tgoalargs([Arg|Args],N,Pil,Pilr,Tin,Tout) :-
	((\+ reg_unchanged(Arg,N),rval_needed(Args,N)) % (reg is not released)
	 ->	getreg(Tin,Nreg), T1 = [Nreg|Tin],
		Pil = [movreg(N,Nreg)|Pil1],
		load_from_nreg(Args,Nargs,N,Nreg)
	 ;	Nargs = Args, T1 = Tin, Pil1 = Pil
	),
				%	Nargs=Args,Pil=Pil1,Tin=T1,
	tputpar(Arg,N,Pil1,Pil2,T1,T2),
	N1 is N + 1,
	tgoalargs(Nargs,N1,Pil2,Pilr,T2,Tout).

% check to see if we are indeed going to overwrite this register
reg_unchanged(varocc(_Vid,Prag),N) :-
	vprag_type1(Prag,t), vprag_loc(Prag,N).

% check to see if contents of register we are about to overwrite are needed.
rval_needed([varocc(_Vid,Prag)|_Args],N) :-
	vprag_type1(Prag, t),
	vprag_loc(Prag,M), nonvar(M), M=:=N,
	vprag_occ(Prag, Occ),
	(Occ=l; Occ=s),
	!.
rval_needed([structure(_,Sargs)|_Args],N) :-
	rval_needed(Sargs,N),!.
rval_needed([_|Args],N) :- rval_needed(Args,N).

% Have moved contents of N to Nreg, now update descriptions
load_from_nreg([],[],_,_).
load_from_nreg([varocc(Vid,Prag)|Args],[varocc(Vid,Nprag)|Nargs],N,Nreg) :-
	vprag_type1(Prag, t), vprag_loc(Prag,M), nonvar(M), N=:=M,
	vprag_occ(Prag, Occ), (Occ=l; Occ=s), !,
	vprag_context(Prag,C), vprag_level(Prag,L), vprag_argno(Prag,A),
	vprag_lastchunk(Prag,Lc), vprag_type(Prag,Ty),
	vprag_use(Prag,Us), vprag_nouse(Prag,Nou),
	% now set new
	vprag_type(Nprag, t), vprag_loc(Nprag,Nreg), vprag_occ(Nprag, Occ),
	vprag_context(Nprag,C), vprag_level(Nprag,L), vprag_argno(Nprag,A),
	vprag_lastchunk(Nprag,Lc), vprag_type(Nprag,Ty),
	vprag_use(Nprag,Us), vprag_nouse(Nprag,Nou),
	load_from_nreg(Args,Nargs,N,Nreg).
load_from_nreg([structure(Sym,Sargs)|Args],[structure(Sym,Nsargs)|Nargs],
								N,Nreg) :-
	!, 
	load_from_nreg(Sargs,Nsargs,N,Nreg),
	load_from_nreg(Args,Nargs,N,Nreg).
load_from_nreg([Arg|Args],[Arg|Nargs],N,Reg) :-
	load_from_nreg(Args,Nargs,N,Reg).


tputpar(integer(V),N,[putnumcon(N,V)|Pil],Pil,Tin,Tin).
tputpar(real(V),N,[putfloat(N,V)|Pil],Pil,Tin,Tin).
tputpar(constant(Sym),N,[putcon(N,Sym)|Pil],Pil,Tin,Tin).
tputpar(string(Sym),N,[putstring(N,Sym)|Pil],Pil,Tin,Tin).
tputpar(varocc(_Vid,Prag),N,Pil0,Pil,Tin,Tout) :-
	( vprag_type(Prag, t), alloc_reg1(Prag,N,Tin,Tout) ; Tin = Tout ),
	putvarinst(Prag,N,Pil0,Pil,Tout).
tputpar(structure(Sym,Args),N,[putstr(N,Sym)|Pil],Pilr,Tin,Tout) :-
    tbldpars(Args,Pil,Pilr,Tin,Tout).
%   telling(X),tell(user),
%   write('putstr: '),write(N),
%   write('  Tin= '),write(Tin),write('  Tout= '),write(Tout),nl,
%   write('  Pil= '),write(Pil),nl,told,tell(X).

%:- import tell/1, telling/1, told/0, write/1, nl/0 from standard.

tbldpars([],Pil,Pil,T,T).
tbldpars([Subpar|Subpars],Pil,Pilr,T1,T2) :-
    tbldpar(Subpar,Pil,Pil1,T1,T3),
    tbldpars(Subpars,Pil1,Pilr,T3,T2).

tbldpar(varocc(Vid,Prag),Pil,Pilr,Tin,Tout) :-
    alloc_reg(Prag,Tin,Tmid),
    bldvarinst(Prag,Pil,Pilr,Tmid),
    ( vprag_context(Prag, i), Tout=Tmid			/* in inline */
    ; vprag_type(Prag,t), vprag_loc(Prag,Loc),
	release_if_done(Vid,Loc,Prag,[],Tmid,Tout)
    ; Tmid = Tout
    ).
tbldpar(integer(V),[bldnumcon(V)|Pilr],Pilr,T,T).
tbldpar(real(V),[bldfloat(V)|Pilr],Pilr,T,T).
tbldpar(constant(Sym),[bldcon(Sym)|Pilr],Pilr,T,T).
tbldpar(string(Sym),[bldstring(Sym)|Pilr],Pilr,T,T).

tgetpar(integer(V),N,[getnumcon(N,V)|Pil],Pil,Tin,Tin).
tgetpar(real(V),N,[getfloat(N,V)|Pil],Pil,Tin,Tin).
tgetpar(constant(Sym),N,[getcon(N,Sym)|Pil],Pil,Tin,Tin).
tgetpar(string(Sym),N,[getstring(N,Sym)|Pil],Pil,Tin,Tin).
tgetpar(varocc(_Vid,Prag),N,Pil0,Pil,Tin,Tout) :-
    ( vprag_type(Prag, vh), Pil0=Pil, Tin=Tout
    ;   ( vprag_type(Prag, t), alloc_reg1(Prag,N,Tin,Tout)
	; Tin = Tout
	),
	getvarinst(Prag,N,Pil0,Pil,Tout)
    ).
tgetpar(structure(Sym,Args),N,[getstr(N,Sym)|Pil],Pilr,Tin,Tout) :-
    tunipars(Args,Pil,Pilr,Tin,Tout).

/* loops through sub fields of a par */
tunipars([],Pil,Pil,T,T).
tunipars([Subpar|Subpars],Pil,Pilr,T1,T2) :-
    tunipar(Subpar,Pil,Pil1,T1,T3),
    tunipars(Subpars,Pil1,Pilr,T3,T2).

/* generates code for a subfield of par */

tunipar(varocc(Vid,Prag),Pil,Pilr,Tin,Tout) :-
    alloc_reg(Prag,Tin,Tmid),
    univarinst(Prag,Pil,Pilr,Tmid), 
    ( vprag_context(Prag, i), Tout=Tmid			/* in inline */
    ; vprag_type(Prag,t), vprag_loc(Prag,Loc),
	release_if_done(Vid,Loc,Prag,[],Tmid,Tout)
    ; Tmid = Tout
    ).
tunipar(integer(V),[uninumcon(V)|Pilr],Pilr,T,T).
tunipar(real(V),[unifloat(V)|Pilr],Pilr,T,T).
tunipar(constant(Sym),[unicon(Sym)|Pilr],Pilr,T,T).
tunipar(string(Sym),[unistring(Sym)|Pilr],Pilr,T,T).


force_tp_goal_loaded.
