/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tp_index.P
  Author(s)		:  Saumya Debray, Terrance Swift
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  June 11, 1992
========================================================================*/


/*======================================================================*/
/*  makeindex(+Sym, +Clauses, -Pil, -PilR, +PredPrag)			*/
/* 	Generates tryp/retry/trust and index instructions for clauses	*/
/*	of the predicate/function.					*/
/*======================================================================*/

/* single clause, no try/index instruction is needed */
makeindex(_Sym, [_SingleClause], Pil, Pil, _Prag,0,_Tabind) :-!.

makeindex(Sym, [clause(_,_,crec(_,_,Label,_))],Pil,Pilr,_Prag,1,
					tabled(Ind,Chs,Rhs,_)) :-
	!,
	sym_name(Sym,_Pred,Arity),
	Pil = [tabletryfail(Arity,Label,Ind),
%	       comptabletryfail(Arity,Label,Ind),
	       tabconfig(Chs,Rhs)|Pilr].
	
/* 0-ary predicate, generate try's only */
makeindex(Sym, Clauses, Pil,Pilr,_Prag,Tabled,Tabind) :-
	sym_name(Sym,_Pred,Arity),
	Arity =:= 0,
	!,
	gennum(_N),
	index_label_list(Clauses,LabelList),
	index_gentrys(Arity,LabelList,Pil,Pilr,Tabled,Tabind).

/* not needed right now -- tls 
 parallel predicates, generate try's only 
makeindex(Sym, Clauses, Pil,Pilr,Prag,Tabled,Tabind) :-
	sym_name(Sym,Pred,Arity),
	member((parallel), Prag), 
	!,
	gennum(N),
	index_label_list(Clauses,LabelList),
	length(LabelList, Size),
	Pil = [partry(Arity, Size)|Pil1],
	index_gentrys(Arity,LabelList,Pil1,Pilr,Tabled,Tabind).
*/
/* all other cases */
makeindex(Sym,Clauses, Pil, Pilr,Prag,Tabled,Tabind) :-
	sym_name(Sym, _Pred, Arity),
	( member(index(Narg), Prag), !
	; Narg = 1				% Default index on arg 1
	),
	index_argpats(Clauses,Narg,ArgList),
	index_split(ArgList, ArgListGroup),
	Parallel = 0,
	( ArgListGroup = [_One] -> NeedLabel = 0, Pil2 = Pil
	; index_extract_label_list(ArgListGroup, LabelList),
	   index_gentrys(Arity, LabelList, Pil, Pil2,Tabled,Tabind),
	   NeedLabel = 1
	),
	index_groups(ArgListGroup, Narg, Pil2, Pilr,NeedLabel,Parallel,Arity,
							Tabled,Tabind).

/*======================================================================*/
/* index_groups(+ArgListGroup, +ArgNum, -Pil, -PilRest, +NeedLabel,	*/
/*		+Parallel, +Arity)					*/
/*======================================================================*/
index_groups([], _Narg, Pil, Pil, _NeedLabel, _Parallel, _Arity,
							_Tabled,_Tabind).
index_groups([arglabel(T,Group,L)|Rest], Narg,Pil,Pilr,
		NeedLabel,Parallel,Arity,Tabled,Tabind) :-
	( T = g ->
	   ( NeedLabel =:= 1 ->
		gennum(N), L = (inde,N), Pil = [label(inde,L)|Pil1]
	   ; Pil = Pil1
           ),
	   index_group(Group, Narg, Pil1, Pil2, Parallel,Arity,Tabled,Tabind)
	; Pil = Pil2
	),
	index_groups(Rest, Narg, Pil2, Pilr, NeedLabel, Parallel, Arity,
							Tabled,Tabind).

/*======================================================================*/
/* index_group(+Group, +Narg, -Pil, -PilRest, +Parallel, +Arity)	*/
/*	Generate index and try/retry/trust instructions for a group.	*/
/*======================================================================*/

index_group([arglabel(T1,_,L1),arglabel(T2,_,L2)],
		Narg,Pil,Pilr,Parallel,Arity,Tabled,Tabind) :-
        ( index_sot_inst(T1,T2,Narg,L1,L2,Inst) -> Pil = [Inst|Pil1]
	; T1 = l, T2 = l -> Pil = Pil1
	),
	( Parallel =:= 1 -> Pil1 = [partry(Arity, 2)|Pil2]
	; Pil1 = Pil2
	),
	index_gentrys(Arity, [L1,L2], Pil2, Pilr,Tabled,Tabind).
index_group(Group, Narg, Pil, Pilr, Parallel, Arity,Tabled,Tabind) :-
	length(Group, Size),
	gennum(N), L = (inde,N),
	Pil = [cases(L, Arity, Size, _)|Pil1],
	index_genarglabel(Group, Pil1, Pil2),
	Pil2 = [switchonbound(Narg, L, Size)|Pil3],
	( Parallel =:= 1 -> Pil3 = [partry(Arity, Size) | Pil4]
	; Pil3 = Pil4
	),
	index_extract_label_list(Group, LabelList),
	index_gentrys(Arity, LabelList, Pil4, Pilr,Tabled,Tabind).

index_genarglabel([], Pil, Pil).
index_genarglabel([ArgLabel|Rest], [ArgLabel|Pil], Pilr) :-
	index_genarglabel(Rest, Pil, Pilr).

/*======================================================================*/
/* index_gentrys(+Arity, +LabelList, -Pil, -PilR)			*/
/*======================================================================*/

index_gentrys(Arity, [Label|Rest], Pil, Pilr,Tabled,tabled(Ind,Chs,Rhs,_)) :-
	( option(shallow) ->  Pil = [trys(Arity,Label)|Pilm]
	;	(Tabled =:= 1 -> Pil = [tabletry(Arity,Label,Ind),
%	                                comptabletry(Arity,Label,Ind),
					tabconfig(Chs,Rhs)|Pilm]
		 ; 		 Pil = [try(Arity,Label)|Pilm])
	),
	index_genretrys(Arity,Rest,Pilm,Pilr,Tabled).

index_genretrys(Arity,[L1],Pil,Pilr,Tabled) :-
	( option(shallow) ->    Pil = [trusts(Arity,L1)|Pilr]
	;	(Tabled =:= 1 -> Pil = [tabletrust(Arity,L1)|Pilr]
		 ; 		 Pil = [trust(Arity,L1)|Pilr])
	).


index_genretrys(Arity,[L1|Rest],Pil,Pilr,Tabled) :-
	Rest = [_|_],
	( option(shallow) ->    Pil = [retrys(Arity,L1)|Pilm]
	; 	(Tabled =:= 1 ->  Pil = [tableretry(Arity,L1)|Pilm]
	;			Pil = [retry(Arity,L1)|Pilm]
	)),
	index_genretrys(Arity,Rest,Pilm,Pilr,Tabled).

/*======================================================================*/
/* special switchonbound cases: two clauses, one is atomic type and the	*/
/* other is compound type.						*/
/*======================================================================*/
index_sot_inst(i,l,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(i,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(n,l,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(n,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(c,l,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(c,s,N,L1,L2,switchonterm(N,L1,L2)).
index_sot_inst(l,i,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(l,n,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(l,c,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(s,n,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(s,i,N,L1,L2,switchonterm(N,L2,L1)).
index_sot_inst(s,c,N,L1,L2,switchonterm(N,L2,L1)).

/*======================================================================*/
/* index_label_list(+Clauses, -LabelList)				*/
/*	Obtain the list of labels of all clauses. The labels are	*/
/*	generated (later) by compile_clause/5.				*/
/*======================================================================*/
index_label_list([],[]).
index_label_list([clause(_,_,crec(_,_,Label,_))|ClRest],[Label|LRest]) :-
	index_label_list(ClRest,LRest).


/*======================================================================*/
/* index_extract_label_list(+ArgPatterns, -LabelList)			*/
/*	Same as index_label_list/2 but the input is a list of arg	*/
/*	patterns.
/*======================================================================*/
index_extract_label_list([],[]).
index_extract_label_list([arglabel(_,_,Label)|Rest],[Label|LRest]) :-
	index_extract_label_list(Rest,LRest).

/*======================================================================*/
/* index_split(+ArgPatterns, +ArgGroups)				*/
/*	Split the argpatterns into a list of arg patterns (ArgGroup)	*/
/*	so that each group contains either a "v" arg or a list of	*/
/*	non-"v" (bound) args.						*/
/*	Precisely, ArgGroups is a list of arglabel(T,Value,Label) 	*/
/*	and when T = g, Value is again a list of arg patterns.		*/
/*======================================================================*/
/* index_split/2 is also used in cp_opt.P.  				*/
/*======================================================================*/

index_split(ArgPatterns, ArgGroups) :-
	index_split0(ArgPatterns, ArgGroups0),
	index_simplify_singleton(ArgGroups0, ArgGroups).

index_split0([], []).
index_split0([One|Rest], [arglabel(g,Group,_L)|GRest]) :-
	( One = arglabel(v, _, _) -> Group = [One], NRest = Rest
	; index_split1([One|Rest], NRest, Group)
	),
	index_split0(NRest, GRest).

index_split1([], [], []).
index_split1([One|Rest], NRest, NA) :-
	( One = arglabel(v, _, _) -> NRest = [One|Rest], NA = []
	; NA = [One|NNA], index_split1(Rest, NRest, NNA)
	).

index_simplify_singleton([], []).
index_simplify_singleton([Group|Rest], [NGroup|NRest]) :-
	Group = arglabel(g, Patterns, _L),
	( Patterns = [One] -> NGroup = One
	; NGroup = Group
	),
	index_simplify_singleton(Rest, NRest).

/*======================================================================*/
/* index_argpats(+Clauses, +IndexArg, -ArgPatList)			*/
/*	IndexArg: the argument number to be indexed			*/
/*	ArgPatList: a list of arglabel(Pattern, HashVal, Label)		*/
/*	  where Label is the label for the clause			*/
/*		Pattern 	HashVal					*/
/*		n (nil symbole)	[]					*/
/*		i (integer)	an integer				*/
/*		c (constant)	a Symbol entry				*/
/*		v (variable)	v					*/
/*		l (".")		l					*/
/*		s (structure)	a Symbol entry				*/
/*======================================================================*/
/* code like index_argpats/3 is used in cp_opt/2.  Please make sure that*/
/* any major indexing changes which aggect index_argpats/3 does not also*/
/* affect cpo_index_argpats/3						*/
/*======================================================================*/
index_argpats([],_,[]).
index_argpats([clause(Arglist,_,crec(_,_,L,_))|CRest], N, 
			[arglabel(Pattern,NewArg,L)|APatternRest]) :-
        nthmember(Arg, Arglist, N),
	index_argpat(Arg,Pattern, NewArg),
	index_argpats(CRest,N,APatternRest).

index_argpat(constant(Sym), n, []) :- sym_name(Sym, [], 0).
index_argpat(integer(X), i, X).
index_argpat(real(X), i, X).
index_argpat(constant(C), c, C).
index_argpat(varocc(_), v, v).
index_argpat(structure(Sym,_), l,l) :- sym_name(Sym, '.', 2).
index_argpat(structure(Sym,_), s,Sym).


force_tp_index_loaded.
