/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  tprog.P
  Author(s)		:  Saumya Debray, Kostis Sagonas, Terrance Swift
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  May 24, 1993
========================================================================*/

/*======================================================================*/
/* compile_clause(+Sym, +Clause, -PilCode, -PilCodeR, +Neck)		*/
/*	This procedure takes a clause in internal source form		*/
/*	clause(_,_,_), translates it into the intermediate form		*/
/*	clause(_,_,_) (done by the predicate inprog/2), and then	*/
/*	produces the list of PIL instructions for that clause.		*/
/*	- PilCode-PilCodeR is the difference list of the resulting	*/
/*			   translation of the clause.			*/
/*	+ Neck = 1: generate "neck", and do not reuse registers in the	*/
/*		    head (shallow backtracking)				*/
/*		 0: do not generate "neck" (no shallow backtracking)	*/
/* 	Refer to the file Format for the format of InClause.		*/
/*======================================================================*/

compile_clause(Sym, Clause, PilCode, PilCodeR, Neck) :-
	inprog(Sym, Clause, InClause),
	tclause(Sym, InClause, PilCode, PilCodeR, Neck).

tclause(Sym, clause(Args,Body,crec(_,ARSize,Label,_)),
		[label(clau,Label)|Pil0], Pilr, Neck) :- 
	( sym_type(Sym,tabled(_,_,_,Tabar)) -> 
			 Tabled = 1, ARSize1 is ARSize + 1
	; Tabar = Tabar, Tabled = 0, ARSize1 = ARSize
	),
	gennum(N), Label = (Sym, N),
	( Tabled =:= 1 -> Pil0 = [allocate,getpbreg(ARSize)|Pil1]
	; ( ARSize1 > 0 -> Pil0 = [allocate|Pil1] ; Pil1 = Pil0 )
	),
	sym_name(Sym,_,Arity),
	reserve(Arity, [], Tin),
	( Neck=:=0, theadpars(1,Args,Pil1,Pil2,Tin,TRegs1), Pil2 = Pil3
	; Neck=\=0, theadpars_hold(1,Args,Pil1,Pil2,Tin,TRegs1),
		    Pil2 = [neck(Arity)|Pil3]
	),
%telling(X),tell(user),write(Args),nl,
%write('PilHead = '),write(Pil1),nl,told,tell(X),
	tbody(Body,ARSize1,Pil3,Pilr,TRegs1,_,lastgoal,no,[],Tabled,Tabar).

/*======================================================================*/
/* theadpars(+Argno, +Args, -Pil-Pilr, +TRin, -TRout)			*/
/*	Generates code for the head. It generates code for one argument	*/
/*	first, and then recursively calls itself to generate code for	*/
/*	the rest of the arguments.					*/
/*		TRin:  list of registers in use at entry.		*/
/*		TRout: list of registers in use at exit.		*/
/*									*/
/* theadpars_hold(+Argno, +Args, -Pil-Pilr, +TRin, -TRout)		*/
/*	It is the same as theadpars/6 except it does not release the	*/
/*	registers used by the parameters. Used for shallow backtracking.*/
/*======================================================================*/

:- index theadpars/6-2.

theadpars(_, [], Pil, Pil, T, T).
theadpars(N, [Par|Rest], Pil, Pilr, TRin, TRout) :-
	release(N, TRin, TRm),
	tgetpar(Par, N, Pil, Pil1, TRm, TRmid),
	N1 is N+1,
	theadpars(N1, Rest, Pil1, Pilr, TRmid, TRout).

:- index theadpars_hold/6-2.

theadpars_hold(_, [], Pil, Pil, T, T).
theadpars_hold(N, [Par|Rest], Pil, Pilr, TRin, TRout) :-
	tgetpar(Par, N, Pil, Pil1, TRin, TRmid),
	N1 is N+1,
	theadpars_hold(N1, Rest, Pil1, Pilr, TRmid, TRout).

/*======================================================================*/
/* ----------- EFFICIENT CODE FOR DISJUNCTIONS/CONDITIONALS ----------- */
/* 
   This is an algorithm to generate efficient code for disjunctions
   and conditionals nested arbitrarily deep.  The emphasis here is to
   avoid chains of branches when different execution paths come together.
   However, instead of tedious scanning of assembly code to detect this,
   we try to avoid it altogether by passing labels around.

   The idea is the following: execution branches need come together only
   if the goals are of the form (c1 OR c2) AND c3.  In this case, when we
   see the AND, we generate a label, which is the label of the place where
   the execution paths should meet.  This is then passed into the routines
   that process the disjunction, as a parameter "meet(Label)".

   The decision on when to actually emit the "meet" label is decided by
   passing around a parameter, Depth.  This can take a value of 0 or 1.
   A depth value of 0 indicates an "outer disjunction", i.e. a goal of
   the form ( (c1 OR c2) AND c3 ).  A depth value of 1 indicates an "inner
   disjunction", e.g. the inner OR in the case ((c1 OR (c2 OR c3)) AND c4).
   This information is used to determine when to generate the label
   corresponding to the "meet" label: this is generated if and only if
   (i) a meet exists, i.e. is nonvariable, and (ii) the depth is 0.  If
   these conditions are met, the meet label is generated and the depth set
   to 1 so that duplicate labels are not produced.

   Things are complicated by the fact that we generally only look at the
   outermost connective (it is expensive to search the tree all the time).
   Thus, it is possible to have a goal of the form
   		( (c1 AND (c2 OR c3) ) AND c4 )
   Here, the paths should come together before c4.  This can be handled as
   before, as it turns out: if the goal is of the form ((c1 OR c2) AND c3),
   then a new meet is generated and passed into the first conjunct (i.e.
   (c1 OR c2), together with a new depth of 0; the meet and depth values
   passed down into the second conjunct c2 are what was passed in from
   above, since this is where the execution paths should subsequently come
   together, if necessary.

 */

/*======================================================================*/
/* tbody(+Goal, +ARSize, Pil-Pilr, +Tin, -Tout, +Meet,+Nested,HoldRegs)	*/
/*	Tin, Tout: the registers in use at the entry and at the exit.	*/
/*	Meet: When processing the nested or-branches, this is the label	*/
/*		to meet. It can be also "lastgoal", in which case all	*/
/*		or-branch (top level or nested) do not generate labels	*/
/*		and jumps. "midgoal" for insignificant cases.		*/
/*	Nested: if "no", this is the top level or-branch (if		*/
/*		we are processing an or-branch);			*/
/*		if "yes", this would be a nested or-branch.		*/
/*======================================================================*/

tbody(and(Goal,Goals,_),A,Pil,Pilr,Tin,Tout,Meet,_,HoldRegs,Tabled,Tabar) :-
    tbody(Goal,A,Pil,Pil1,Tin,Tmid,midgoal,no,HoldRegs,Tabled,Tabar),
    tbody(Goals,A,Pil1,Pilr,Tmid,Tout,Meet,no,HoldRegs,Tabled,Tabar),
    !.
tbody(or(Goal,Goals,_),A,Pil,Pilr,Tin,[],Meet,Nested,HoldRegs,Tabled,Tabar) :-
    ( Nested = no -> 
	gennum(LabId1), DSym = disj(LabId1),
	Pil = [call(DSym,A), label(disj,(DSym,LabId1)),
			trymeelse(0,Label2)|Pil1],
	( Meet = lastgoal -> ThisMeet = Meet, Pil5 = Pilr
	; gennum(LabId4), ThisMeet = (disj,LabId4),
	  Pil5 = [label(meet, ThisMeet)|Pilr]
	)
    ; 	Pil = [retrymeelse(0,Label2)|Pil1],
	Pil5 = Pilr, ThisMeet = Meet
    ),
    tbody(Goal,A, Pil1,Pil2,Tin,_,Meet,no,HoldRegs,Tabled,Tabar),
    gennum(LabId2), Label2 = (disj,LabId2),
    ( Meet = lastgoal ->
	Pil2 = [label(disj,Label2)|Pil3]
    ;	Pil2 = [jump(ThisMeet),label(disj,Label2)|Pil3]
    ),
    ( more_branches(Goals) -> Pil3=Pil4 ; Pil3 = [trustmeelsefail(0)|Pil4] ),
    tbody(Goals,A,Pil4,Pil5,Tin,_,ThisMeet,yes,HoldRegs,Tabled,Tabar).
tbody(if_then_else(Test,TGoal,FGoal,_),A,Pil,Pilr,Tin,Tout,M,_D,HoldRegs0,
								Tabled,Tabar) :-
    gennum(TLabId), TLabel = label(then,(disj,TLabId)),
    gennum(FLabId), FLabel = label(else,(disj,FLabId)),
    tvar_hold(Test, HoldRegs0, HoldRegs1),
%telling(X),tell(user), write('  Tin= '),write(Tin),nl,told,tell(X),
    tcond(Test,TLabel,FLabel,Pil,[TLabel|Pil1],Tin,Tmid,HoldRegs1),
    tbody(TGoal,A,Pil1,Pil2,Tmid,Tout0,M,no,HoldRegs1,Tabled,Tabar),
    ( M = lastgoal ->
	Pil2 = [FLabel|Pil3], Pil4 = Pilr
    ;	Pil2 = [jump(ThisMeet),FLabel|Pil3],
	gennum(MLabId), ThisMeet = (meet,MLabId),
	Pil4 = [label(meet, ThisMeet)|Pilr]
    ),
%telling(X),tell(user),
%write('  Tin= '),write(Tin),write('  Tmid= '),write(Tmid),
%write('  Tout0= '),write(Tout0),nl, told,tell(X),
    merge(Tmid,Tout0,Tout1),
    tbody(FGoal,A,Pil3,Pil4,Tout1,Tout2,M,no,HoldRegs0,Tabled,Tabar), 
    /* tvar may be in */
    merge(Tout1,Tout2,Tout), !.		/* branches of an i-t-e */
tbody(nil,_,Pil,Pilr,T,T,_,_,_,Tabled,Tabar) :-
    ( Tabled =:= 0 -> Pil = [proceed|Pilr]
    ; % handle tabled preds wo body.
      Pil = [addtotable(Tabar,2),deallocate,proceed|Pilr] 
    ).

%:- import tell/1, telling/1, told/0, write/1, nl/0 from standard.

tbody(goal(Sym,Args,_CPrag),A,Pil,Pilr,Tin,Tout,Meet,_,_HoldRegs,
							Tabled,Tabar) :-
    ( Meet = lastgoal -> LastLit = 1 ; LastLit = 0 ),
    sym_name(Sym, Pred, Arity),
    reserve(Arity,Tin,T1), Tout = [],
    tgoalargs(Args,1,Pil,Pil1,T1,_),
%telling(X),tell(user),
%write('tgoalargs called: '),write(Args),
%write('  Tin= '),write(Tin),write('  T1= '),write(T1),
%write('  Tout= '),write(Tout),nl,
%write('  Pil= '),write(Pil),nl,told,tell(X),
    ( option(sysmod), inline_builtin(Pred, Arity, BltNum), !,
	( LastLit =:= 1, 
	  ( A > 0,  Pil1 = [deallocate,builtin(BltNum),proceed|Pilr]
	  ; A =< 0, Pil1 = [builtin(BltNum),proceed | Pilr]
	  )
        ; LastLit =\= 1, Pil1 = [builtin(BltNum)|Pilr]
        )
    ; ( LastLit =:= 1 ->
	( Tabled =:= 1 ->      /* if tabled = 1, A>0 */
		A1 is A-1, Pil1 = [call(Sym,A),addtotable(Tabar,A1),
				   deallocate,proceed|Pilr]
	 ; ( A > 0 -> Pil1 = [deallocate,execute(Sym)|Pilr]
	   ;          Pil1 = [execute(Sym)|Pilr] )
	 )
      ; Pil1 = [call(Sym,A)|Pilr] )   /* LastLit =\= 1 --  A: ARSize */
    ),
    !.
tbody('_$complete'(goal(Sym,Args,_CPrag)),A,Pil,Pilr,Tin,Tout,Meet,_,_HoldRegs,
							Tabled,Tabar) :-
    ( Meet = lastgoal -> LastLit = 1 ; LastLit = 0 ),
    sym_name(Sym, _Pred, Arity),
    reserve(Arity,Tin,T1), Tout = [],
    tgoalargs(Args,1,Pil,Pil1,T1,_),
    ( LastLit =:= 1 ->
	( Tabled =:= 1 ->      /* if tabled = 1, A>0 */
		A1 is A-1, Pil1 = [complete_call(Sym,A),addtotable(Tabar,A1),
				   deallocate,proceed|Pilr]
	 ; ( A > 0 -> Pil1 = [complete_call(Sym,A),deallocate,proceed|Pilr]
	   ;          Pil1 = [complete_execute(Sym)|Pilr] )
	 )
    ; Pil1 = [complete_call(Sym,A)|Pilr]	/* LastLit =\= 1 -- A: ARSize */
    ),
    !.
tbody(inlinegoal(Pred,_Ar,Args,_CPrag),A,Pil,Pilr,Tin,Tout,Meet,_,HoldRegs,
							Tabled,Tabar) :-
    ( Meet = lastgoal -> LastLit = 1 ; LastLit = 0 ),
    geninline(Pred,Args,HoldRegs,Pil,Pil1,Tin,Tout),
    ( LastLit =:= 1 ->
	( Tabled =:= 1 ->      /* if tabled = 1, A>0 */
		A1 is A-1, Pil1 = [addtotable(Tabar,A1),deallocate,proceed|Pilr]
	 ; ( A > 0 -> Pil1 = [deallocate,proceed|Pilr]
	   ;          Pil1 = [proceed|Pilr] )
	 )
    ; Pil1 = Pilr /* LastLit =\= 1 */
    ).
tbody(fungoal(Sym,Res,Args,_CPrag),A,Pil,Pilr,Tin,Tout,Meet,_,_HoldRegs,
							      _Tabled,_Tabar) :-
    ( Meet = lastgoal -> LastLit = 1 ; LastLit = 0 ),
    sym_name(Sym, _Pred, Arity),
    reserve(Arity,Tin,T1), Tout = [],
    tgoalargs(Args,1,Pil,[userfunc(A,Sym)|Pil3],T1,T2),
    tgetpar(Res, 0, Pil3, Pil1, T2, _),
    ( LastLit =:= 1, 
	( A > 0,  Pil1 = [deallocate,proceed|Pilr]
	; A =< 0, Pil1 = [proceed | Pilr]
	)
    ; LastLit =\= 1, Pil1 = Pilr		% A: ARSize
    ).
tbody(funbody(_Res, Body),A,Pil,Pilr,Tin,_Tout,_Meet,_Nested,HoldRegs,
								Tabled,Tabar) :-
    tbody(Body,A,Pil,Pil1,Tin,_Tmid,midgoal,no,HoldRegs,Tabled,Tabar),
    ( A > 0,  Pil1 = [deallocate,proceed|Pilr]
    ; A =< 0, Pil1 = [proceed | Pilr]
    ).

more_branches(or(_,_,_)).

force_tprog_loaded.
