/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  writeasm.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu
  Modified by		:  Jiyang Xu
  Last modification	:  June 14, 1993
========================================================================*/


write_asmheader(AFile, ModName, SymTab, IsModule) :-
	tell(AFile),
	( IsModule =:= 1 -> write(modname(ModName)) ; write(globalmod) ), nl,
	write_symbols(SymTab).

/* I made the following "SymTab" structure indepedent (Kostis 7/16/92) */

write_symbols(SymTab) :-
	sort_by_index(SymTab, Sorted_SymList),
	!,
	( rest_write_symbols(Sorted_SymList),
	  fail	% undo the binding of the tail of the hash table part of
		% the symbol table (caused by the sorting of the symbols).
	; true
	).

rest_write_symbols(SymTab) :-
	member(Sym, SymTab),
	write_symbol(Sym),
	fail.	% get next symbol from SymTab and reclaim the space used.
rest_write_symbols(_).

write_symbol(Sym) :-
	sym_name(Sym, Name, Arity),
	sym_offset(Sym, Offset),
	sym_env(Sym, Env),
	sym_category(Sym, Cat),
	write('#'), write(Offset), write(': '), write(Name/Arity),
	tab(8), write(Cat), tab(1), write(Env), tab(4),
	( sym_type(Sym, Type) ->  write_type(Type) ; true ), tab(8),
	( sym_type(Sym, tabled(_,_,_,_)) -> write('TABLED'), nl
	                           ;  write('UNTABLED'), nl ),
	!.

write_type(X) :- var(X), !, write(X).
write_type([T1]) :- !, write_type(T1).
write_type([T1|T2]) :- write_type(T1), write('; '), nl, tab(16), write_type(T2).
write_type(arrow(T1, T2)) :- write_type(T1), write(' -> '), write_type(T2).
write_type(X) :- write(X).

write_asm_instructions([]).
write_asm_instructions([One|Rest]) :-
	write_asm_inst(One),
	write_asm_instructions(Rest).

write_asm_inst(label(T,L)) :- 
	!, 
	simplify_label(L, Name-Id),
	write(Id), write(':  label '), 
	( T = pred -> write(Name) ; write(T) ), nl.
write_asm_inst(call(Sym,A)) :- 
	!, 
	( sym_name(Sym, Name, Arity), !, N = Name/Arity ; N = Sym ),
	write('    call '), write(A), put(0' ), write(N), nl.
write_asm_inst(complete_call(Sym,A)) :- 
	!, 
	( sym_name(Sym, Name, Arity), !, N = Name/Arity ; N = Sym ),
	write('    complete_call '), write(A), put(0' ), write(N), nl.
write_asm_inst(execute(Sym)) :- 
	!,
	( sym_name(Sym, Name, Arity), !, N = Name/Arity ; N = Sym ),
	write('    execute '), write(N), nl.
write_asm_inst(switchonbound(_Narg,Lab,_NCls)) :- 
	!,
	write('    switchonbound '), write(Lab), nl.
write_asm_inst(switchon3bound(_N1,_N2,_N3,Lab,_NCls)) :- 
	!,
	write('    switchon3bound '), write(Lab), nl.
write_asm_inst(cases(Lab,Arity,N1,_N2)) :- 
	!,
	write('    cases '), write(Lab), put(0' ),
	write(Arity), put(0' ), write(N1), nl.
write_asm_inst(arglabel(T,V,L)) :- 
	!,
	write('    arglabel '), write(T), put(0' ),
	( T = i -> write(V)
	; T = c -> sym_name(V, Name, Arity), write(Name/Arity)
	; T = s -> sym_name(V, Name, Arity), write(Name/Arity)
	; true
	),
	simplify_label(L, _N-Id), put(0' ), write(Id), nl.
write_asm_inst(Inst) :- 
	functor(Inst, Op, A),
	functor(Inst0, Op, A),
	asm_inst(Inst0, _, _, ArgPattens),
	write('    '), write(Op), write_inst_args(ArgPattens, Inst, 0), nl.

write_inst_args([], _, _). 
write_inst_args([One|Rest], Inst, I) :-
	( One=p, I1 is I
	; One=pp, I1 is I
	; One=ppp, I1 is I
	; I1 is I + 1, arg(I1, Inst, Arg), write_inst_arg(One, Arg)
	),
	write_inst_args(Rest, Inst, I1).

write_inst_arg(v(_V), Loc) :- write(' v'), write(Loc).
write_inst_arg(r(_R), Loc) :- write(' r'), write(Loc).
write_inst_arg(s(_I), Sym) :- sym_name(Sym, P, N), put(0' ), write(P/N).
write_inst_arg(c(_I), Sym) :- sym_name(Sym, P, _), put(0' ), write(P).
write_inst_arg(i(_N), Int) :- put(0' ), write(Int).
write_inst_arg(n(_N), Int) :- put(0' ), write(Int).
write_inst_arg(a(_A), Int) :- put(0' ), write(Int).
write_inst_arg(l(_L), Lab) :- simplify_label(Lab, _N-Id), put(0' ), write(Id).
write_inst_arg(t(_T), Arg) :- put(0' ), write(Arg).

simplify_label((Sym,LabId), Name/Arity-LabId) :- sym_name(Sym, Name, Arity).
simplify_label((Other,LabId), Other-LabId).
simplify_label(abs(-1), fail - (-1)).


