/************************************************************************/
/*                                                                      */
/* XSB System                                                           */
/* Copyright SUNY at Stony Brook, 1986, ECRC 1990                       */
/* Copyright (C) SUNY at Stony Brook, 1993                              */
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB, */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along   */
/* with XSB so you can know your rights and responsibilities.           */
/* It should be in a file named LICENSE.                                */
/* Among other things, this notice must be preserved on all copies.     */
/*                                                                      */
/************************************************************************/

/*======================================================================
  File			:  dynload.c
  Author(s)		:  Jiyang Xu, Kostis Sagonas, Harald Schroepfer
  Last modification	:  July 29, 1993
========================================================================*/

#include "config_make.h"	/* This should appear BEFORE the ifdef!	*/

#ifdef FOREIGN

#ifndef SOLARIS
#include <a.out.h>
#else
#include <dlfcn.h>
#endif

#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include "aux.h"
#include "inst.h"
#include "psc.h"

#define BUFFEXTRA 1024

shapri static char *tfile = "/tmp/xsb-dyn";	/* tmp file for "ld" */
shapri static char *executable = EXECUTABLE;	/* name of exec in emu */
shapri extern byte proceed_inst;

byte *load_obj_dyn(pofilename, cur_mod, ld_option)
char *pofilename, *ld_option;
Psc cur_mod;
{
#ifdef SOLARIS
	char	*name;
	Pair	search_ptr;
	char	sofilename[128];
	int 	strl = strlen(pofilename);
	void	*handle;
	void	*funcep;
	bool	dummy();
	
	/* (1) create filename.so */

	strcpy(sofilename, pofilename );
	sofilename[strl-1] = 's';
	sofilename[strl]   = 'o';
	sofilename[strl+1] = '\0';

	/* (2) open the needed object */

	if (( handle = dlopen(sofilename, RTLD_LAZY)) == 0 )
	{
		fprintf( stderr, "%s\n", dlerror());
		return 0;
	}

	/* (3) find address of function and data objects
	**
	** dyn_link_all(loc, cur_mod);
	*/

	search_ptr = (Pair)get_ep(cur_mod);

	while (search_ptr)
	{
		name = get_name(search_ptr->psc_ptr);

		if (get_type(search_ptr->psc_ptr) == T_FORN)
		{
			if ((funcep = (int (*)) dlsym(handle, name)) == NULL)
			{
				fprintf(stderr, "%s\n", dlerror());
				fprintf(stderr, "++Warning: cannot find foreign procedure %s\n", name);
				set_ep(search_ptr->psc_ptr, (byte *)(dummy));
			}
			else
			{ 
				set_ep(search_ptr->psc_ptr, (byte *)(funcep));
			}
				
		}
		search_ptr = search_ptr->next;
	}
	return &proceed_inst;
#else
	shapri extern char *install_dir;
	int i, buffsize, start, fd, error_occur = 0, loadsize, loc;
	struct exec header;
	char buff[256], subfile[128];
	struct stat statbuff;

      /* first step: get the header entries of the *.o file, in order	*/
      /* to obtain the size of the object code and then allocate space	*/
      /* for it.							*/
        if (strlen(pofilename) >= 127) return 0;
	strcpy(subfile, pofilename);
	subfile[strlen(subfile)-1] = 'o';	/* change *.O to *.o */
	fd = open(subfile, O_RDONLY, 0);
	if (fd < 0) {
	    fprintf(stderr, "Cannot find the C object file: %s\n", subfile);
	    return 0;
	}
	read(fd, &header, sizeof(struct exec));
	close(fd);

      /* second step: run incremental ld and generate a temporary 	*/
      /* object file (including orginal *.o and libraries) ready to be	*/
      /* read in.							*/
	buffsize = header.a_text + header.a_data + header.a_bss;
	start = mem_alloc(buffsize);
      /* The "-T hex" option of ld starts the text segment at location	*/
      /* hex. Specifying -T is the same as using the -Ttext option.	*/
	sprintf(buff, "/bin/ld -N -A %s/emu/%s -T %x -o %s %s %s -lc", 
		install_dir, executable, start, tfile, subfile, ld_option);
	if (strlen(buff) > 510) {
	    fprintf(stderr, "++Error: buffer for ld command is too small\n");
            exit (-1);
	}
	system(buff);

      /* third step: check if the size of the buffer just allocated is	*/
      /* big enough to load the object (when the object code uses other	*/
      /* libraries, the buffer may not be big enough). If this is the	*/
      /* case, redo the second step with a bigger buffer.		*/
	fd = open(tfile, O_RDONLY, 0);
	if (fd < 0) {
	    fprintf(stderr, "++Error: The file is not generated by the loader\n");
	    return 0;
	}
	read(fd, &header, sizeof(struct exec));
	loadsize = header.a_text + header.a_data + header.a_bss;
	if (loadsize > buffsize) {
	  close(fd);			/* need to reallocate buffer */
	  mem_dealloc(start, buffsize);
	  start = mem_alloc(loadsize+BUFFEXTRA);
	  sprintf(buff, "/bin/ld -N -A %s/emu/%s -T %x -o %s %s %s -lc",
		install_dir, executable, start, tfile, subfile, ld_option);
	  system(buff);
	  fd = open(tfile, O_RDONLY, 0);
	  read(fd, &header, sizeof(struct exec));
	}

      /* fourth step: read in the intermediate object files.		*/
      /* load text and data segment */
	loadsize = header.a_text + header.a_data;
	lseek(fd, N_TXTOFF(header), 0);
	read(fd, start, loadsize);
      /* load symbol table and string table */
	fstat(fd, &statbuff);
	loadsize = statbuff.st_size-N_SYMOFF(header);
	loc = mem_alloc(loadsize+8);
	*(int *)loc = loadsize+8;
	*(int *)(loc+4) = header.a_syms/sizeof(struct nlist);
	lseek(fd, N_SYMOFF(header), 0);
	read(fd, loc+8, loadsize);
	close(fd);

      /* fifth step: link C procedure names with Prolog names.		*/
	dyn_link_all(loc, cur_mod);
	mem_dealloc(loc, loadsize+8);
	return &proceed_inst;
#endif
}

bool dummy()
{
    fprintf(stderr, "++Error: trying to use an undefined foreign procedure\n");
    return FALSE;
}

#ifndef SOLARIS
dyn_link_all(symtab, cur_mod)
char *symtab;
Psc cur_mod;
{
  int count, i, value;
  char *ptr, *strtab;
  struct nlist *sym_entry;
  char *name;
  Pair search_ptr;

  count = *(int *)(symtab+4);
  symtab += 8;
  strtab = symtab + count*sizeof(struct nlist);
  search_ptr = (Pair)get_ep(cur_mod);
  while (search_ptr) {
    name = get_name(search_ptr->psc_ptr);
/* Jiyang changed it to the form ``module_pred'':
    sprintf(name, "%s_%s", get_name(cur_mod), get_name(search_ptr->psc_ptr));
 */
    if (get_type(search_ptr->psc_ptr) == T_FORN) {
      for (i=0; i<count; i++) {
	sym_entry = (struct nlist *)(symtab + i * sizeof(struct nlist));
	ptr = strtab + sym_entry->n_un.n_strx;
	if (*ptr++ == '_' && strcmp(name, ptr)==0) { 
	  set_ep(search_ptr->psc_ptr, (byte *)(sym_entry->n_value));
	  break;
	}
      }
      if (i>= count) {          /* does not find the name */
	  fprintf(stderr, "++Warning: cannot find foreign procedure %s\n", name);
	  set_ep(search_ptr->psc_ptr, (byte *)(dummy));
      }
    }
    search_ptr = search_ptr->next;
  }
  return 0;
}
#endif

byte *load_obj(pofilename, cur_mod, ld_option)
char *pofilename, *ld_option;
Psc cur_mod;
{
/*
    if (static_foreign_libraries(get_name(cur_mod)))
        return load_obj_sta(cur_mod);
    else
 */
        return load_obj_dyn(pofilename, cur_mod, ld_option);
}

#endif
