/************************************************************************/
/*									*/
/* PSB-Prolog System							*/
/* Copyright (C) SUNY at Stony Brook, 1986; ECRC, 1990			*/
/* Copyright (C) SUNY at Stony Brook, 1993       			*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute	*/
/* PSB-Prolog, but only under the conditions described in the		*/
/* PSB-Prolog License Agreement.  A copy of this license is supposed to	*/
/* have been given to you along with PSB-Prolog so you can know your	*/
/* rights and responsibilities.  It should be in a file named LICENSE. 	*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  emusubs.i
  Author(s)		:  Warren, Xu, Swift, Sagonas
  Last modification	:  July 1993
========================================================================*/

/* change tls - shallow backtracking taken out */

/*======================================================================*/

/* change tls case DVAR taken out 3/20/91 */

nunify_with_con: /* op1,  op2=(untagged)con */
    IFTHENELSE_DEREF(op1, nunify_lcon)
	/* op1 is FREE */
#ifdef CONASSTRING
	      bind_string((CPtr)(op1), (char *)op2);
    }
    else if (isstring(op1)) {
	      if (string_val(op1) == (char *)op2) goto contcase; else Fail1;
	 }
	 else Fail1;
#else
	      bind_cs((CPtr)(op1), (Pair)op2);
    }
    else if (isconstr(op1) /* || isdelay(op1) */ ) {
	      (Pair)(CPtr)op1 = cs_val(op1);
	      if ((Pair)(CPtr)op1 != (Pair)op2) {	/* a != b */
		  if (((Pair)op2)->psc_ptr != ((Pair)(CPtr)op1)->psc_ptr)
							/* 0(a) != 0(b) */
			{Fail1;}
	      } else goto contcase;
	 } 
	 else Fail1;
#endif
    goto contcase;	/* end case nunify_with_con */

/*======================================================================*/

/* change tls 3/20/91 case DVAR taken out */

nunify_with_num: /* op1 is general, op2 has number (untagged) */
    IFTHENELSE_DEREF(op1, nunify_lnum)
	/* op1 is FREE */
	    bind_int((CPtr)(op1), (int)op2);
    }
    else if (isinteger(op1) /* || isdelay(op1) */) {
	    if (int_val(op1) == (int)op2) goto contcase; else Fail1;
	 }
	 else Fail1;	/* op1 is STRING, FLOAT, CS, or LIST */
    goto contcase; /* end of nunify_with_num */

/*======================================================================*/

nunify_with_float:  /* op1 is general, op2 has float (untagged) */
    IFTHENELSE_DEREF(op1, nunify_lfloat)
        /* op1 is FREE */
            bind_float(vptr(op1), asfloat(op2));
    }
    else if (isfloat(op1) /* || isdelay(op1) */) {
	      /*if (float_val(op1) == asfloat(op2)) goto contcase; else Fail1;*/
	      float_unification_exception(lpcreg); Fail1;
	 }
#ifndef DOMAIN
	 else Fail1;	/* op1 is INT, STRING, CS, or LIST */
#else
         else if (isdomain(op1)) {
		   if (cell((CPtr)domain_val(op1))!=op1) {
			op1 = cell((CPtr)domain_val(op1));
			goto nunify_lfloat;
		   }
                   else Fail1;
	      }
	      else Fail1;	/* op1 is INT, STRING, CS, or LIST */
#endif
    goto contcase; /* end of nunify_with_float */

/*======================================================================*/

nunify_with_nil: /* op1, nil_sym(tagged) */
    IFTHENELSE_DEREF(op1, nunify_lnil)
	/* op1 is FREE */
#ifdef CONASSTRING
	    bind_string((CPtr)(op1), 0);
    }
    else if (isstring(op1)) {
	    if (isnil(op1)) goto contcase; /* a == [] */
	    else Fail1;
	 }
	 else Fail1;	/* op1 is LIST, INT, DVAR, FLOAT, or DELAY */
#else
	    bind_cs((CPtr)(op1), nil_sym);
    }
    else if (isstring(op1) || isconstr(op1) ||
					isdelay(op1)) {
	    if (isnil(op1)) goto contcase;	/* a == [] */
	    else Fail1;
	 }
	 else Fail1;	/* op1 is LIST, INT, DVAR, or FLOAT */
#endif
    goto contcase;	/* end case nunify_with_nil */

/*======================================================================*/

nunify_with_str: /* (op1, op2 as psc_ptr) */
    /* struct psc_rec *str_ptr; using op2 */
    IFTHENELSE_DEREF(op1, nunify_lstr)
	/* op1 is FREE */
	    bind_cs((CPtr)(op1), (Pair)hreg);
	    new_heap_functor(hreg, (Psc)op2);
	    flag = WRITE;
    }
    else if (isconstr(op1) /* || isdelay(op1) */ ) {	
	    op1 = (Cell)(cs_val(op1));
	    if (*((Psc *)op1) == (Psc)op2) {
		flag = READFLAG;
		sreg = (CPtr)op1 + 1;
	    }
	    else Fail1;
	 }
	 else Fail1;
    goto contcase;	/* end of case for nunify_with_str */

/*======================================================================*/
    
nunify_with_list_sym: /* (op1) */
    IFTHENELSE_DEREF(op1, nunify_llist)
	/* op1 is FREE */
	    bind_list((CPtr)(op1), hreg);
	    flag = WRITE;
    }
    else if (islist(op1)) {	
	    sreg = clref_val(op1);
	    flag = READFLAG;
	 }
	 else Fail1;
    goto contcase;	/* end nunify_with_list_sym */

/*======================================================================*/

nbldval:
    deref(op1);
    if (isnonvar(op1)) {new_heap_node(hreg, op1);}
    else {		/* Need to make sure no backward pointers */
			/* This can be optimized by some static analysis */
	bind_ref((CPtr)(op1), hreg);
	new_heap_free(hreg);
    }
    goto contcase;
    
/*======================================================================*/

call_sub: /* (psc)*/

  if (*asynint_ptr > 0) {                /* interrupt detected */
	if (*asynint_ptr >= DELAYINT_MARK) { /* DELAYINT_MARK or KEYINT_MARK */
	    if (*asynint_ptr & KEYINT_MARK)
	         synint_proc(psc, MYSIG_KEYB, lpcreg-8);
	    else synint_proc(psc, MYSIG_DELAY, lpcreg-8);
	    lpcreg=pcreg;
	} else lpcreg = (byte *)get_ep(psc);
	*asynint_ptr = 0;
  } else switch (get_type(psc)) {
	case T_PRED:
	    lpcreg = (pb)get_ep(psc);
	    break;
	case T_FUNC:
	    lpcreg = (pb)get_ep(psc);
	    break;
	case T_DYNA:
	    lpcreg = (pb)get_ep(psc);
	    break;
	case T_FORN:
#ifdef sequent
	    dyn_pred = (int *)get_ep(psc);
#else
	    dyn_pred = (PFI)get_ep(psc);
#endif
	    dyn_pred();
	    lpcreg = cpreg;		/* "proceed" */
	    break;
	case T_UDEF:
	    synint_proc(psc, MYSIG_UNDEF, lpcreg-8);
	    lpcreg=pcreg;
	    break;
	case T_UFUN:
	    synint_proc(psc, MYSIG_UNDEF, lpcreg-8);
	    lpcreg=pcreg;
	    break;
	default:
	    synint_proc(psc, MYSIG_UNDEF, lpcreg-8);
	    lpcreg=pcreg;
	    break;
  }
#if (defined(TRACING) || defined(DEBUG))
  if (call_intercept) {               /* for debugging or for statistics */
  	pcreg = lpcreg;
/*	ereg = le_reg;		 for debugging only */
	intercept(psc);
	lpcreg = pcreg;
  }
#endif

  goto contcase;

/*======================================================================*/

/*======================================================================*/

complete_call_sub: /* (psc)*/
/* interrupts may need some work here */
  if (*asynint_ptr > 0) {                /* interrupt detected */
	if (*asynint_ptr >= DELAYINT_MARK) { /* DELAYINT_MARK or KEYINT_MARK */
	    if (*asynint_ptr & KEYINT_MARK)
	         synint_proc(psc, MYSIG_KEYB, lpcreg-8);
	    else synint_proc(psc, MYSIG_DELAY, lpcreg-8);
	    lpcreg=pcreg;
	} else lpcreg = (byte *)get_ep(psc);
	*asynint_ptr = 0;
  } else switch (get_type(psc)) {
	case T_PRED:
	    lpcreg = (pb)get_ep(psc);
	    break;
/*
	case T_DYNA:
	    lpcreg = (pb)get_ep(psc);
	    break;
*/
	case T_UDEF:
	    synint_proc(psc, MYSIG_UNDEF, lpcreg-8);
	    lpcreg=pcreg;
	    break;
	default:
	    synint_proc(psc, MYSIG_UNDEF, lpcreg-8);
	    lpcreg=pcreg;
	    break;
  }
  comptableflag = 1;
#if (defined(TRACING) || defined(DEBUG))
  if (call_intercept) {              
  	pcreg = lpcreg;
	intercept(psc);
	lpcreg = pcreg;
  }
#endif
  goto contcase;

/*======================================================================*/

/*======================================================================*/

neck_sub:

  if (!dreg) {
	CPtr b;

	cp_cpreg(breg) = cpreg;
	cp_hreg(breg) = hbreg;
	cp_ebreg(breg) = ebreg;
	 /* save arguments in the registers onto stack */
	b = breg + sizeof(struct choice_point)/sizeof(CellStr);
	save_registers(b, (int)op1, i, rreg);
  }
  dreg = 1;
  goto contcase;

/*======================================================================*/

/*======================================================================*/

neckcut_sub:

    if (dreg>0) breg = (CPtr)int_val(op2);
    else {
	breg += sizeof(struct choice_point)/sizeof(CellStr) + (int)op1;
	dreg = 1;
    }
    hbreg = cp_hreg(breg);
    ebreg = cp_ebreg(breg);
    goto contcase;

