/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  ham.P
  Author(s)		:  Some ECRC guy
  Modified by		:  Kostis Sagonas
  Last modification	:  April 12, 1993
========================================================================*/

/************************************************************************/
% one solution to the queens problem 
% in order to restrict the search space, a list of possible values is passed
% through the generation procedure; if a row is selected from this list for
% a given column, then this row-value will be discarded from the list.
% thus, it is no more necessary to check the horizontal and vertical
% constraints.
/************************************************************************/

go :-	cputime(T1), ( go(8, _), fail; true ), 
	cputime(T2), T is T2 - T1, 
	write('Time used: '), write(T), write(' sec'), nl.

demo :- cputime(T1), demo1, cputime(T2), T is T2 - T1,
	write('Time used: '),  write(T), write(' sec'), nl.
demo1 :-
	go(8, Soln), 
	write(Soln), nl, fail.
demo1.

go( Size, Solution ):-
	poss_rows( Size, PossList ), 
	solve( PossList, 0, [], Solution ).

solve( [Poss1|RestPoss], Col, Current, Final ):-
	Col1 is Col+1, 
	delete( [Poss1|RestPoss], Row, NewPossRows ), 
	safe( Col1, Row, Current ), 
	solve( NewPossRows, Col1, [s( Col1, Row ) | Current], Final ).
solve( [], _, Final, Final ).

delete( [Y|L], X, [Y|M] ):-
	delete( L, X, M ).
delete( [X|L], X, L ).

threaten( X, Y, I, J ):-
	U is X+Y, 
	U is I+J.
threaten( X, Y, I, J ):-
	U is X-Y, 
	U is I-J.
		
safe( X, Y, [s( I, J )|Rest] ):-
	not( threaten( X, Y, I, J ) ), 
	safe( X, Y, Rest ).
safe( _, _, [] ).

poss_rows( N, R ) :- poss_rows( N, [], R ).

poss_rows( 0, L, L ).
poss_rows( N, K, L ) :- N > 0,
	N1 is N-1, 
	poss_rows( N1, [N|K], L ).
