/***********************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  array1.P
  Author(s)		:  Jiyang Xu, Kostis Sagonas
  Modified by		:  Jiyang Xu, Kostis Sagonas
  Last modification	:  April 10, 1993
========================================================================*/


/*======================================================================*/
/* A simpler array implementation: only one version is valid.		*/
/*    --  Older versions are replaced by an invalid flag.		*/
/*    --  backtrackable							*/
/*======================================================================*/

:- export array_new/2, array_elt/3, array_update/4.

:- import functor/3, arg/3 from standard.
:- import term_set_arg/4 from machine.

/* -- array : backtrackable --------------------------------------*/

array_new(A, Size) :- functor(A, array, Size).

array_elt(A, I, V) :- arg(I, A, V).

array_update(A, I, V, A) :- term_set_arg(A, I, V, 0).

/* implementation without value trail --- 

array_update(A, I, V, A) :-
	arg(I, A, VL),
	( var(VL), !, VL = V			% if var, just set it
	;
	   ( term_set_arg(A, I, V, 1)
	   ; term_set_arg(A, I, VL, 1), fail
	   )
	).
*/

/* TEST

import array_new/2 from array1.
import array_elt/3 from array1.
import array_update/4 from array1.
array_new(A, 4), array_update(A,1,1,B), array_update(B,2,2,C),
	( array_update(C,3,3,D), array_elt(D,3,E4)
	; array_update(C,3,6,E), array_elt(E,3,E5)
	; array_update(C,3,7,F), array_elt(F,3,E6)
	),
	array_elt(A,3,E1),array_elt(B,3,E2),array_elt(C,3,E3).

*/
