/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  basics.P
  Author(s)		:  David S. Warren, Kostis F. Sagonas
  Modified by		:  Kostis F. Sagonas
  Last modification	:  February 16, 1993
========================================================================*/


:- export ground/1, copy_term/2, 
	  append/3, member/2, memberchk/2, length/2, ith/3, reverse/2,
	  select/3, load_dyn/1.

:- import buff_alloc/3, buff_set_cell/4, buff_cell/3, buff_word/3,
          buff_set_var/4, buff_set_word/3, buff_dealloc/4,
	  term_psc/2, psc_arity/2, term_arg/3, term_type/2,
	  expand_file_name/2
     from machine.
:- import add_hilog_symbol/1 from hilog_symbol.
:- import seeing/1, see/1, seen/0, call/1, functor/3,
	  expand_term/2, '=='/2, read/1, op/3, (table)/3
     from standard.
:- import predicate_property/2 from current_symbols.
:- import assert/1, abolish/1 from assert.
:- import warning/1 from aux.
:- import (import)/1 from interpreter.
:- import default_call_size/1, default_ret_size/1 from config.

% ground/1 checks for ground term
ground(T) :- nonvar(T), term_type(T,Ty),
	(Ty =:= 3
	 ->	T = [A1|A2],
	 	ground(A1), ground(A2)
	 ; Ty =:= 1
	 ->	term_psc(T,PSC),
	 	psc_arity(PSC,N),
		ground(T,N)
	 ;	true
	).
ground(T,N) :-
	N =:= 0
	 ->	true
	 ;	term_arg(T,N,A),
	 	ground(A),
		N1 is N-1,
		ground(T,N1).


% good ole member
member(X,[X|_]).
member(X,[_|L]) :- member(X,L).


memberchk(X,[X|_]) :- !.
memberchk(X,[_|L]) :- memberchk(X,L).


% append! ta-da!
append([],L,L).
append([X|L1],L2,[X|L3]) :- append(L1,L2,L3).


% A not so naive reverse
reverse(L, R) :- reverse_acc(L, [], R).

reverse_acc([], Acc, Acc).
reverse_acc([Head|Tail], Acc, Reversed) :-
	reverse_acc(Tail, [Head|Acc], Reversed).


% Some Prologs like to call this delete!
select(Element, [Element|Rest_Elements], Rest_Elements).
select(Element, [Element1|Rest_L1], [Element1|Rest_L2]) :-
	select(Element, Rest_L1, Rest_L2).


% length/2 that works both ways
length(L,N) :-
	var(N)
	 ->	length1(L,N)
	 ;	length2(L,N).

length1([], 0).
length1([_|R], N) :- length1(R, N1), N is N1 + 1.

length2(L,N) :-
	N =< 0
	 ->	L=[]
	 ;	N1 is N-1, L=[_|L1],
		length2(L1,N1).


% ith/3 that works both ways
% ith(Index,List,Element)
ith(Index,List,Element) :-
	term_type(Index,T),
	(T =:= 2		% integer
	 ->	ith0(Index,List,Element)
	 ;	ith1(List,1,Index,Element)
	).

ith0(I,[X|L],Y) :-
	I > 0,
	(I =< 1
	 ->	Y=X
	 ;	I1 is I-1,
		ith0(I1,L,Y)
	).

ith1([X|_],I,I,X).
ith1([_|L],I1,I,X) :- I2 is I1+1, ith1(L,I2,I,X).

% copy_term/2 creates a new variant of a term. It does NOT copy ground
% subterms.
/*      Term: to copy, 
        Buff: to copy it into, 
        Worddisp: to put answer or ptr to anser,
        Start: of free space to put subterms,
        End: (returned) space used on subterms,
        Size: of buffer to know direction to point pointers.
        Grd: var if subterms are ground, 1 if not.
*/

copy_term(X,Y) :- 
	%%\+ (\+ (X=Y)),
	term_type(X,Type),
	(Type =:= 0	%var
	 ->	true
	 ; Type =:= 1	% str
	 ->	buff_alloc(65536,Buff,0),       %heap
		copy_term1(X,Y,Buff)
	 ; Type =:= 3	% list
	 ->	buff_alloc(65536,Buff,0),       %heap
		copy_term1(X,Y,Buff)
	 ;	X=Y
	).

copy_term1(X,_Y,Buff) :-
        copytermng(X,Buff,0,4,Size,65536,_),
	buff_set_word(Buff,65532,Size),
	fail.
copy_term1(X,Y,Buff) :-
        buff_word(Buff,65532,Size),
        (Size =:= 4
         ->     buff_dealloc(Buff,65536,0,0), X=Y
         ;      buff_cell(Buff,0,Z),
                buff_dealloc(Buff,65536,Size,0),
                Z=Y
        ).

copytermng(Term,Buff,Worddisp,Start,End,Size,Grd) :-
	term_type(Term,Type),
	(Type =:= 0	% var
	 ->	Grd=1, buff_set_var(Buff, Worddisp, Size, Term), End=Start
	 ; Type =:= 3	% list
	 ->	ListStr is Buff + Start,
		buff_set_cell(Buff,Worddisp,3, ListStr), % ptr to list rec		
		%Wordsize=4,
		Arg2loc is Start+4,
		Newstart is Arg2loc+4, % reserve rec space
		Term=[A1|A2],
		copytermng(A1,Buff,Start,Newstart,End0,Size,Grd1),
		copytermng(A2,Buff,Arg2loc,End0,End1,Size,Grd1),
		(var(Grd1) 
 		->	End=Start, buff_set_cell(Buff,Worddisp,7,Term)
		;	End=End1,Grd=Grd1
		)
	 ; Type =:= 1	% con/str
	 ->	Str is Buff + Start,
		buff_set_cell(Buff,Worddisp,1,Str), % ptr to str rec
		term_psc(Term, Psc),
		buff_set_word(Buff, Start, Psc),
		%Wordsize=4,     %wordsize(Wordsize),
		psc_arity(Psc,Arity),
		Argsloc is Start + Arity*4,
		Newstart is Argsloc+4,
		copyargs(Term,Arity,Buff,Argsloc,Newstart,End1,Size,Grd1),
		(var(Grd1) 
		 ->	End=Start, buff_set_cell(Buff,Worddisp,7,Term)
		 ;	End=End1,Grd=Grd1
		)
	 ;	buff_set_cell(Buff, Worddisp, Type,Term), End=Start
	 
	).
        
copyargs(Term,Argno,Buff,Argloc,Start,End,Size,Grd) :- 
	Argno =:= 0
	 ->	End=Start
	 ;	term_arg(Term,Argno,Arg),
	 	copytermng(Arg,Buff,Argloc,Start,Mid,Size,Grd),
		%Wordsize=4,    %wordsize(Wordsize),
		Nargno is Argno-1, Nargloc is Argloc-4,
		copyargs(Term,Nargno,Buff,Nargloc,Mid,End,Size,Grd).

/*-------------define load_dyn for reading in dynamic predicates -----*/

load_dyn(Filename) :-
	seeing(Oldfile),
	expand_file_name(Filename, FileName),
	see(FileName),
	read_and_assert,
	seen,
	see(Oldfile),
	fail.		% reclaim all space, only side-effects
load_dyn(_).

read_and_assert :-
	read_all(Clauses,[],Commands,[]),
	retract_clauses(Clauses,Commands),
	assert_clauses(Clauses),
	call_all(Commands).

read_all(Clauses0,Clauses,Commands0,Commands) :-
	read(Term),
	(Term == end_of_file
	 ->	Clauses = Clauses0, Commands = Commands0
	 ; Term = :-(Dir)
	 ->	proc_directive(Dir,Commands0,Commands1),
		read_all(Clauses0,Clauses,Commands1,Commands)
	 ;	expand_term(Term,Term1),
		Clauses0 = [Term1|Clauses1],
		read_all(Clauses1,Clauses,Commands0,Commands)
	).

retract_clauses([],_).
retract_clauses([Clause|Clauses],Commands) :-
	split(Clause,Head,_Body),
	(retractable(Head)
	 ->	functor(Head,F,A),
		(memberchk(:-(multifile(F/A)),Commands)
		 ->	true
		 ;	abolish(F/A)
		)
	 ;	true
	),
	retract_clauses(Clauses,Commands).

split((Head:-Body),Head,Body) :- !.
split(Head,Head,true).

retractable(:-(_)) :- !,fail.
retractable(X) :- \+ predicate_property(X, built_in).

assert_clauses([]).
assert_clauses([Clause|Clauses]) :-
	assert(Clause),
	assert_clauses(Clauses).

call_all([]).
call_all([:-(Cmd)|Commands]) :-
	(Cmd = multifile(_)
	 ->	true
	 ; call(Cmd)
	 ->	true
	 ;	warning(('Command failed: ', Cmd))
	),
	call_all(Commands).


/* directives (should be integrated somehow with the parser) */

proc_directive(export(_X), Dirs, Dirs) :- !,
	warning('export directive ignored.').
proc_directive(import(from(X, Mod)), Dirs, Dirs) :- !,
	import(from(X, Mod)).
proc_directive(type(_X), Dirs, Dirs) :- !,
	warning('type directive ignored.').
proc_directive(index(_X), Dirs, Dirs) :- !,
	warning('index directive ignored.').
proc_directive(mode(_X), Dirs, Dirs) :- !,
	warning('mode directive ignored.').
proc_directive(parallel(_X), Dirs, Dirs) :- !,
	warning('parallel directive ignored.').
proc_directive(fun(_X), Dirs, Dirs) :- !,
	warning('fun directive ignored.').
proc_directive(file(_X), Dirs, Dirs) :- !,
	warning('file directive ignored.').
proc_directive(table(Pred), Dirs, Dirs) :- !,
	default_call_size(Chs), default_ret_size(Rhs),
	table(Pred,Chs,Rhs).
proc_directive(table(Pred,Chs), Dirs, Dirs) :- !,
	default_ret_size(Rhs), table(Pred,Chs,Rhs).
proc_directive(table(Pred,Chs,Rhs), Dirs, Dirs) :- !,
	table(Pred,Chs,Rhs).
proc_directive(table_all, Dirs, Dirs) :- !, 
	warning('table_all directive ignored. Use table/n').
proc_directive(op(P,T,S), Dirs, Dirs) :- !, op(P,T,S).
proc_directive(hilog(X), Dirs, Dirs) :- !, add_hilog_symbol(X).
proc_directive(Dir, [:-(Dir)|Dirs], Dirs).

/* --------------------- end of file basics.P ------------------------- */
