/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  consult.P
  Author(s)		:  Kostis F. Sagonas, David S. Warren
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  July 19, 1993
========================================================================*/

:- export consult/1, consult/2, '.'/2, 
	  reconsult/1, reconsult/2,
	  compile/1, compile/2.

:- import libpath/1 from loader.
:- import compile/4 from compile.
:- import compile_cH/5 from foreign.
:- import rm/1, sys_link/3 from unix.
:- import file_write0/2 from file_write.
:- import xsb_flag/2 from current_symbols.
:- import str_substring/4, str_cat/3 from string.
:- import abort/1, call/1, file_exists/1, atom/1, '=='/2 from standard.
:- import code_load/3, code_call/3, file_put/2, buff_byte/3, str_len/2,
	  expand_file_name/2, file_time/2
     from machine.

/*======================================================================*/
/*======================================================================*/
compile(Path) :- compile(Path, []).	% compile with no options

compile(X, _) :-
	var(X), !,
	abort('Uninstantiated argument 1 of compile/[1,2], cannot compile!').
compile([], _) :- !.
compile([H|T], Options) :- !,
	\+ (\+ (compile0(H,Options))),
	compile(T, Options).
compile(Path, Options) :- \+ (\+ (compile0(Path,Options))).

compile0(Path, Options) :-
	atom(Path),
	expand_file_name(Path, Expanded_Path),
	search_module(Expanded_Path, Dir, Mod, SExt, _HasObj, Base, _Obj),
	SExt =\= 0'O,
	!,
	compile_f(SExt, Base, Options, _, Mod, Dir).
compile0(Path, _) :-
	file_write0(2, 'Cannot find the file/module '),
	file_write0(2, Path), file_put(2, 10), fail.

compile_f(0'P, Base, Opts, ExList, Mod, _) :- 
	compile(Mod, Base, Opts, ExList).
compile_f(0'c, Base, Opts, ExList, Mod, Dir) :- 
	compile_cH(Mod, Base, Opts, ExList, Dir).
compile_f(0'x, Base, Opts, ExList, Mod, _) :-
	str_cat(Base, '.P', FileName),
	sys_link(Base, FileName, _),
	compile(Mod, Base, Opts, ExList),
	rm(FileName).

/*======================================================================*/
/* Predicates consult/[1,2] are thereafter similar to reconsult/[1,2].	*/
/*======================================================================*/

consult(Path) :- reconsult(Path, []).
consult(Path, Options) :- \+ (\+ (reconsult(Path, Options))).

[File|Rest] :- reconsult(File), ( Rest == [] -> true ; call(Rest) ).

/*======================================================================*/
/*======================================================================*/

reconsult(Path) :- reconsult(Path, []).

reconsult(Path, Options) :- 
	( nonvar(Path) -> 
		( atom(Path) -> \+ (\+ (reconsult0(Path, Options)))
		; abort('Wrong type in argument 1 of reconsult/[1,2], cannot reconsult!')
		)
	; abort('Uninstantiated argument 1 of reconsult/[1,2], cannot reconsult!')
	).

reconsult0(user, Options) :- !, compile_load_user(Options).
reconsult0(Path, Options) :- 
	expand_file_name(Path, Expanded_Path),
        search_module(Expanded_Path, Dir, Mod, SExt, HasObj, Base, Obj),
	!,
	( (SExt =\= 0'O , i_want_compile(SExt, Base, Obj, HasObj)) ->
		compile_f(SExt, Base, Options, _ExList, Mod, Dir)
	; true
	),
	load_exec(Obj, Mod).
reconsult0(Path, _) :-
        file_write0(2, 'Cannot find the file/module '),
        file_write0(2, Path), file_put(2, 10), fail.

i_want_compile(SExt, Base, Obj, HasObj) :-
	( HasObj =:= 0 -> true ; needs_recompile(SExt, Base, Obj) ).

needs_recompile(0'P, Base, Obj) :-
	str_cat(Base, '.P', PFileName),
	file_time(Obj, OTime), 
	file_time(PFileName, PTime),
	OTime < PTime.
needs_recompile(0'x, Base, Obj) :-
	file_time(Obj, OTime), 
	file_time(Base, PTime),
	OTime < PTime.
%%%needs_recompile(0'o, _Base, _Obj).
% As a temporary solution a foreign module needs to be recompiled
% if either the .c file or if the .H file needs to be recompiled.
needs_recompile(0'c, Base, Obj) :-
	( ( xsb_flag(host_type, 'Solaris') -> str_cat(Base, '.so', C_Obj)
	  ; str_cat(Base, '.o', C_Obj)
	  ),
	  ( file_exists(C_Obj) -> 
		file_time(C_Obj, COTime),
		str_cat(Base, '.c', CFileName),
		file_time(CFileName, CTime),
		COTime < CTime
	  ; true
	  )
	)
	; 
	( str_cat(Base, '.H', HFileName),
	  file_time(Obj, OTime), 
	  file_time(HFileName, HTime), 
	  OTime < HTime
	).


/*======================================================================*/
/*  search_module(+FileName, -Dir, -Mod, -SExt, -HasObj, -Base, -Obj)	*/
/*                                                                      */
/*	Given a FileName (must be an atom), search_module/7 finds the	*/
/*	location of the corresponding module and returns:		*/
/*		1) the directory (Dir),					*/
/*		2) the module name (Mod),				*/
/*		3) the extension (SExt) of the source file		*/
/*		   (P/o/c/x[other name]/O[no source file]),		*/
/*		4) whether an object code file exists (HasObj = 0,1),	*/
/*		5) the base name of the file (Base = Dir+Mod), and	*/
/*		6) the object file name (Obj)				*/
/* 	The call will fail if the specified file cannot be found.	*/
/*======================================================================*/

search_module(Name, Dir, Mod, SExt, HasObj, Base, Obj) :-
	str_len(Name, Len),
	search_slash(Name, Len, Slash), % If Slash =:= -1 no slash is found
	Loc1 is Len - 2,	   % The following is obviously a non-general
	buff_byte(Name, Loc1, B1), % way of searching for the '.?' suffix
	( B1 =:= 0'. -> Dot = Loc1 ; Dot = Len ),
	( Slash < 0, Dot =:= Len -> 
		Mod = Name, real_search_module(Mod, Dir, Base, SExt)
	; 				% is a full file name
	  ( Dot =:= Len -> 
		% if the file name has no extension first check for the
		% existence of the file with the extension.
		( str_cat(Name, '.P', PName), file_exists(PName), SExt = 0'P
	%%%	; str_cat(Name, '.o', PName), file_exists(PName), SExt = 0'o
		; str_cat(Name, '.c', PName), file_exists(PName), SExt = 0'c
		; file_exists(Name), SExt = 0'x
		; str_cat(Name, '.O', PName), file_exists(PName), SExt = 0'O
		)
	  ; file_exists(Name), Loc3 is Len - 1, buff_byte(Name, Loc3, SExt)
	  ),
	  ( Slash < 0 -> 
		Dir = './', 
		str_substring(Name, 0, Dot, Mod), 
		str_cat(Dir, Mod, Base)
	  ; str_substring(Name, 0, Slash, Dir),
	    Loc2 is Slash + 1, Till_Dot is Dot - Loc2,
	    str_substring(Name, Loc2, Till_Dot, Mod),
	    str_substring(Name, 0, Dot, Base)
	  )
	),
	str_cat(Base, '.O', Obj),
	( file_exists(Obj) -> HasObj = 1 ; HasObj = 0 ).

/*======================================================================*/
/* real_search_module(+ModName, -Dir, -Base, -Ext)			*/
/*======================================================================*/

real_search_module(ModName, Dir, Base, Ext) :-
        libpath(Dir),
        str_cat(Dir, ModName, Base),
	( str_cat(Base, '.P', Path), file_exists(Path), Ext = 0'P
	%%%; str_cat(Base, '.o', Path), file_exists(Path), Ext = 0'o
	; str_cat(Base, '.c', Path), file_exists(Path), Ext = 0'c
	; file_exists(Base), Ext = 0'x		% no suffix
	; str_cat(Base, '.O', Path), file_exists(Path), Ext = 0'O
	).

/*======================================================================*/
/* search_slash(+FileName, +InLoc, -SlashLoc)				*/
/*	Search for the occurrence of a '/' from the InLoc'th character	*/
/*	of FileName (backwards), and return the offset of the last '/'	*/
/*	character. If no '/' exists, a -1 is returned.			*/
/*======================================================================*/

search_slash(FileName, I, SlashLoc) :-
	( I =:= 0 -> SlashLoc = -1
	; I1 is I - 1,
	  buff_byte(FileName, I1, Byte),
	  ( Byte =:= 0'/ -> SlashLoc = I1 
	  ; search_slash(FileName, I1, SlashLoc) 
	  )
	).	

/*======================================================================*/
/* load_exec(+File, +ModName)						*/
/*	Changed to use the C loader (code_load/3) that automatically	*/
/*	imports all exported predicates of the module into the current	*/
/*	working module when its third argunemt is 1.  - Kostis (4/3/93)	*/
/*======================================================================*/

load_exec(File, ModName) :-	
	code_load(File, Init, 1),	% use the changed C loader.
	( Init > 0			% load succeeded
	; Init =< 0, file_write0(2, 'Error in loading file '),
		file_write0(2, File),
		file_put(2, 10), fail
	),
	!,
	file_write0(1, '['), file_write0(1, ModName),
	file_write0(1, ' loaded]'), file_put(1, 10),
	code_call(Init, true, 2).	% call '_$main'

compile_load_user(Options) :-
	compile(user, user, Options, _), !,
	load_exec('user.O', user),
	rm('user.O').

