/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  copyterm.P
  Author(s)		:  David S. Warren
  Modified by		:  Jiyang Xu
  Last modification	:  August 13, 1990
========================================================================*/

:- export copyterm/4, copyterm/6, termsize/2.

:- import buff_alloc/3, buff_set_cell/4, buff_set_var/4, buff_set_word/3,
	term_psc/2, psc_arity/2, term_type/2, stat_flag/2
    from machine.

:- import atomic/1, arg/3, atom/1, integer/1, real/1
    from standard.

/*======================================================================*/
/* function that determines the size of a word.				*/
/* 	will be implemented as function. But this library is made	*/
/*	by V 3.3, which does not support functions.			*/
/*======================================================================*/

wordsize(8) :- stat_flag(18, Word), Word = 3, !.
wordsize(4).

/*======================================================================*/
/*======================================================================*/
/* This routine copies a term into a buffer. It is passed:
	Term: the term to copy,
	Buffer: the buffer to copy it into,
	Worddisp: the word of the buffer in which to put the copy (or
		a pointer to the copy.)
	Start: offset of the next free location in the buffer, before the 
		copy is done.
	End: (returned) offset of the first free location after the
		copying.

	Variables are copied into the buffer and the copied variables are
	pointed into the buffer and trailed. Thus later binding of these 
	`outside' variables will cause the copied variables to be changed,
	too. If, however, the copyterm call is failed over, the variables
	in the buffer will be ``disconnected'' from the outer variables.

	Copyterm is a prime candidate for moving down into the emulator
	as a builtin written in C.
*/

copyterm(Term, Perm, Buff, SizeByte) :- 
	termsize(Term, Size),
	wordsize(Wordsize),
	SizeByte is Size * Wordsize,
	buff_alloc(SizeByte, Buff, Perm),
	copyterm(Term, Buff, 0, Wordsize, _End, SizeByte).

copyterm(Term, Buff, Worddisp, Start, Start, Size) :-
	var(Term),!, buff_set_var(Buff, Worddisp, Size, Term).
copyterm(Term, Buff, Worddisp, Start, Start, _Size) :-
	integer(Term),!, buff_set_cell(Buff, Worddisp, 2, Term).
copyterm(Term, Buff, Worddisp, Start, Start, _Size) :-
	atom(Term), !, buff_set_cell(Buff, Worddisp, 5, Term).
				% assume atom as string
copyterm(Term, Buff, Worddisp, Start, Start, _Size) :-
	real(Term), !, buff_set_cell(Buff, Worddisp, 6,Term).
copyterm(Term, Buff, Worddisp, Start, End, Size) :-
	Term=[_|_],!,
	ListStr is Buff + Start,
	buff_set_cell(Buff, Worddisp, 3, ListStr), /* ptr to list rec */
	wordsize(Wordsize),
	Newstart is Start + 2 * Wordsize, /* reserve rec space */
	copyargs(Term, 1, 2, Buff, Start, Newstart, End, Size).
copyterm(Term, Buff, Worddisp, Start, End, Size) :- 
	is_structure(Term), !,
	Str is Buff + Start,
	buff_set_cell(Buff, Worddisp, 1, Str), /* ptr to str rec */
	term_psc(Term, Psc),
	buff_set_word(Buff, Start, Psc),
	wordsize(Wordsize),
	Argsloc is Start + Wordsize,
	psc_arity(Psc,Arity),
	Newstart is Argsloc+Arity*Wordsize, /* reserve rec space*/
	copyargs(Term, 1, Arity, Buff, Argsloc, Newstart, End, Size).
	
copyargs(Term, Argno, Maxargs, Buff, Argloc, Start, End, Size) :- 
	Argno > Maxargs,
	 Start=End;
	Argno =< Maxargs,
	 arg(Argno, Term, Arg),
         copyterm(Arg, Buff, Argloc, Start, Mid, Size),
	 wordsize(Wordsize),
	 Nargno is Argno+1, Nargloc is Argloc+Wordsize,
	 copyargs(Term, Nargno, Maxargs, Buff, Nargloc, Mid, End, Size).

/*======================================================================*/
/* termsize(+Term, -Size)						*/
/*	Return the number of words (cells) used by the term.		*/
/*======================================================================*/

termsize(Term, Size) :- 
	termsize0(Term, Size1),
	Size is Size1 + 1.

termsize0(Term, 0) :- var(Term),!.
termsize0(Term, 0) :- atomic(Term),!.
termsize0(Term, Size) :- Term=[_|_],!,
	termargsize(Term, 1, 2, S1),
	Size is S1 + 2.
termsize0(Term, Size) :- 
	term_psc(Term, Psc),
	psc_arity(Psc, Arity),
	termargsize(Term, 1, Arity, S1),
	Size is S1 + Arity + 1.
	
termargsize(_Term, Argno, Maxargs, 0) :- Argno > Maxargs.
termargsize(Term, Argno, Maxargs, Size) :- 
	Argno =< Maxargs,
	arg(Argno, Term, Arg),
	termsize0(Arg, S1),
	Nargno is Argno+1,
	termargsize(Term, Nargno, Maxargs, S2),
	Size is S1 + S2.


is_structure(X) :- term_type(X, Type), Type =:= 1.
