/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  db.P
  Author(s)		:  David S. Warren
  Modified by		:  Kostis Sagnas, Jiyang Xu, David S. Warren
  Last modification	:  May 18, 1993
========================================================================*/


/*======================================================================*/
/* This module contains routines that handle predicate and clause	*/
/* references, in order to support assert and retract in our system.	*/
/*									*/
/* A predicate reference (Prref) is a database reference to a sequence  */
/* of clauses. Normally a Prref is associated with a psc-entry 		*/
/* (in the e.p.  field), the psc entry of the main functor symbol of 	*/
/* all the clauses.  But that need not be the case. A Prref can be	*/
/* created, asserted to, and called explicitly. 			*/
/*									*/
/* See the description of the formats of these records in assert.P	*/
/*======================================================================*/

:- export db_new_prref/3.

:- import dbgen_inst/4 from dbcmpl.
:- import buff_alloc/3, buff_set_word/3, psc_set_type/3, psc_set_prop/3,
	tip_prop/4, psc_arity/2
	from machine.
:- import '_$table'/3 from file_op.

/*======================================================================*/
/* db_new_prref(+PSC,-Prref): 						*/
/*	creates an empty Prref, i.e.  one with no clauses in it.  	*/
/*	If called, it will simply fail.  Prref must be a variable at 	*/
/*	the time of call. It points the entry point of the psc record	*/
/*	pointed to by PSC to this Prref. So in effect it makes it	*/
/*	dynamic.	 						*/
/*	dsw:								*/
/*	Prref ->:							*/
/*		-4: Addr of last Clref on ALL chain			*/
/*		0: BC instruction: fail (if empty),			*/
/*			jump and save breg (if nonempty)		*/
/*		4: Addr of first Clref on ALL chain			*/
/*======================================================================*/
db_new_prref(PSC,PHead,Prref) :-
	buff_alloc(16, Prref0, 1),	/* use multiple of 8 */
	subbuffer(Prref0, 4, Prref),
	buff_set_word(Prref, -4, Prref),
	dbgen_inst(fail,Prref,0,_),
	buff_set_word(Prref, 8, Prref),
	psc_set_type(PSC, 1, 1),	/* Make it dynamic */
	('_$table'(PHead,Callsize,Retsize)
	 ->	buff_alloc(40, Tabinsbuf, 1),
		buff_alloc(16, Tip, 1),
		psc_arity(PSC,Arity),
		Alloccode is Tabinsbuf+12,
		dbgen_inst(tabletryfail(Arity,Alloccode,Tip),Tabinsbuf,0,L0),
		dbgen_inst(allocate,Tabinsbuf,L0,L1),
		dbgen_inst(getpbreg(2),Tabinsbuf,L1,L2),
		dbgen_inst(calld(Prref,3),Tabinsbuf,L2,L3),
		dbgen_inst(addtotable(Arity,2),Tabinsbuf,L3,L4),
		dbgen_inst(deallocate,Tabinsbuf,L4,L5),
		dbgen_inst(proceed,Tabinsbuf,L5,_),
		tip_prop(0,4,0,First_tip),
		tip_prop(0,4,1,Tip),	% set new first tip
		buff_set_word(Tip,0,First_tip),	% fill in tip
		buff_set_word(Tip,4,0),
		buff_set_word(Tip,8,Callsize),
		buff_set_word(Tip,12,Retsize),
		psc_set_prop(PSC, Tabinsbuf, 1)	% set psc to point to table ins
	 ;	psc_set_prop(PSC, Prref, 1)	/* set ep to Prref */
	).

/*======================================================================*/
/* -- Aux routines ------------------					*/
/*======================================================================*/

subbuffer(Buff,Offset,SubBuff) :- SubBuff is Buff+Offset.

