/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB License Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  db_clause.P
  Author(s)		:  Kostis F. Sagonas, David S. Warren
  Last modification	:  May 18, 1993
========================================================================*/

:- set_inthandler('_$clause_int'(_),16).	% MYSIG_CLAUSE = 16


/************************************************************************/
/* Module dbclause defines routines to step through clauses and 	*/
/* return them, and also to delete clauses for retracting. Space is	*/
/* reclaimed. (See module assert for formats of data structures.	*/
/************************************************************************/

% db_retract(+Head,+Body,+Prref) deletes clauses and reclaims space

db_retract(Head,Body,Prref) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Prref,0,Head,Clref,Typ),
	db_clause_match(Head,Tbody,Clref,Typ),
	db_erase(Clref).

% Don't reclaim space
db_retract_nr(Head,Body,Prref) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Prref,0,Head,Clref,Typ),
	db_clause_match(Head,Tbody,Clref,Typ),
	db_erase_nr(Clref).	% Don't reclaim space

% transform a goal to turn cuts (!) into '_$cutto's. Differs from
% goal_cut_trans in that variables are not wrapped with call/1,
% but are left unchanged.

goal_cut_trans_a(X,X,_) :- var(X),!.
goal_cut_trans_a([X|Y],Z,_) :- !, hack_consult([X|Y],Z). %  a HACK!
goal_cut_trans_a(!,'_$cutto'(Cutpoint),Cutpoint) :- !.
goal_cut_trans_a((A,B),(Na,Nb),Cutpoint) :- !,
	goal_cut_trans_a(A,Na,Cutpoint),
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a((A;B),(Na;Nb),Cutpoint) :- !,
	goal_cut_trans_a(A,Na,Cutpoint),
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a((A->B),(A->Nb),Cutpoint) :- !,
	goal_cut_trans_a(B,Nb,Cutpoint).
goal_cut_trans_a(X,X,_).

hack_consult([-X],reconsult(X)) :- !.
hack_consult([X],consult(X)) :- !.
hack_consult([-X|Y],(reconsult(X),Z)) :- !,hack_consult(Y,Z).
hack_consult([X|Y],(consult(X),Z)) :- !,hack_consult(Y,Z).


% db_retractall(+Call,+Prref) deletes all clauses matching a pattern

db_retractall(Call,Prref) :-
	db_retract(Call,_,Prref),fail.
db_retractall(_,_).

% db_abolish(+Call,+Prref) deletes all clauses of a predicate.
%	(currently same as retractall, will need extending.)

db_abolish(Call,Prref) :-
	db_retractall(Call,Prref),
	buff_byte(Prref,0,Op1),
	(Op1 =:= 197	% tabletryfail
	 ->	buff_word(Prref,24,Prrefa),
		Prrefab is Prrefa-4,
		buff_dealloc(Prrefab,16,0,1),
		buff_dealloc(Prref,40,0,1)
	 ;	Prrefab is Prref-4,
		buff_dealloc(Prrefab,16,0,1)
	).

% db_clause(+Head,+Body,+Prref,-Clref) retrieves a clause.

db_clause(Head,Body,Prref,Clref) :-
	goal_cut_trans_a(Body,Tbody,_),
	db_get_clauses(Prref,0,Head,Clref,Typ),
	(Typ =:= 2	% compiled code
	 ->	Body = '$compiled'	
	 ;	db_clause_match(Head,Tbody,Clref,Typ)
	).

:- buff_alloc(8,Buff,1),asserti(db_body_save(Buff),0).

% db_clause_match(+Call,+Body,+Clref,+Typ)
%	Determines whether Clref contains a clause whose head matches Call.
%	Type indicates the kind	of Clref (indexed or not). 

db_clause_match(Head,Body,Clref,Typ) :-
	(Typ =:= 1
	 ->	Entry is Clref + 24
	 ;	Entry is Clref + 8
	),
	db_body_save(Buff),
	buff_set_cell(Buff,0,5,true),	% set default
	(db_do_call(Entry,Head)
	 ->	stat_set_flag(24,0,0),
	 	db_body_save(Buff),
		buff_cell(Buff,0,Body0),
		Body = Body0
	 ;	stat_set_flag(24,0,0),
	 	fail
	).

% db_do_call(+CodeEntry, +Head)
%	Makes the call to Head on the heap after setting the interrupt flag
%	that will retrieve the code from the heap.  By this time we finally
%	know that Head is "dynamic" (psc_type = 1) so there is no need to
%	retrieve the type.  If at some later version we need the psc type 
%	the following should appear:
%			conpsc(Head, PSC), psc_type(PSC, Type),
%	BEFORE stat_set_flag/3.  Under NO circumstances, should code be put
%	between stat_set_flag/3 and code_call/3, because the hack of building
%	the body of the call on the heap will not work!!.

db_do_call(Entry, Head) :-
	stat_set_flag(24, 1, 0),	% set clause interrupt on (CLAUSE_INT)
	code_call(Entry, Head, 1).	% Type =:= 1 means dynamic predicate.

db_body_save(_) :- fail.	% gotta fix this, by having dynamic!

'_$clause_int'(Call) :-
	stat_set_flag(24,0,0),	% turn clause intercept off
 	db_body_save(Buff),
	buff_set_var(Buff,0,8,X),
	X=Call.

% db_get_clauses(+Prref,+Ind, +Term,-Clref,-Typ) returns Clrefs for each clause
%	whose head might unify with Term. It uses the SOB index blocks to
%	efficiently find them. It may return clauses whose head won't unify
%	with Term. N starts at 0, and is incremented each time one goes
%	through an SOB for a multiple index. So it indicates the number of the
%	index used.

db_get_clauses(Prref,Ind,Term,Clref,Typ) :-
	buff_byte(Prref, 0, Op1),
	(Op1 =:= 197	% tabletryfail, skip table instr
	 ->	buff_word(Prref,24,Prrefa),	% where the prref is
		buff_byte(Prrefa, 0, Op)
	 ;	Prrefa = Prref,
		Op = Op1
	),
	% must be jump and save breg or jump if Prref is SOB 
	(Op =:= 170 -> true ; Op =:= 240 -> true),
	buff_word(Prrefa, 4, Fclref),
	get_clauses(Fclref, Ind, Term, Clref, Typ).

% get_clauses(+Clref0,+Ind,+Term,-Clref,-T) returns Clrefs that contain code.
%	Clref will be a buffer address (at 0), and T=0 if nonindexed, T=3 if
%	indexed, T=2 if header for compiled code.

get_clauses(Clref0,Ind,Term,Clref,T) :-
	buff_word(Clref0, -8, Tword),
	T1 is Tword /\ 3,
	get_clauses1(T1, Clref0, Ind, Term, Clref, T).
get_clauses(Clref0, Ind, Term, Clref, T) :-
	buff_byte(Clref0, 0, Inst),
	another_clause(Inst, Clref0, Ind, Term, Clref, T).

% another_clause(+Op,+Clref,+Ind,+Term,-Clref,-T) gets the next clause, if 
%	there is one.

another_clause(160 /*try*/ , Clref0, Ind, Term, Clref, T) :- 
	buff_word(Clref0, 4, Clref1),
	get_clauses(Clref1, Ind, Term, Clref, T).
another_clause(161 /*retry*/ , Clref0, Ind, Term, Clref, T) :- 
	buff_word(Clref0, 4, Clref1),
	get_clauses(Clref1, Ind, Term, Clref, T).

% get_clauses1(+T1,+Clref0,+Ind,+Term,-Clref, -T) returns one clause if 
%	unindexed; otherwise returns all clauses from SOB block.

get_clauses1(0, Clref, _, _, Clref, 0).	% return this unindexed code
get_clauses1(1, Clref0, Ind, Term, Clref, T) :-	
	buff_byte(Clref0, 20, Op),	% sob block
	Op =\= 248,	% not a fail inst, so has clauses
	buff_byte(Clref0, 9, Argno0),
	buff_byte(Clref0, 10, Argno1),
	buff_byte(Clref0, 11, Argno2),
	buff_word(Clref0,16,TabSize),
	get_joint_hash(Argno0,Argno1,Argno2,Term,TabSize,Hashval),
	(Hashval < 0
	 ->	buff_word(Clref0, 24, Clref1),	% all chain or another SOBindex
		buff_word(Clref1, -8, Blt), 
		((Blt /\ 3) =:= 1	% another index
		 ->	Sobprref is Clref0 + 20, Ind1 is Ind+1,
			db_get_clauses(Sobprref,Ind1,Term,Clref,T)
		 ; 	T = 3, get_clauses_in_sob(Clref1, -1, Clref)
		)
	 ;	Bucket is 32 + 4 * Hashval,
		buff_word(Clref0, Bucket, Clref1),
		Faddr is Clref0 + 32 + 4 * TabSize,	%% dsw
		Clref1 =\= Faddr, 		/* something in it */
		T = 1,
	 	get_clauses_in_sob(Clref1, Ind, Clref)
	).
get_clauses1(2, Clref, _, _, Clref, 2).	% return header for compiled code


get_joint_hash(Argno1,Argno2,Argno3,Term,TabSize,Hashval) :-
	Argno1 =:= 0
	 ->	(Argno2 =:= 0
		 ->	arg(Argno3,Term,Arg),
			(nonvar(Arg)
			 ->	term_hash(Arg,TabSize,Hashval)
			 ;	Hashval = -1
			)
		 ;	arg(Argno2,Term,Arg1), arg(Argno3,Term,Arg2),
			((nonvar(Arg1),nonvar(Arg2))
			 ->	term_hash(Arg1,TabSize,H1),
				term_hash(Arg2,TabSize,H2),
				Hashval is ((H1 << 1) + H2) mod TabSize
			 ;	Hashval = -1
			)
		)
	 ;	arg(Argno1,Term,Arg1), arg(Argno2,Term,Arg2),
		arg(Argno3,Term,Arg3),
		((nonvar(Arg1),nonvar(Arg2),nonvar(Arg3))
		 ->	term_hash(Arg1,TabSize,H1),term_hash(Arg2,TabSize,H2),
			term_hash(Arg3,TabSize,H3),
			Hashval is ((H1 << 2) + (H2 << 1) + H3) mod TabSize
		 ;	Hashval = -1
		)
	.



% get_clauses_in_sob(+Clref0,+Ind,-Clref) returns clauses in SOB block.

get_clauses_in_sob(Clref0, Ind, Clref) :- Clref is Clref0 - (Ind+1)*16.
get_clauses_in_sob(Clref0, Ind, Clref) :-
	buff_byte(Clref0, 0, Inst),
	next_sob(Inst, Clref0, Ind, Clref).

% next_sob(+Op,+Clref0,+Ind,-Clref) returns next sob clause, if there is one.

next_sob(160 /*try*/ , Clref0, Ind, Clref) :-
	buff_word(Clref0, 4, Clref1),
	get_clauses_in_sob(Clref1, Ind, Clref).
next_sob(161 /*retry*/ , Clref0, Ind, Clref) :-
	buff_word(Clref0, 4, Clref1),
	get_clauses_in_sob(Clref1, Ind, Clref).

/************************************************************************/
/* The following routines delete clauses from Prrefs, but do NOT	*/
/* reclaim the space.							*/
/************************************************************************/

% db_erase_nr(+Clref) removes the clause Clref from the all.index chains but
%	does NOT reclaim the space.
db_erase_nr(Clref) :-
	buff_word(Clref,-8,Lenflg),
	Flag is Lenflg /\ 3,
	db_erase_type_nr(Flag,Clref).

db_erase_type_nr(0,Clref) :-		% unindexed
	dbgen_inst(fail,Clref,8,_).	% set clause to fail
db_erase_type_nr(2,_Clref) :-		% compiled code header
	abort('Cannot erase compiled code').
db_erase_type_nr(3,Iclref) :-		% indexed
	buff_byte(Iclref,11,Toskip),
	Addr is 12 + Toskip*2,
	dbgen_inst(fail,Iclref,Addr,_).	% set clause to fail

/************************************************************************/
/* The following routine cleans up by reclaiming space for all clauses	*/
/* that were erased without reclaiming their space.			*/
/************************************************************************/

db_reclaim_space(Head,Prref):-
	db_get_clauses(Prref,0,Head,Clref,Typ),
	erase_if_failed(Typ,Clref),
	fail.
db_reclaim_space(_,_).

erase_if_failed(0,Clref) :-		% nonindexed ref
	buff_byte(Clref, 8, Op),	% must be a fail instruction
	Op =:= 248,
	db_erase(Clref).		% if so, erase it
erase_if_failed(3,Clref) :-		% an indexed ref
	buff_byte(Clref,11,Toskip),
	Addr is 12 + Toskip*2,
	buff_byte(Clref, Addr, Op),	% must be a fail instruction
	Op =:= 248,
	db_erase(Clref).		% if so, erase it.

/************************************************************************/
/* The following routines delete clauses from Prrefs, and reclaim the	*/
/* space. They are modeled on the idea of erase, and should be extended	*/
/* in that direction.							*/
/************************************************************************/

% db_erase(+Clref) removes the clause Clref from the all.index chains and
%	reclaims the space

db_erase(Clref) :-
	buff_word(Clref,-8,Lenflg),
	Flag is Lenflg /\ 3,
	db_erase_type(Flag,Clref),
	Len is Lenflg /\ 8'777777774,
	Buffer is Clref - 8,
	buff_dealloc(Buffer, Len, 0, 1).

% db_erase_type(+Typ,+Clref) removes the Clref. Typ=1 if indexed, =0 if not.

db_erase_type(0,Clref) :-		% unindexed
	buff_byte(Clref, 0, Inst),
	buff_word(Clref, -4, Pclref),	% previous
	buff_word(Clref, 4, Nclref),	% next
	dbgen_inst(fail,Clref,8,_),	% set fail for protection
	buff_set_word(Pclref, 4, Nclref),
	(Inst =:= 161 			% retrymeelse
	 ->	buff_set_word(Nclref, -4, Pclref)  % true
	 ; Inst =:= 160		% trymeelse
	 ->	buff_set_word(Nclref, -4, Pclref),
		buff_byte(Nclref, 0, B1),
		(B1 =:= 186		% dyntrustmeelse->noop
		 ->	dbgen_inst(noop(2), Nclref, 0, _)
		 ;	B1 =:= 161,		% retrymeelse->trymeelse
			dbgen_opcode(trymeelse(_,_),Nclref,0)
		)
	; Inst =:= 186		% dyntrustmeelsefail
	->	buff_byte(Pclref, 0, B2),
		(B2 =:= 161	% retrymeelse
		 ->	dbgen_opcode(dyntrustmeelsefail(_,_),Pclref,0)
		 ;	B2 =:= 160,	% trymeelse
		 	dbgen_inst(noop(2), Pclref, 0, _) % prev is only, noop
		),
		buff_set_word(Nclref, 8, Pclref)
	;	Inst =:= 249,	% deleting last
		buff_byte(Pclref,0,Pop),
		(Pop =:= 170	% jumptbreg so prref
		 ->	dbgen_opcode(fail,Pclref,0),
			buff_set_word(Nclref, 8, Pclref)
		 ;	Psobbuf is Pclref - 20,
			buff_word(Psobbuf, -8, Lenflg),	% turn reg and delet
			Nlenflg is Lenflg /\ 8'777777774,
			buff_set_word(Psobbuf, -8, Nlenflg),
			db_erase(Psobbuf)
		)
	).
db_erase_type(2,_Clref) :-		% compiled code header
	abort('Cannot erase compiled code').
db_erase_type(3,Clref) :-		% indexed
	db_erase_type_i(Clref),
	db_erase_type_a(Clref).

% db_erase_type_i(+Iclref) erases the indexed clref.
	
db_erase_type_i(Iclref) :-
	buff_byte(Iclref, 11, Toskip),	% number of halfwords to skip
	Clref is Iclref + 16,
	buff_byte(Clref, 0, Inst),	% process index-chain
	dbgen_inst(fail,Iclref,24,_),	% set to fail for safety
	buff_word(Clref, -4, Pclref),	% previous
	buff_word(Clref, 4, Nclref),	% next
	(Inst =:= 161 			% retrymeelse
	 ->	buff_set_word(Pclref, 4, Nclref),
	 	buff_set_word(Nclref, -4, Pclref)
	 ; Inst =:= 160		% trymeelse
	 ->	buff_set_word(Pclref, 0, Nclref),	% set first bucket ptr
	 	buff_set_word(Nclref, -4, Pclref),
	 	buff_byte(Nclref, 0, B1),
		(B1 =:= 186		% dyntrustmeelse->noop
		 ->	dbgen_inst(noop(2), Nclref, 0, _)
		 ;	B1 =:= 161,		% retrymeelse->trymeelse
			dbgen_opcode(trymeelse(_,_),Nclref,0)
		)
	; Inst =:= 186		% dyntrustmeelsefail
	->	buff_set_word(Pclref, 4, Nclref),	% prev's next is bucket
		buff_byte(Pclref, 0, B2),
		(B2 =:= 161	% retrymeelse
		 ->	dbgen_opcode(dyntrustmeelsefail(_,_),Pclref,0)
		 ;	B2 =:= 160,	% trymeelse
		 	dbgen_inst(noop(2), Pclref, 0, _) % prev is only, noop
		)
	;	Inst =:= 249,	% deleting last
		find_fail_addr(Pclref,Faddr),
		buff_set_word(Pclref, 0, Faddr)		% pnt buckt to failinst
	),
	(Toskip =< 6	% and then recurse for next index
	 ->	true
	 ;	db_erase_type_i(Clref)
	).

% find_fail_addr(+Bucket,-Faddr) finds the fail address for bucket with
%	address Bucket. It scans from the current bucket toward the end
%	of the table, checking to see if any bucket contains the fail
%	address. This is somewhat of a hack, but seems necessary with 
%	the current representation. Maybe someday it will be fixed?

find_fail_addr(Bucket, Faddr) :-
	Nbuck is Bucket + 4,
	buff_byte(Nbuck, 0, Op),
	(Op =:= 248	%fail?
	 ->	Faddr = Nbuck
	 ;	buff_word(Nbuck, 0, Addr),
		buff_byte(Addr, 0, Op1),
		(Op1 =:= 248
		 ->	Faddr = Addr
		 ;	find_fail_addr(Nbuck,Faddr)
		)
	).
	
% db_erase_type_a(+Clref) erases the unindexed clref.

db_erase_type_a(Clref) :-
	buff_byte(Clref, 0, Inst),	% process all-chain
	buff_word(Clref, -4, Pclref),	% previous
	buff_word(Clref, 4, Nclref),	% next
	(Inst =:= 161 			% retrymeelse
	 ->	buff_set_word(Pclref, 4, Nclref),
	 	buff_set_word(Nclref, -4, Pclref)
	 ; Inst =:= 160		% trymeelse
	 ->	buff_set_word(Pclref, 24, Nclref),	% set SOB first ptr
	 	buff_set_word(Nclref, -4, Pclref),
	 	buff_byte(Nclref, 0, B1),
		(B1 =:= 186		% dyntrustmeelse->noop
		 ->	dbgen_inst(noop(2), Nclref, 0, _)
		 ;	B1 =:= 161,		% retrymeelse->trymeelse
			dbgen_opcode(trymeelse(_,_), Nclref, 0)
		)
	; Inst =:= 186		% dyntrustmeelsefail
	->	buff_set_word(Pclref, 4, Nclref),
		buff_set_word(Nclref, 28, Pclref),	% set SOB last ptr
		buff_byte(Pclref, 0, B2),
		(B2 =:= 161	% retrymeelse
		 ->	dbgen_opcode(dyntrustmeelsefail(_,_),Pclref,0)
		 ;	B2 =:= 160,	% trymeelse
		 	dbgen_inst(noop(2), Pclref, 0, _) % prev is only, noop
		)
	;	Inst =:= 249,	% deleting last in entire SOB block
		buff_word(Pclref, -8, Lenflg),	% turn into reg clref and delet
		Nlenflg is Lenflg /\ 8'777777774,
		buff_set_word(Pclref, -8, Nlenflg),
		db_erase(Pclref)
	).

