/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  dbcmpl.P
  Author(s)		:  David S. Warren, Kostis F. Sagonas
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  May 18, 1993
========================================================================*/


:- export dbgen_inst/4.
:- export dbgen_opcode/3.

:- import asm_inst/4 from asm_inst.
:- import buff_set_byte/3, term_psc/2,
	  buff_set_word/3, term_type/2,
	  buff_set_cell/4
    from machine.

/*======================================================================*/
/* dbgen_inst: Generate an instruction in the buffer			*/
/*======================================================================*/

dbgen_inst(Inst, Buff, LocIn, LocOut) :-
	asm_inst(Inst, _Len, Opcode, Args),
	buff_set_byte(Buff, LocIn, Opcode),
	LocMid is LocIn + 1,
	dbgen_args(Args, Buff, LocMid, LocOut).

dbgen_args([], _, LocIn, LocIn).
dbgen_args([One|More], Buff, LocIn, LocOut) :-
	dbgen_arg(One, Buff, LocIn, LocMid),
	dbgen_args(More, Buff, LocMid, LocOut).

dbgen_arg(p, _Buff, LocIn, LocOut) :- LocOut is LocIn + 1.
dbgen_arg(pp, _Buff, LocIn, LocOut) :- LocOut is LocIn + 2.
dbgen_arg(ppp, _Buff, LocIn, LocOut) :- LocOut is LocIn + 3.
dbgen_arg(s(Str), Buff, LocIn, LocOut) :- LocOut is LocIn + 4,
	term_psc(Str, PSC),
	buff_set_word(Buff, LocIn, PSC).
dbgen_arg(c(Sym), Buff, LocIn, LocOut) :- LocOut is LocIn + 4,
	buff_set_word(Buff, LocIn, Sym).
	     /* Trick here! Treat Term as an integer or string.  */
             /* May cause problem if modifying format.           */
dbgen_arg(g(_Sym), _Buff, LocIn, LocOut) :- LocOut is LocIn + 4.
dbgen_arg(n(I), Buff, LocIn, LocOut) :- 
	term_type(I,TT), 
	(TT =:= 2
	 ->	buff_set_word(Buff, LocIn, I)	% integer
	 ;	buff_set_cell(Buff, LocIn, 6, I)	% float
	),
	LocOut is LocIn + 4.
dbgen_arg(i(I), Buff, LocIn, LocOut) :-  LocOut is LocIn + 4, 
	buff_set_word(Buff, LocIn, I).
dbgen_arg(l(L), Buff, LocIn, LocOut) :- LocOut is LocIn + 4,
	buff_set_word(Buff, LocIn, L).
dbgen_arg(r(I), Buff, LocIn, LocOut) :- LocOut is LocIn + 1, 
	buff_set_byte(Buff, LocIn, I).
dbgen_arg(v(I), Buff, LocIn, LocOut) :- LocOut is LocIn + 1, 
	buff_set_byte(Buff, LocIn, I).
dbgen_arg(a(I), Buff, LocIn, LocOut) :- LocOut is LocIn + 1, 
	buff_set_byte(Buff, LocIn, I).
/* dbgen_arg(t(T), Buff, LocIn, LocOut) :- LocOut is LocIn + 1. */



/*======================================================================*/
/* dbgen_opcode: Generate an opcode into a buffer			*/
/*======================================================================*/

dbgen_opcode(Inst,Buff,Disp) :-
	asm_inst(Inst,_Len,Opcode,_Args),
	buff_set_byte(Buff,Disp,Opcode).
