/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
   File			:  domain.P
   Author(s)		:  Jiyang Xu, Kostis Sagonas
   Modified by		:  Jiyang Xu
   Last modification	:  April 10, 1993
========================================================================*/


:- export ':'/2, int/3, int/1, intype/1, '/='/2, plus/3, times/3, in/2.
:- export deleteff/3, deleteff_p/3, generate/1, generate_ff/1.

:- import integer/1, arg/3, write/1, nl/0, call/1, op/3 from standard.
:- import dom_range/3, dom_size/2, dom_elem/2, dom_diff/3, dom_enum/2,
	dom_plus/3, dom_times/3, dom_type/2, dom_min/2,
	term_type/2, term_psc/2, term_new/2, psc_arity/2, dom_esize/2
    from machine.

% To compile this file the following operator declarations are needed.

:- op(700,xfx,in).
:- op(700,xfx,'/=').

/*== Type Checking ':'/2 ============================================ */

X : [M, N] :- !, 		/* integer subrange type */
	integer(M),
	integer(N),
	term_type(X, TX),
	( TX =:= 0 -> dom_range(X, M, N)
	; TX =:= 2 -> X >= M, X =< N
	; TX =:= 14 -> dom_type(X, TD), 
	  ( TD =:= 17 ->
	    dom_min(X, M0), dom_size(X, S),
	    ( M0 = M, N is M0 + S - 1, !
	    ; dom_range(Y, M, N), Y = X
	    )
	  ; TD =:= 16
	  )
	).
X : int(_) :- !,		/* integer type */
	term_type(X, TX),
	( TX =:= 0 -> dom_range(X, -100000000, 100000000)
	; TX =:= 2 -> true
	; TX =:= 14 -> dom_type(X, TD),
	   ( TD =:= 16 -> true
	   ; TD =:= 17 -> true
	   )
	).
_X : float(_) :- !,		/* floating point type */
	write('"X : float(_)" is not implemented yet.'), nl.
X : Type :-			/* Assume finite type for now */
	term_type(Type, TT), TT = 1,
	term_psc(Type, Psc),
	term_type(X, TX),
	( TX =:= 0 -> dom_enum(X, Psc)
	; TX =:= 5 -> X in Type
	; TX =:= 14 -> dom_type(X, TD), TD = 18,	% DOMATOM
		dom_min(X, Psc0), Psc=Psc0
	).


/*== Generic Type Predicate in/2 =====================================*/

X in [M, N] :- !, int(X, M, N).
X in Type :- term_type(Type, TT), TT = 1,
	term_psc(Type, Psc),
	term_new(Psc, Call),
	psc_arity(Psc, A),
	call_args(A, Type, Call),
	arg(1, Call, X),
	!,
	call(Call).

call_args(1, _, _).
call_args(I, Type, Call) :- I > 1,
	arg(I, Type, X),
	arg(I, Call, X),
	J is I - 1,
	call_args(J, Type, Call).

/*== Builtin Type int/3 ===============================================*/

int(X, M, N) :- integer(M), integer(N),
	term_type(X, T),
	( T =:= 2 -> X >= M, X =< N			% integer
	; T =:= 14 -> dom_type(X, DT), 
	  ( DT =:= 17 ->		% DOMRANGE
	    dom_size(X, S),
	    dom_elem_enum(X, 0, S)
	  ; DT =:= 19 ->		% DOMVAR
	    dom_range(Y, M, N),
	    S is N-M+1,
	    dom_elem_enum(Y, 0, S),
	    X = Y
	  )
	; T =:= 0 ->					% variable
	    dom_range(X, M, N),
	    S is N-M+1,
	    dom_elem_enum(X, 0, S)
	).

dom_elem_enum(X, I, _) :- dom_elem(X, I).
dom_elem_enum(X, I, N) :- J is I + 1, J < N, dom_elem_enum(X, J, N).

/*=== Builtin Type int/1 ============================================*/

int(X) :- term_type(X, T),
	( T =:= 2 -> true
	; T =:= 14 -> dom_type(X, DT), DT =:= 17,
		dom_size(X, S), dom_elem_enum(X, 0, S)
	; T =:= 0 -> enum_int(0, X)
	).

enum_int(X, X).
enum_int(X, Y) :- X1 is X + 1, enum_int(X1, Y).

/*=== Constraint Builtins ===========================================*/

intype(X) :- term_type(X, T), 
	( T =:= 14 -> dom_size(X, N), dom_elem_enum(X, 0, N)
	; true
	).

plus(X, Y, Z) :- dom_plus(X, Y, Z).

times(X, Y, Z) :- dom_times(X, Y, Z).

X /= Y :- var(Y), !, dom_diff(X, Y, 0).
X /= Y+D :- !, integer(D), dom_diff(X, Y, D).
X /= Y-D :- !, integer(D), dom_diff(Y, X, D).
X /= Y :- !, dom_diff(Y, X, 0).

deleteff_p(List, X, Rest) :-
	deleteff_p_findmin(List, 256, Min),
	deleteff_p_select(List, X, Rest, Min).
	
deleteff_p_findmin([], Min, Min).
deleteff_p_findmin([(_,H)|T], Min0, Min1) :-
	term_type(H, TT),
	( TT =:= 14 -> dom_esize(H, DS),
		( DS < Min0 -> deleteff_p_findmin(T, DS, Min1)
		; deleteff_p_findmin(T, Min0, Min1)
		)
	; TT =:= 0 -> fail
	; TT = 1
	).

deleteff_p_select([(ID,H)|T], (ID,H), T, Min) :- 
	term_type(H, TT),
	( TT =:= 14 -> dom_esize(H, DS), DS =< Min
	; Min =:= 1
	),
	!.
deleteff_p_select([H|T], X, [H|NT], Min) :- 
	deleteff_p_select(T, X, NT, Min).

deleteff(List, X, Rest) :- 
	deleteff_findmin(List, 256, Min),
	deleteff_select(List, X, Rest, Min).

deleteff_findmin([], Min, Min).
deleteff_findmin([H|T], Min0, Min1) :-
	term_type(H, TT),
	( TT =:= 14 -> dom_esize(H, DS),
		( DS < Min0 -> deleteff_findmin(T, DS, Min1)
		; deleteff_findmin(T, Min0, Min1)
		)
	; TT =:= 0 -> fail
	; TT = 1
	).

deleteff_select([H|T], H, T, Min) :- 
	term_type(H, TT),
	( TT =:= 14 -> dom_esize(H, DS), DS =< Min
	; Min =:= 1
	),
	!.
deleteff_select([H|T], X, [H|NT], Min) :- 
	deleteff_select(T, X, NT, Min).

generate([]).
generate([H|T]) :- intype(H), generate(T).

generate_ff([]).
generate_ff([H|T]) :- 
	deleteff([H|T], X, R),
	intype(X), 
	generate_ff(R).

