/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  foreign.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  April 3, 1993
========================================================================*/


:- export compile_cH/5, compile_c/3, compile_H/4.

:- import unix/1 from unix.
:- import str_cat/3 from string.
:- import str_len/2 from machine.
:- import write_int/1 from prof_sbp.
:- import xsb_flag/2 from current_symbols.
:- import attach/2, length/2 from listutil.
:- import message/1, time_message/2 from aux.
:- import read/1, see/1, seen/0, write/1, put/1, tell/1, told/0,
	  '=='/2, file_exists/1, cputime/1, abort/1
     from standard.

compile_cH(Modname, Base, Options, ExList, Dir) :-
	compile_c(Modname, Base, Dir),
	compile_H(Modname, Base, Options, ExList).

compile_c(Modname, Base, Dir) :-
	message(('[ Compiling C file ', Base, '.c ]')),
	( xsb_flag(host_type, 'Solaris') ->
		unix(['cc -O2 -s -o ', Base, '.so', ' -G ', Base, '.c'])
	; unix(['cc -c ', Base, '.c']),
	  /* The following appears here because C compilers usually put the 
	     object file in the current directory and not in the source's one.*/
	  ( Dir == './' -> true ; unix(['/bin/mv ', Modname, '.o ', Dir]) )
	).

compile_H(Modname, Base, Options, ExList) :-
	str_cat(Base, '.H', HFile),
	( file_exists(HFile) ->
		( Options = [_|_] -> M = ' (compile options ignored)' ; M = '' ),
		message(('[ Compiling Foreign Module ', Base, M, ' ]')),
		cputime(TimeStart),
		see(HFile),
		read_clauses(ExList, LdOption),
		seen,
		length(ExList, Count),
		str_cat(Base, '.O', OFile),
		tell(OFile),
		co_asmhead(Modname, LdOption, Count),
		co_asmsym(ExList),
		told,
		time_message(TimeStart, (Modname, ' compiled'))
	; abort(('File ', HFile, ' for foreign module does not exist'))
	),
	!.

read_clauses(ExList, LdOption) :-
	read(Clause),
	( Clause == end_of_file -> true
        ;   ( Clause = (':-'(export(Exports))) ->
		add_to_exlist(Exports, ExList)
	    ; Clause = (':-'(ldoption(Opt))) ->
		( LdOption=Opt ; true )
	    ; true			% all other clauses are ignored
	    ),
            read_clauses(ExList, LdOption)
        ).

add_to_exlist(P/A, ExList) :- !, attach(P/A, ExList).
add_to_exlist((Pred, Preds), ListType) :- !,
        add_to_exlist(Pred, ListType),
        add_to_exlist(Preds, ListType).

co_asmhead(Modname, LdOption, PSC_Count) :-
	co_magic(8),
	str_len(Modname, Length),
	put(Length),
	write(Modname),
	( var(LdOption) -> put(0)	% No ldoption was specified.
	; str_len(LdOption, LdoLen), put(LdoLen), write(LdOption)
	),
	write_int(PSC_Count).

co_asmsym([]).
co_asmsym([One|ExList]) :-
	co_one(One),
	co_asmsym(ExList).

co_one(P/A) :-
	str_len(P, L),
	put(0),		% domain = ex(ported)
	put(13),	% type = T_FORN
	put(A),		% arity = a
	put(L),		% length
	write(P).		

co_magic(N) :- put(17), put(18), put(19), put(N).

	
/*======================================================================*/
/* format: 
		+-----------------------+
		| magic = 8		|
		+-----------------------+------------------+
		| len  |  module name (len size)	   |
		+-----------------------+----------+-------+
		| len  |  ld option		   |
		+-----------------------+----------+
		| psc count		|
		+-----------------------+

		+-----------------------+-------------------------------+
	count:	| 0 | 13 | 0 |arity|len |  symbol name			|
		+-----------------------+-------------------------------+
		... ... ... ...
		... ... ... ...

	Note that a foreign module is always a Module (len size =/= 0).
*/
/*======================================================================*/
