/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  numbervars.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu, David S. Warren
  Last modification	:  May 21, 1992
========================================================================*/


%   numbervars(+Term, +FirstN, ?LastN) 
%   Each variable in the HiLog Term is instantiated to a term of the form 
%   '$VAR'(N), where N is an integer starting from FirstN. FirstN is used
%   as the value of  N  for the first variable in Term (starting from the 
%   left).  The second distinct variable in  Term  is given a value  of N
%   satisfying "N is FirstVar + 1";  the third distinct variable gets the 
%   value FirstVar+2, and so on.  The last variable in Term has the value 
%   LastVar-1.

numbervars(Y,I,J) :- term_type(Y,T),
	(T =:= 0
	 ->	Y='$VAR'(I), J is I+1
	 ; T =:= 3
	 ->	Y=[A1|A2],
	 	numbervars(A1,I,I1), numbervars(A2,I1,J)
	 ; T =:= 1
	 ->	term_psc(Y,PSC), psc_arity(PSC,N),
	 	numbervars(Y,I,J,1,N)
	 ;	J=I
	).
numbervars(Y,I,J,N,A) :-
	(N > A
	 ->	J=I
	 ;	term_arg(Y,N,Arg),
	 	numbervars(Arg,I,I1),
	 	N1 is N+1, numbervars(Y,I1,J,N1,A)
	).

%   numbervars(+Term)
%   makes a HiLog Term ground by binding variables to subterms $VAR(N) with
%   integer values of N ranging from 0 up. This predicate was written since
%   most of the calls to numbervars/3 are of the form: numbervars(Term,0,_)
 
numbervars(Term) :-
        numbervars(Term, 0, _).

