/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1993				*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  prof_lib.P
  Author(s)		:  Jiyang Xu
  Modified by		:  Kostis F. Sagonas, Jiyang Xu
  Last modification	:  September 21, 1992
========================================================================*/


:- export oneof/1, measure/1, par_member/2, mapset/3, '&'/2, maplist/3.

:- import length/2 from basics.
:- import findall/3 from setof.
:- import stat_sta/1 from machine.
:- import call/1, arg/3, '=..'/2 from standard.
:- import par_member_try/3, term_arg/3, term_new/2, psc_insert/3, 
	  term_type/2
     from machine.

/*======================================================================*/
/*======================================================================*/

oneof(Call) :- call(Call), !.

/*======================================================================*/
/*======================================================================*/

measure(Call) :- stat_sta(2), oneof(Call), fail.
measure(_Call) :- stat_sta(3).

/*======================================================================*/
/*======================================================================*/

par_member(X, L) :-
	term_type(L, TT),
	TT =:= 3,			% is a list
	length(L, N),
	( N > 240 -> split_member(L, LL),
		par_member(L1, LL),
		par_member(X, L1)
	; psc_insert(parmemtry, N, Psc),
	  term_new(Psc, Term),
	  copy_args(L, Term, 1),
	  par_member_try0(X0, Term, N),
	  X = X0
	).

par_member_try0(X, T, N) :- par_member_try(X, T, N).
	/* in order to put par_member_try as the last goal */

copy_args([], _, _).
copy_args([H|T], Term, I) :-
	term_arg(Term, I, X), X = H,
	I1 is I + 1,
	copy_args(T, Term, I1).

split_member([], []) :- !.
split_member(L, [L1|LLR]) :-
	split_1(240, L, L1, LR),
	split_member(LR, LLR).

split_1(I, L, L1, LR) :-
	( I =:= 0 -> L1 = [], LR = L
	; L = [] -> L1 = [], LR = []
	; L = [X|T], L1 = [X|T1], I1 is I - 1,
	  split_1(I1, T, T1, LR)
	).

/*======================================================================*/
/*======================================================================*/

mapset(Call, InList, OutList) :-
	findall(X, mapset_one(X, InList, Call), OutList).

mapset_one(X, InList, Call) :-
	par_member(Y, InList),
	arg(1, Call, Y),
	arg(2, Call, X),
	oneof(Call).


/*======================================================================*/
/*======================================================================*/

maplist(Goal, InList, OutList) :-
	Goal =.. [Pred, _, _|Others],
	Call =.. [Pred, _, _|Others],
	map_tag(InList, InListT, OutList, OutListT, 1),
	findall(Y, mapset_one(Call, InListT, Y), OutListU),
	map_sort(OutListU, OutListT).

mapset_one(Call, InList, I-Y) :-
	par_member(I-X, InList),
	arg(1, Call, X),
	arg(2, Call, Y),
	oneof(Call).

map_tag([], [], [], [], _).
map_tag([X|Xs], [I-X|IXs], [Y|Ys], [I-Y|IYs], I) :-
	I1 is I + 1,
	map_tag(Xs, IXs, Ys, IYs, I1).

map_sort([], _).
map_sort([I-Y|R], OutListT) :-
	s_member(I-Y, OutListT),
	map_sort(R, OutListT).

/*======================================================================*/
/*======================================================================*/

&(X, Y) :- 
	findall(I-Z, parand_solve(I, X, Y, Z), Bag),
	s_member(a-X, Bag),
	s_member(b-Y, Bag).

:- parallel parand_solve/4. 

parand_solve(a, X, _, X) :- call(X).
parand_solve(b, _, Y, Y) :- call(Y).

s_member(X, [X|_]).
s_member(X, [_|L]) :- s_member(X, L).

/*======================================================================*/
/* test routines							*/
/*======================================================================*/
/*
test(R) :- maplist(p(2, 4, 1), [1,2,3], R).

p(X, Y, C) :- Y is 2 * X + C.

*/
