/************************************************************************/
/*                                                                      */
/* XSB System                                                           */
/* Copyright (C) SUNY at Stony Brook, 1993                              */
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB, */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along   */
/* with XSB so you can know your rights and responsibilities.           */
/* It should be in a file named LICENSE.                                */
/* Among other things, this notice must be preserved on all copies.     */
/*                                                                      */
/************************************************************************/

/*======================================================================
  File                  :  tables.P
  Author(s)		:  Swift 
  Last modification	:  August, 1993
======================================================================*/

:- export 
	  abolish_all_tables/0,
          abolish_table_call/1, 	% not yet implemented
	  abolish_table_pred/1, 
	  analyze_table/1, 	
	  comp_call/1,
	  etnot/1,
	  get_returns_for_call/2,
	  get_calls_for_table/3,
	  robust_term_psc/2,
	  set_pil_on/0, set_pil_off/0, /* convienence */
	  set_prof_on/1, set_prof_off/0, /* convienence */
	  set_xtrace_on/0, set_xtrace_off/0, /* convienence */
	  show_table/1, 		% debugging.
	  table_complete/2, 		% obsolete
	  table_copy/2, 
	  table_prop/3, table_prop/4,
	  table_state/2,
	  tfindall/3, tnot/1,
	  zero_out_profile_table/0,write_out_profile_table/0,
	
	  get_call_structure_nc/2.	%not for general use.

%	  get_tip_val/3, 
%	  set_tip_size/3, 
%	  table_inst/2, 
% can be found in old_tables.P

:- import 

	abolish_table_info/0,
	analyze_predicate_table/1, 
	buff_alloc/3, buff_cell/3, buff_dealloc/4,
	buff_set_word/3, buff_word/3 , 
	call_hash/5, comp_code_call/2, conpsc/2, cut_if_leader/3, 
	get_osp_breg/2, 
	heap_copyarg/2,
%        incomplete/2, 
	is_incomplete/2, 
	print_predicate_table/1,
%         psc_arity/2, 
        psc_insert/4, psc_tabled/2, psc_type/2,
	stat_flag/2, stat_set_flag/3, 
%	suspend/3,
	table_intern/2,
	term_new/2,  
%	term_psc/2, 
	term_type/2,tip_prop/4,
	zero_out_profile/0,write_out_profile/0
	from machine.

:- import write/1, '=='/2, ttywritenl/1, callable/1, call/1,
	  '\=='/2, integer/1, functor/3
	  from standard.  

:- import  abort/1, 
           call_expose/1,
	   goal_cut_trans/3
	   from standard.

:- import load_pred/1 from loader.

:- import ground/1 from basics.

:- import copyterm/6 from copyterm.

/* need to come to a decision about F/A */

zero_out_profile_table:-
	zero_out_profile.

write_out_profile_table:-
	write_out_profile.

robust_term_psc(Term,Psc):-
	term_type(Term,Type),
	(Type =:= 1 -> conpsc(Term,Psc)
	 ; Type =:=5 -> conpsc(Term,Psc)
	  ; abort('Error: first arg to term_psc/1 is not callable')).

analyze_table('/'(F,A)):-!,
	functor(Term,F,A),
	robust_term_psc(Term,Psc),
	analyze_predicate_table(Psc).

get_calls_for_table(Skeleton,Call,Empty) :-
	robust_term_psc(Skeleton,PSC),
	psc_tabled(PSC,Tip),
	Tip =\= 0,
	buff_alloc(0,Toaddr,0),  		%get hreg addr.
	term_new(PSC,Call),
        functor(Call,_,Arity),
	tip_offset(call_hash_addr,CHAOFF),
	tip_offset(call_hash_size,CHSOFF),
	buff_word(Tip,CHAOFF,Hashtabaddr),
	buff_word(Tip,CHSOFF,CHS),
	call_buckets(Hashtabaddr,CHS,Bucket),
	get_entry_info(Bucket,Toaddr,Arity,Empty).

% given a hashtable address, return addr of first element in each bucket
call_buckets(Hashtabaddr,CHS,Bucket) :- 
	Hashtabaddr =\= 0,
	call_buckets_1(Hashtabaddr,CHS,Bucket).

call_buckets_1(Hashtabaddr,N,Bucket) :- 
	N > 0,
	(buff_word(Hashtabaddr,0,Bucket)
	 ;
	 N1 is N-1, 
	 Hashtabaddr1 is Hashtabaddr - 4,
	 call_buckets_1(Hashtabaddr1,N1,Bucket)
	).

% given a call_entry address, return ndet addr of each call on chain.
get_entry_info(Bucketaddr,Toaddr,Arity,Empty) :-
	Bucketaddr =\= 0,
	(Calladdr = Bucketaddr,
	 Addr is Calladdr-24,	
% skip next call, ret ptr, sus ptr, RHS, openptr.
	 load_args(Addr,Toaddr,1,Arity),
 	 call_structure_offset(retptr,Offset),
  	 buff_word(Calladdr,Offset,Return_hash),
	 (Return_hash < 0 -> Empty = empty ; Empty = notempty)
	 ;
	 buff_word(Bucketaddr,0,Nextcall),
	 get_entry_info(Nextcall,Toaddr,Arity,Empty)
	).

return_buckets(Hashtabaddr,RHS,Bucket) :- 
	Hashtabaddr > 0,
	return_buckets_1(Hashtabaddr,RHS,Bucket).

return_buckets_1(Hashtabaddr,N,Bucket) :- 
	N > 0,
	(buff_word(Hashtabaddr,0,Bucket)
	 ;
	 N1 is N-1, Hashtabaddr1 is Hashtabaddr - 4,
	 return_buckets_1(Hashtabaddr1,N1,Bucket)
	).

get_another_return(Bucketaddr,Toaddr,Arity) :-
	Bucketaddr =\= 0,
	(Addr is Bucketaddr-8,	
	 load_args(Addr,Toaddr,1,Arity)
	 ;
	 buff_word(Bucketaddr,0,Nextcall),
	 get_another_return(Nextcall,Toaddr,Arity)
	).

load_args(Addr,To,I,Arity) :-
	I > Arity
	 ->	true
	;	
		To1 is To + 4,
%		set_pil_on,
		heap_copyarg(Addr,To1),
		I1 is I+1, Addr1 is Addr-4,
		load_args(Addr1,To1,I1,Arity).

abolish_all_tables:- abolish_table_info.

abolish_table_pred('/'(F,A)):-!,
	functor(Term,F,A),
	abolish_table_pred_1(Term).
abolish_table_pred(Term):-
	abolish_table_pred_1(Term).

abolish_table_pred_1(Term):-
	robust_term_psc(Term,Psc),
	psc_tabled(Psc,Tip),
	(Tip =:= 0 -> abort('Error: Predicate not tabled')
	; 
 	 tip_transform(tab_entry,Num),
	 tip_prop(Tip,Num,1,0)).		% 1 = 'set'.

abolish_table_call(_Call):-
	write('abolish table call not yet imlemented').
/*
	get_tip_val(Call,call_hash_size,Chs),
	get_tip_val(Call,tab_entry,Ep),
	integer(Chs),integer(Ep),
  	Ep =/= 0,
	call_hash(Call,Ep,Ar,Chs,Call_str_ptr),
Need last
*/

table_prop(Pred,Type,Old,New):-
	get_tip_val(Pred,Type,Old),
	set_tip_size(Pred,Type,New).

table_prop(Pred,Type,Val):-
	get_tip_val(Pred,Type,Val).

set_tip_size('/'(F,A),Type_size,Val):-!,
	functor(Term,F,A),
	set_tip_size_1(Term,Type_size,Val).
set_tip_size(Term,Type_size,Val):-
	set_tip_size_1(Term,Type_size,Val).

set_tip_size_1(Term,Type_size,Val):-
	robust_term_psc(Term,Psc),
	psc_tabled(Psc,Tip),
	tip_transform(tab_entry,Num),
	tip_prop(Tip,Num,0,Val1),	% 0 = 'get'.
	(Val1 =:= 0 ->
	  tip_transform(Type_size,Num2),
	  tip_prop(Tip,Num2,1,Val)	% 1 = 'set'.
	;
	  abort('Error: Cannot reset call or return hash sizes unless table is abolished')).


get_tip_val('/'(F,A),Type,Val):-!,
	functor(Term,F,A),
	get_tip_val_1(Term,Type,Val).
get_tip_val(Term,Type,Val):-
	get_tip_val_1(Term,Type,Val).

get_tip_val_1(Term,Type,Val):-
	var(Val),
	robust_term_psc(Term,Psc),
	(Psc =:= 0 -> Val = undef
	 ; psc_tabled(Psc,Tip),
	   (Tip =:= 0 -> Val = undef
  	    ; tip_transform(Type,Num2),
	      tip_prop(Tip,Num2,0,Val) ) ).	% 0 = 'get'.

table_complete(Call,Result):-
	table_state(Call,Result).

table_state(Call,Result):-
	get_tip_val(Call,call_hash_size,Chs),
	get_tip_val(Call,tab_entry,Ep),
	( integer(Chs),integer(Ep) -> 
  	  (Ep =:= 0 -> Result = no_call_yet
	   ; functor(Call,_,Ar),
	     call_hash(Call,Ep,Ar,Chs,Call_str_ptr),
	     (Call_str_ptr \== 0 -> 
  	       (call_structure_offset(susptr,Offset),
	        buff_word(Call_str_ptr,Offset,Complete),
	        (Complete =:= -1 -> Result = complete
	         ; Result = incomplete ) )
	      ; Result = no_call_yet))
  	  ; Result = nottabled).

get_returns_for_call(Call,Ret):-
	robust_term_psc(Call,PSC),
	psc_tabled(PSC,Tip),
	Tip =\= 0,
	buff_alloc(0,Toaddr,0),  		%get hreg addr.
	term_new(PSC,Ret),
	get_tip_val(Call,call_hash_size,Chs),
	get_tip_val(Call,ret_hash_size,RHS),
	get_tip_val(Call,tab_entry,Ep),
	integer(Chs),integer(RHS),Ep =\= 0,
	functor(Call,_,Ar),
	call_hash(Call,Ep,Ar,Chs,Call_str_ptr),
        Call_str_ptr =\= 0,
  	call_structure_offset(retptr,Offset),
        buff_word(Call_str_ptr,Offset,Return_hash),
	Return_hash > 0,
        return_buckets(Return_hash,RHS,Ret_bucket),
        get_another_return(Ret_bucket,Toaddr,Ar).
	
call_structure_offset(retptr,-4).
call_structure_offset(susptr,-8).

tip_offset(call_hash_addr,4).
tip_offset(call_hash_size,8).
tip_offset(ret_hash_size,12).

%retstructsize(3).

tip_transform(tab_entry,1).
tip_transform(call_hash_size,2).
tip_transform(ret_hash_size,3).

table_copy(Ptr,New):-
	ground(Ptr),
	var(New),
	table_intern(Ptr,New).
	
tfindall(T, Call, Result) :-
	table_state(Call,Res),
	(Res == complete -> tfindall1(T,Call,Result)
	 ; Res == no_call_yet -> (call(Call),fail ; tfindall(T,Call,Result))
	 ; Res == incomplete -> comp_call(Call),tfindall1(T,Call,Result)).

tfindall1(T, Call, Result) :-
	buff_alloc(65536, Buff, 0),
	findall_1(T, Call, Result, Buff, 65536).

findall_1(T, Call, _Result, Buff, BuffLen) :-
	wordsize(Wordsize),
	copyterm([], Buff, 8, _, _, BuffLen), /* init result list to empty */
	buff_set_word(Buff, 0, 8), /* init where to put next answer */
	FirstFree is 8 + Wordsize,
	buff_set_word(Buff, 4, FirstFree), /* init first free place */
        Place=Place, Start=Start,End=End,
	call_expose(Call),
%	write(Call),nl,
%	comp_call(Call),
%	write(passed),write(Call),nl,
	buff_word(Buff, 0, Place), /* get where to put answer */
	buff_word(Buff, 4, Start), /* get first free place */
	copyterm([T], Buff, Place, Start, End, BuffLen),
	Tailloc is Start+Wordsize,
	buff_set_word(Buff, 0, Tailloc), /* where to put next answer */
				/* point to the tail [] */
	buff_set_word(Buff, 4, End), /* next first free place */
	End > BuffLen, ttywritenl('Findall buffer overflow'),
	fail.
findall_1(_, _Call, Result, Buff, BuffLen) :-
	buff_word(Buff, 4, Length),	/* Length =\= 12 fail if [] */
	buff_dealloc(Buff, BuffLen, Length, 0),
	buff_cell(Buff, 8, Result0),
	Result = Result0.

wordsize(8) :- stat_flag(18, Word), Word = 3, !.
wordsize(4).

tnot(X):-
     	(call(X),fail ;
	is_incomplete(X,Ret_ptr),
	buff_word(Ret_ptr,0,Val),
	Val =< 0).

etnot(X):-
	get_osp_breg(Os,Breg),
	(call(X),
	  cut_if_leader(X,Os,Breg),
	  fail 
	;
	  is_incomplete(X,Ret_ptr),
	  buff_word(Ret_ptr,0,Val),
	  Val =< 0).

comp_call(Goal) :- 
	( callable(Goal) ->
		'_$savecp'(C), 
		goal_cut_trans(Goal, Trans_Goal, C), 
		comp_call_c(Trans_Goal)
	; var(Goal) -> 
		abort('Error: Uninstantiated argument 1 of call/1')
	; abort('Error: Wrong type in argument 1 of call/1 (callable term expected)')
	).

comp_call_c(X) :-
	term_type(X, TT),
	( TT =:= 1 -> robust_term_psc(X, PSC)
	; TT =:= 5 -> psc_insert(X, 0, PSC, 0)
	),
	psc_type(PSC, T),
	( T =:= 2 -> true		% Regular predicate
	; load_pred(PSC)
	),
	psc_tabled(PSC,Tabled),Tabled \== 0,
	comp_code_call(PSC, X).

%%%%%%%%%%%%%%%%%%%%%%%%%%
% Used by compiler, so written w.o. frills.
%%%%%%%%%%%%%%%%%%%%%%%%%
get_call_structure_nc(Call,Ptr):-
	robust_term_psc(Call,Psc),
	psc_tabled(Psc,Tip),
        tip_prop(Tip,1,0,Ep),	% 0 = 'get', 1 = ep.
        tip_prop(Tip,2,0,Chs),	% 0 = 'get', 2 = chs.
	functor(Call,_,Ar),
	call_hash(Call,Ep,Ar,Chs,Ptr).

%%%%%%%%%%%%%%%%%%%%%%%%%%
% For testing: not for general use
%%%%%%%%%%%%%%%%%%%%%%%%%
/* obsolete 
hcopyarg(From):-
	set_pil_on,
	heap_copyarg(From,0).
*/

set_prof_on(Level):-
	stat_set_flag(52,Level,1).

set_prof_off:-
	stat_set_flag(52,0,1).

set_pil_on:-
	stat_set_flag(0,1,1).

set_pil_off:-
	stat_set_flag(0,0,1).

set_xtrace_on:-
	stat_set_flag(51,1,1).

set_xtrace_off:-
	stat_set_flag(51,0,1).

show_table('/'(F,A)):-!,
	functor(Term,F,A),
	robust_term_psc(Term,Psc),
	print_predicate_table(Psc).

/* needs to be updated
output_table(Table,File):-
	tell(File),
	output_table_1(Table),
	told.

output_table_1(Table):-
	table_inst(Table,Ret),
	numbervars(Ret,0,_),
	alt_write_canonical(Ret),	nl,
	fail.
output_table_1(_).
*/
end_of_file.

tnot(Goal):-
	comp_call(Goal),
	(call(Goal) -> fail ; true).

