/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  unix.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu
  Modified by		:  Jiyang Xu (August 13, 1990)
  Last modification	:  May 18, 1993
========================================================================*/


:- export unix/2, cd/1, unix/1, ls/0, rename/2, rm/1, edit/1.
:- export sys_chdir/2, sys_access/3, sys_stat/3, sys_link/3.

:- import str_cat/3 from string.
:- import atom/1, abort/1 from standard.
:- import psc_name/2, sys_system/2, term_psc/2, term_type/2,
	  expand_file_name/2
    from machine.


unix(Command) :- 
	term_type(Command, T),
	( T =:= 3 -> constr_command(Command, Command1), unix(Command1, Res)
	; T =:= 5 -> unix(Command, Res)
	),
	Res =:= 0.

constr_command([H], H).
constr_command([H|T], String) :-
	constr_command(T, SubString),
	str_cat(H, SubString, String).
	
unix(Command, Result) :-
	term_type(Command, T),
	( T =:= 5 -> sys_system(Command, R), Result = R
	; T =:= 1 -> term_psc(Command, Psc), psc_name(Psc, RCommand), 
		     sys_system(RCommand, R), Result = R
	).

cd(Path) :- 
	( atom(Path) -> 
		expand_file_name(Path, Dir), sys_chdir(Dir, Res), Res=:=0
	; var(Path) ->
		abort('Uninstantiated argument 1 of cd/1')
	; abort('Wrong type in argument 1 of cd/1')
	).

rename(Old, New) :- sys_rename(Old, New, Res), Res =:= 0.

ls :- unix('ls -F', Res), Res =:= 0.

rm(Name) :- sys_unlink(Name, Res), Res =:= 0.

edit(File) :- 
	expand_file_name(File, FileName), 
	unix(['$EDITOR ', '$EDITOR_OPTIONS ', FileName, '.P']).

/*
sys_exit() :- sys_syscall(1).
 */
sys_link(Source, Dest, Res) :-	sys_syscall(  9, Res, Source, Dest).
sys_unlink(Path, Res) 	    :-	sys_syscall( 10, Res, Path).
sys_chdir(Dir, Res)	    :- 	sys_syscall( 12, Res, Dir).
sys_access(File, Mode, Res) :- 	sys_syscall( 33, Res, File, Mode).
sys_stat(Path, Buff, Res)   :-	sys_syscall( 38, Res, Path, Buff).
sys_rename(Old, New, Res)   :-	sys_syscall(128, Res, Old, New).


/*
sys_syscall(_CallNo, _Res)				 :- '_$builtin'(55).
 */
sys_syscall(_CallNo, _Res, _A1)				 :- '_$builtin'(55).
sys_syscall(_CallNo, _Res, _A1, _A2)			 :- '_$builtin'(55).
/*
sys_syscall(_CallNo, _Res, _A1, _A2, _A3)		 :- '_$builtin'(55).
sys_syscall(_CallNo, _Res, _A1, _A2, _A3, _A4)		 :- '_$builtin'(55).
sys_syscall(_CallNo, _Res, _A1, _A2, _A3, _A4, _A5)	 :- '_$builtin'(55).
sys_syscall(_CallNo, _Res, _A1, _A2, _A3, _A4, _A5, _A6) :- '_$builtin'(55).
 */
