/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  interpreter.P
  Author(s)		:  David S. Warren, Terrance Swift, 
			   Kostis Sagonas, Jiyang Xu
  Last modification	:  April 5, 1993
========================================================================*/


/*======================================================================*/
/*  This file contains the main program of the interpreter.		*/
/*======================================================================*/
/*  WARNING: Should you change this file better consult the comment in	*/
/*	     predicate hack_libpath_changed/0 first!			*/
/*======================================================================*/

:- interpreter.

set_inthandler(Handler, Number) :-
	Flag is Number + 32,
	term_psc(Handler, PSC),
	stat_set_flag(Flag, PSC, 1).

interpreter :-
	set_inthandler('_$load_undef'(_), 0),           % MYSIG_UNDEF
	set_inthandler('_$keyboard_int'(_), 1),         % MYSIG_KEYB
	stat_flag(20, SP),
	( SP =:= 1, set_inthandler('_$stackfull'(_), 2) % MYSIG_OFMEM
	; SP =\= 1
	),
	( SP =:= 1, set_inthandler('_$stackfull'(_), 5) % MYSIG_OFTC
	; SP =\= 1
	),
	set_inthandler('_$calltonum'(_), 3),            % MYSIG_SPY
	set_inthandler('_$wakedelay'(_), 11),		% MYSIG_DELAY
	stat_set_flag(12, 0, 1),	% set current_module to "usermod"
	conpsc(current_module, PSC2),
	conpsc(usermod, PSC),
	psc_set_prop(PSC2, PSC, 1),
	conpsc(userin, Uin),		% set current_input to "userin"
	psc_set_prop(Uin, 0, 1),
	psc_set_type(Uin, 5, 1),
	stat_set_flag(10, 0, 1),
	conset(current_input, Uin),
	conpsc(userout, Uout), 		% set current_output to "userout"
	psc_set_prop(Uout, 1, 1),
	psc_set_type(Uout, 5, 1),
	conset(current_output, Uout),
	stat_set_flag(11, 1, 1),
	initial_import,			% import all standard predicates
	version_message,
	load_library_directories,	% check for user's eccentricities
	set_prompt('| ?- '),
	repeat,
		'_$savecp'(Cp),			% to set abort cut point
                conset('_$abort_cutpoint', Cp),
		stat_set_flag(5, 1, 1),		% hide call interceptions.  
/*		stat_set_flag(2,0,1),	% set flag to intercept stack oflow */
		stat_set_flag(7, 0,1),		% reset trace number
		conset('_$break_level', 0),
		print_debug_state,
		get_prompt(Prompt),
		file_write(1, Prompt),
		file_read(0, X, Vars),
		( nonvar(X), read_in_stop(X), !	
			% end loop; pop a break level, out if at top
		; process_query(X, Vars)
		).

read_in_stop(halt) :- halt. /* quit entire system, regardless of break level */
read_in_stop(end_of_file).	/* pop a break level, out if at top */

process_query(X, []) :-		/* no variables appeared in the input */
	!,
	call_query(X),
	!,
	file_put(1, 10), file_write(1, yes), file_put(1, 10),
	fail.
process_query(X, Vars) :- 
	call_query(X),
	print_answer(Vars),
/*	print_delaygoals,  */	/* need to test the delay mode;	      */
				/* or put the predicate in syslib !!! */
	no_more,
	close_open_tables,
	!,
	file_put(1, 10), file_write(1, yes), file_put(1, 10),
	fail.

no_more :- file_get(0, X), X =:= 10, !.
no_more :- no_more, fail.

call_query(X) :- 
%%%	'_$savecp'(C),			% moved into call/1, affect trace?
	func_trans(X, Y),
	( stat_set_flag(5, 0, 1) 		% expose tracing
	; stat_set_flag(5, 1, 1), fail		% hide again when backtrack
	),
	'_$call'(Y), 
	( stat_set_flag(5, 1, 1)		% hide tracing when succeed
	; stat_set_flag(5, 0, 1), fail		% expose when backtrack to user
	).
call_query(_) :- 
	file_put(1,10), file_write(1, no), file_put(1, 10), fail.

% hidden version of call
'_$call'(Y) :- 
	'_$savecp'(C), 
	goal_cut_trans(Y,X,C),
	call_expose(X).



print_answer([]) :- !.
print_answer([vv(Name, Val)|Tail]) :-
	file_put(1, 10), file_write(1, Name), 
	file_write(1, ' = '), file_write(1, Val),
	print_answer(Tail).


/* === abort and break ================================================= */

abort :- conval('_$abort_cutpoint', Cp), '_$cutto'(Cp), 
	stat_set_flag(2, 0, 1), 
        close_open_tables, fail.

break :- conval('_$break_level', Blevel),
	Nblevel is Blevel+1, conset('_$break_level', Nblevel),
	file_write(1, '[ Break (level '), file_write(1, Nblevel),
	file_write(1, ') ]'),file_put(1, 10),
	readl_brklp1,
	conset('_$break_level', Blevel),
	file_write(1, '[ End break (level '),file_write(1, Nblevel),
	file_write(1, ') ]'),file_put(1, 10).

readl_brklp1 :- 
	repeat, 
		conval('_$break_level', Blevel),
		file_write(1, Blevel), file_write(1, ': ?- '),
		file_read(0, X, Vars),
		( nonvar(X), read_in_stop(X), !
			% end read loop; pop a break level, out if X=halt.
		; process_query(X, Vars)
		).


/* === import ========================================================== */

import(from(Syms, Mod)) :- !,
	conname(Mod, MName),
	import1(Syms, MName).
% import(Mod) :- import all preds in that module. not implemented yet.

import1([], _) :- !.
import1((One,Rest), MName) :- !,
	import1(One, MName), import1(Rest, MName).
import1([One|Rest], MName) :- !,
	import1(One, MName), import1(Rest, MName).
import1(P/A, MName) :-
	conname(P, PName),
	is_integer(A),
	psc_import(PName, A, MName).


/* === hide and expose ================================================= */

'_$hide'(N) :- stat_flag(5, N), NewN is N + 1, stat_set_flag(5, NewN, 1).
			% N = hide_state++
'_$hide'(N) :- stat_flag(5, N), NewN is N - 1, stat_set_flag(5, NewN, 1), fail.
			% hide_state--
'_$expose'(N) :- stat_set_flag(5, N, 1).
			% hide_state = N
'_$expose'(N) :- N1 is N + 1, stat_set_flag(5, N1, 1), fail.
			% hide_state = N+1, fail


/* === initial_import ================================================== */

initial_import :-
	standard_symbol(P, A, [im(Mod)|_]),
	import(from(P/A, Mod)),
	fail.	% Take next one and do not leave any choice points behind.
initial_import.


/* === load_library_directories ======================================== */

load_library_directories :-
	stat_flag(49, HomeDir),	% Flag[49] contains user's home dir.
	conname('/.xsbrc.P', File),
	str_cat(HomeDir, File, FileName),
	machine_file_exists(FileName),
	!,
	reconsult(FileName),		% The file is reconsulted so that 
					% it is compiled if a .O file does
					% not exist or is obsolete.
	hack_libpath_changed,
	stat_set_flag(50, 1, 1).	% Setting flag[50] notifies the loader
					% that the .xsbrc file was consulted.
load_library_directories.

%-----------------------------------------------------------------------%
%   hack_libpath_changed/0						%
%									%
%   Forces all the files that are used by the XSB interpreter to be	%
%   loaded into the system before setting the flag that notifies the 	%
%   loader that it must first check the user-defined libraries when 	%
%   searching the file system for a module. This prevents a module in a	%
%   user-defined library named say "file_read", to be consulted instead %
%   of the "syslib/file_read" module.					%
%   For every goal appearing after the load_library_directories/0 goal	%
%   in predicate interpreter/0 above, that is imported from another 	%
%   system module not yet loaded in the system, care must be taken to	%
%   force the loading in the system of the file containing the		%
%   predicate before setting the flag. This can be easily done by	%
%   importing a dummy predicate from that module.			%
%   At the time when the comment is written the following modules have	%
%   already been loaded in the system before consulting the .xsbrc file	%
%	1) loader	2) interpreter	3) std_sbp	4) machine	%
%	5) consult	6) standard	7) file_write			%
%   The only modules loaded by this interpreter after setting the flag	%
%   are:								%
%	1) file_read	2) file_op					%
%   and the following predicate ensures that is loaded in the system	%
%   before setting the flag.						%
%-----------------------------------------------------------------------%

hack_libpath_changed :-
	force_file_read_loaded,
	force_file_op_loaded.


/* === interrupt handlers ============================================== */

/* ------ undefined predicate interrupt handler ------------------------ */

'_$load_undef'(Term) :- 
	conpsc(Term, PSC),
	load_pred(PSC),
	psc_prop(PSC, Addr),
	psc_type(PSC, T),
	code_call(Addr, Term, T).

/* ------ the keyboard interrupt handler ------------------------------- */

'_$keyboard_int'(Call) :- 
	is_xwammode(Flag),
	(Flag =:= 1 -> abort ; 	break, call_c(Call)).

'_$calltonum'(_) :- file_write(1, 'Illegal call: to a number!'), 
	file_put(1, 10), abort.

/* ------ delay awaking handler ---------------------------------------- */

'_$wakedelay'(CurGoal) :- 
	solve_wakengoals,
	term_psc(CurGoal, Psc),
	psc_prop(Psc, Addr),
	psc_type(Psc, T),
	code_call(Addr, CurGoal, T).

/* ------ stack over flow handler -------------------------------------- */

'_$stackfull'(_) :- file_write(1, 'Stack Overflow!!'),file_put(1, 10),abort.

/* === version message ================================================= */

version_message :-
	stat_flag(16, Major),
	stat_flag(17, Minor),
	stat_flag(18, Word), version_word(Word, WordM),
	stat_flag(19, Mode), version_mode(Mode, ModeM),
	stat_flag(20, Para), version_para(Para, ParaM),
	stat_flag(22, Date),
	file_write(1, 'XSB (Beta) Version '),
	file_write(1, Major),
	file_put(1, 0'.), 
	file_write(1, Minor),
	Year is Date >> 16,
	Month is (Date >> 8) /\ 255, 
        Day is Date /\ 255,
	file_write(1, ' ('),
	file_write(1, Year),
	file_put(1, 0'/),
	file_write(1, Month),
	file_put(1, 0'/),
	file_write(1, Day),
	file_put(1, 0')), 
	file_put(1, 10),
	file_write(1, ParaM), file_write(1, WordM), file_write(1, ModeM),
	file_put(1, 10),
	fail.		% I do not wanna leave all choice points!
version_message.


version_para(1, '[sequential, ').
version_para(2, '[parallel, ').

version_word(1, 'single word, ').
version_word(2, 'one & half word, ').
version_word(3, 'double word, ').

version_mode(1, 'optimal mode]').
version_mode(2, 'debug mode]').
version_mode(3, 'profile mode]').

/* === utilities ======================================================= */

conpsc(Con, Psc) :-
	term_type(Con, T),
	( T =:= 1 -> term_psc(Con, Psc)
	; T =:= 5 -> psc_insert(Con, 0, Psc, 0)
	).
conname(Con, Name) :-
	term_type(Con, T),
	( T =:= 1 -> term_psc(Con, Psc), psc_name(Psc, Name)
	; T =:= 5 -> Name = Con
	).

/*
consetonce(Con, Val) :- 
	conpsc(Con, PSC), 
	psc_prop(PSC,Old),
	(Old =:= 0 -> psc_set_prop(PSC, Val, 1) ; true).
*/
conset(Con, Val) :- conpsc(Con, PSC), psc_set_prop(PSC, Val, 1).
conval(Con, Val) :- conpsc(Con, PSC), psc_prop(PSC, Val).


repeat.
repeat :- repeat.

prompt(New, Old) :- get_prompt(Old), set_prompt(New).

get_prompt(Prompt) :-
	term_psc(prompt(_), Psc),
	psc_prop(Psc, Buff),	
	str_insert(Buff, String),	% temp solution, since ALIAS not work
	Prompt = String.		% for double word.

set_prompt(Prompt) :-
	term_psc(prompt(_), Psc),
	psc_set_type(Psc, 8, 1),			% T_CSET
	conname(Prompt, String),
	psc_set_prop(Psc, String, 1).

print_debug_state :-
	stat_flag(6,Trace),
	stat_flag(4,Debug),
	( Trace > 0 ->	file_write(1,[trace]), file_put(1,10)
	; Debug > 0 ->	file_write(1,[debug]), file_put(1,10)
	; true
	).
		
/*
func_trans(X, X).
*/

% The following should be a syntactic transformation 
% to handle functional syntax. dsw

func_trans(X, Y) :- func_trans(X, Z, Z, Y).

func_trans(X, Y, O, N) :- 
	term_type(X, T),
	( T =:= 0 -> Y = X, N = O
	; T =:= 3 -> X = [X1|X2], Y = [Y1|Y2],
		func_trans(X1, Y1, O, M), func_trans(X2, Y2, M, N)
	; T =:= 1 ->
		term_psc(X, Psc),
		psc_type(Psc, Type),
		( Type =:= 14, M = ((Y is Z), O)	% Y is free
		; Type =\= 14, M = O, Z = Y
		),
		term_new(Psc, Z),
		psc_arity(Psc, A),
		func_transargs(A, X, Z, M, N)
	; /* T=:=5; T=:=2; T=:=6 */ Y = X, N = O
	).

func_transargs(0, _, _, O, O) :- !.
func_transargs(I, X, Y, O, N) :-
	term_arg(X, I, X1),
	term_arg(Y, I, Y1),
	func_trans(X1, Y1, O, M),
	I1 is I - 1,
	func_transargs(I1, X, Y, M, N).

