/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File				:  loader.P
  Author(s)			:  Kostis F. Sagonas, Jiyang Xu
  Modified by			:  Jiyang Xu (August 13, 1990)
  Made libpath/1 dynamic by	:  Kostis F. Sagonas (May 28, 1992)
  Last modification		:  April 3, 1993
========================================================================*/



:- load(interpreter).


libpath(P) :- 
	libraries_loaded, 
	library_directory(Dir), 
	expand_file_name(Dir, AbsoluteDir),
	ll_conname(AbsoluteDir, P).
libpath(P) :- 
	stat_flag(23, Dir),
	ll_conname('/lib/', Sub),
	str_cat(Dir, Sub, P).
libpath(P) :- 
	stat_flag(23, Dir),
	ll_conname('/syslib/', Sub),
	str_cat(Dir, Sub, P).
libpath(P) :- 
	stat_flag(23, Dir),
	ll_conname('/cmplib/', Sub),
	str_cat(Dir, Sub, P).
libpath('./').


/* Flag[50] set indicates that the ~/.xsbrc module has been loaded.	*/

libraries_loaded :- 
	stat_flag(50, Lib_Loaded),
	Lib_Loaded =:= 1.


load_pred(PSC) :- \+ (\+ (load_pred0(PSC))).

load_pred0(PSC) :- 
	stat_flag(1, Hitrace), 	
	stat_set_flag(1, 0, 1),			% turn off hitrace
	load_pred1(PSC),
	stat_set_flag(1, Hitrace, 1).		% resume hitrace status if 
						% successful

load_pred1(PSC) :-
	'_$savecp'(B),
	psc_name(PSC, Name),
	psc_arity(PSC, Arity),
	psc_env(PSC, Env), 
	( Env =:= 2 ->			% is "imported"
		psc_prop(PSC, ModPSC),
		psc_name(ModPSC, ModName),
		psc_type(ModPSC, ModType), 
		( ModType =:= 4
		; ModType =\= 4,	% module is not loaded yet
		  load(ModName)		% can fail here <--- MAY BE A BUG HERE 
		),
		( psc_type(PSC, NType),
		  ( NType =:= 2; NType =:= 13; NType =:= 14 ), 
			% module loaded, and pred/func/foreign-pred is there
		  '_$cutto'(B)
		 % module loaded, but pred/func/foreign-pred is not there
	 	; ll_prompt('The predicate/function '),
		  ll_prompt(Name), ll_prompt('/'), ll_writearity(Arity),
		  ll_prompt(' is not defined in the module '),
		  ll_prompt(ModName), l_nl, fail
		)
	; ll_prompt('Undefined predicate/function: '),  % no module, or not pred
	  ll_prompt(Name), ll_prompt('/'), ll_writearity(Arity), l_nl, 
	  fail
	).

load(ModName) :-
	'_$savecp'(B),
	stat_flag(4, Debug_on),
	( Debug_on =:= 1 ->
		ll_prompt('[ load module '),
		ll_prompt(ModName), 
		ll_prompt(' ]'), l_nl
	; Debug_on =\= 1
	),
	( search_file(ModName, FileName),
	  code_load(FileName, Init, 0),	% use the C loader
	  Init > 0,			% load succeeded
	  '_$cutto'(B),
	  code_call(Init, true, 2)	% call '_$main'
	; ll_prompt('Cannot find the module '), ll_prompt(ModName), l_nl,
	  fail
	).

search_file(ModName, FileName) :- 
	expand_file_name(ModName, ExpandedModName),
	ll_conname(ExpandedModName, MName),
	ll_conname('.O', OName),
	buff_byte(MName, 0, Char),
	( Char =:= 0'/ ->	% we already have an absolute path name
		WholePath = MName
	; libpath(LibPath),
	  ll_conname(LibPath, PName),
	  str_cat(PName, MName, WholePath)
	),
	str_cat(WholePath, OName, FileName).

/*======================================================================*/
/* Auxiliary predicates.						*/
/*	Many of them already appear somewhere else in the system, but	*/
/*	they also appear here due to the fact that they	cannot be	*/
/*	imported in this module (the loader should be self-contained).	*/
/*======================================================================*/

l_nl :- file_put(1, 10).

ll_prompt(String) :- ll_prompt(0, String).

ll_prompt(N, String) :-
	term_type(String, T), 
	( T =:= 5 -> String0 = String
	; T =:= 1 -> term_psc(String, PSC), psc_name(PSC, String0)
	),
	buff_byte(String0, N, Char),
	( Char =:= 0 -> true
	; file_put(1, Char), N1 is N+1, ll_prompt(N1, String0)
	).

ll_writearity(Int) :- 
	( Int < 10 -> Char is Int + 48, file_put(1, Char)
	; file_put(1, 42)	% when arity > 9, write an asterisk *
	).	

ll_conname(Con, Name) :- 
	term_type(Con, T),
	( T =:= 1 -> term_psc(Con, Psc), psc_name(Psc, Name)
	; T =:= 5 -> Name = Con
	).

