/************************************************************************/
/*									*/
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*									*/
/* Everyone is granted permission to copy, modify and redistribute XSB,	*/
/* but only under the conditions described in the XSB License Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.			 	*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  machine.P
  Author(s)		:  Jiyang Xu, Kostis F. Sagonas, Terrance Swift
			   David S. Warren.
  Last modification	:  July, 1992
========================================================================*/


/*======================================================================*/
/* This file contains definitions for predicates implemented by the	*/
/* underline machine (hard builtins).					*/
/*======================================================================*/

psc_name(_PSC, _String) :- '_$builtin'(1).
psc_arity(_PSC, _Arity) :- '_$builtin'(2).
psc_type(_PSC, _Type) :- '_$builtin'(3).
psc_prop(_PSC, _Term) :- '_$builtin'(4).
psc_set_type(_PSC, _Type, _Perm) :- '_$builtin'(5).
psc_set_prop(_PSC, _Term, _Perm) :- '_$builtin'(6).
file_open(_NameString, _RWMode, _File) :- '_$builtin'(7).
file_close(_File) :- '_$builtin'(8).
file_get(_File, _Char) :- '_$builtin'(9).
file_put(_File, _Char) :- '_$builtin'(10).
term_psc(_Term, _PSC) :- '_$builtin'(11).
term_type(_Term, _Type) :- '_$builtin'(12).
term_compare(_Term1, _Term2, _Res) :- '_$builtin'(13).
term_new(_PSC, _Term) :- '_$builtin'(14).
term_arg(_Term, _Index, _Arg) :- '_$builtin'(15).
term_set_arg(_Term, _Index, _Arg, _Perm) :- '_$builtin'(16).
stat_flag(_Flag, _Value) :- '_$builtin'(17).
stat_set_flag(_Flag, _Value, _Perm) :- '_$builtin'(18).
buff_alloc(_Size, _Buffer, _Perm) :- '_$builtin'(19).
buff_word(_Buffer, _Disp, _Value) :- '_$builtin'(20).
buff_set_word(_Buffer, _Disp, _Value) :- '_$builtin'(21).
buff_byte(_Buffer, _Disp, _Value) :- '_$builtin'(22).
buff_set_byte(_Buffer, _Disp, _Value) :- '_$builtin'(23).
code_call(_CodeAddr, _Term, _Type) :- '_$builtin'(24).
str_len(_String, _Length) :- '_$builtin'(25).
str_cpy(_String1, _Buffer) :- '_$builtin'(26).
str_cat(_String1, _String2, _String3) :- '_$builtin'(27).
str_cmp(_String1, _String2, _Res) :- '_$builtin'(28).
str_hsh(_String, _Arity, _Size, _Res) :- '_$builtin'(29).
str_insert(_StringIn, _StringOut) :- '_$builtin'(30).

/* some other builtins that might need hard implementation */

stat_sta(_X) :- '_$builtin'(32).
stat_cputime(_X) :- '_$builtin'(33).

code_load(ByteCodeFileName, InitAddr) :- 
	code_load(ByteCodeFileName, InitAddr, 0).
code_load(_ByteCodeFileName, _InitAddr, _Exp) :- '_$builtin'(34).
buff_set_var(_Buff, _Disp, _BufSize, _Var) :- '_$builtin'(35).
buff_dealloc(_Buff, _OldSize, _NewSize, _Perm) :- '_$builtin'(36).
buff_cell(_Buff, _Disp, _Cell) :- '_$builtin'(37).
buff_set_cell(_Buff, _Disp, _Tag, _Value) :- '_$builtin'(38).

/* for efficiency reasons, the following predicates are also implemented */

file_getword(_File, _Word) :- '_$builtin'(40).
file_putword(_File, _Word) :- '_$builtin'(41).
psc_insert(_Name, _Arity, _PSC, _MName) :- '_$builtin'(42).
psc_import(_Pname, _Arity, _MName) :- '_$builtin'(43).
file_getbuf(_File, _Bytes, _Buf, _Disp) :- '_$builtin'(44).
file_putbuf(_File, _Bytes, _Buf, _Disp) :- '_$builtin'(45).
psc_insertmod(_ModName, _Def, _PSC) :- '_$builtin'(46).
load_seg(_SegNo, _TextBytes, _IndexBytes, _File, _InitAddr) :- '_$builtin'(47).
file_gettoken(_File, _PrevCh, _Type, _Value, _NextCh) :- '_$builtin'(48).
file_puttoken(_File, _Type, _Value) :- '_$builtin'(49).
term_hash(_Term, _Size, _Value) :- '_$builtin'(50).
unload_seg(_CodeSeg) :- '_$builtin'(51).
load_obj(_FileName, _Mod, _LdOption, _InitAddr) :- '_$builtin'(52).
expand_file_name(_FileName, _ExpandedFileName) :- '_$builtin'(53).
getenv(_EnvVar, _Value) :- '_$builtin'(54).
/* sys_syscall(CallNo, Res, Arg, ......) :- '_$builtin'(55). */
sys_system(_Command, _Res) :- '_$builtin'(56).
sys_gethost(_Name, _Buffer) :- '_$builtin'(57).
sys_errno(_Errno) :- '_$builtin'(58).
sys_brocall(_CallNo, _ArgBuff, _Res) :- '_$builtin'(59).
file_time(_File, _Time) :- '_$builtin'(60).

/* domain stuff out for now
dom_size(_Dom, _Size) :- '_$builtin'(80).
dom_type(_Dom, _Type) :- '_$builtin'(81).
dom_range(_Dom, _Min, _Max) :- '_$builtin'(82).
dom_elem(_Dom, _Index) :- '_$builtin'(83).
dom_diff(_X, _Y, _Z) :- '_$builtin'(84).
dom_plus(_X, _Y, _Z) :- '_$builtin'(85).
dom_times(_X, _Y, _Z) :- '_$builtin'(86).
dom_min(_Dom, _Min) :- '_$builtin'(87).
dom_enum(_Dom, _TypePsc) :- '_$builtin'(88).
dom_esize(_Dom, _Size) :- '_$builtin'(89).
*/

buff_assign_word(_Buffer, _Disp, _Value) :- '_$builtin'(96).

psc_env(_PSC, _Env) :- '_$builtin'(100).
psc_spy(_PSC, _Spy) :- '_$builtin'(101).

psc_tabled(_PSC, _Tabled) :- '_$builtin'(102).
tip_prop(_Tip, _Field, _Get_or_set, _Value):- '_$builtin'(103).
is_incomplete(_Tab,_Flag):- '_$builtin'(104).
get_osp_breg(_Osp,_Breg):- '_$builtin'(105).
cut_if_leader(_Call,_Osp,_Breg):- '_$builtin'(106).

pair_psc(_Pair, _PSC) :- '_$builtin'(110).
pair_next(_Pair, _Next_Pair) :- '_$builtin'(111).
next_bucket(_Bucket, _Next_Bucket) :- '_$builtin'(112).

suspend(_Call,_Tab,_Arity):- '_$builtin'(113).
incomplete(_Tab,_Flag):- '_$builtin'(114).
heap_copyarg(_From,_To):- '_$builtin'(115).
table_intern(_Ptr,_Intptr) :- '_$builtin'(116).
is_xwammode(_Flag) :- '_$builtin'(117).
close_open_tables :- '_$builtin'(118).
call_hash(_Term,_Ht_base,_Ar,_Chs,_Call_str):- '_$builtin'(119).
print_predicate_table(_Psc):- '_$builtin'(120).
analyze_predicate_table(_Psc):- '_$builtin'(124).
comp_code_call(_Psc,_Call):- '_$builtin'(125).
abolish_table_info:- '_$builtin'(126).
zero_out_profile:- '_$builtin'(127).
write_out_profile:- '_$builtin'(128).
assert_code_to_buff(_Clause,_Size) :- '_$builtin'(129).
assert_buff_to_clref(_Head,_NAri,_Prref,_AZ,_RInd,_HashTabSize,_Clref) :-
	'_$builtin'(130).
fmt_write(_File,_Format,_ValsList) :- '_$builtin'(131).
fmt_read(_File,_Format,_TypeList,_ValsListVar,_Ret) :- '_$builtin'(132).


fmt_read(Format, TypeList, ValsListVar, Ret) :-
	stat_flag(10, FileDescriptor),
	fmt_read(FileDescriptor, Format, TypeList, ValsListVar, Ret).
fmt_write(Format, ValsListVar) :-
	stat_flag(11, FileDescriptor),
	fmt_write(FileDescriptor, Format, ValsListVar).
	
psc_insert(Name, Arity, PSC) :- psc_insert(Name, Arity, PSC, 0). 

conname(Con, Name) :-
	term_type(Con, Type),
	( Type =:= 1 -> term_psc(Con, Psc), psc_name(Psc, Name)
	; Type =:= 5 -> Name = Con
	).

conpsc(Con, Psc) :-
	term_type(Con, Type),
	( Type =:= 1 -> term_psc(Con, Psc)
	; Type =:= 5 -> psc_insert(Con, 0, Psc, 0)
	).

is_integer(Term) :- term_type(Term, Type), Type =:= 2.

machine_file_exists(File) :- 
	sys_syscall(33, Res, File, 0), Res =:= 0.

/* The following is extremely DIRTY and should be used very very	*/
/* cautiously!.  For the moment it is only used in symtab.P when the	*/
/* symbol that was inserted into the symbol table may be retracted and	*/
/* the count of the symbols has to remain the same.			*/

backtrackable_buff_set_word(Buff, Bytes, NewValue, _) :- 
	buff_set_word(Buff, Bytes, NewValue). 
backtrackable_buff_set_word(Buff, Bytes, _, OldValue) :- 
	buff_set_word(Buff, Bytes, OldValue), fail.

sys_syscall(_CallNo, _Res, _A1, _A2) :- '_$builtin'(55).


/* --------------------- end of file machine.P ------------------------ */
