/************************************************************************/
/*                                                                      */
/* XSB System								*/
/* Copyright (C) SUNY at Stony Brook, 1986, 1993; ECRC, 1990		*/
/*                                                                      */
/* Everyone is granted permission to copy, modify and redistribute XSB  */
/* but only under the conditions described in the XSB Licence Agreement.*/
/* A copy of this license is supposed to have been given to you along	*/
/* with XSB so you can know your rights and responsibilities.		*/
/* It should be in a file named LICENSE.				*/
/* Among other things, this notice must be preserved on all copies.	*/
/*									*/
/************************************************************************/

/*======================================================================
  File			:  string.P
  Author(s)		:  Kostis F. Sagonas, Jiyang Xu
  Modified by		:  Jiyang Xu
  Last modification	:  April 3, 1993
========================================================================*/


%  str_fromlist(-Atom, +List)
%  Given a list of integers representing characters returns the 
%  corresponding atom. This is more like atom_chars/2 that works
%  only in the opposite direction.

str_fromlist([], "[]") :- 
	!.	% Due to a quite strange representation of the empty list.
str_fromlist(String, List) :-
	str_fl_check(List, Length),
	buff_alloc(Length, Buff, 0),	% alloc in heap 
	str_fl_put(Buff, 0, List),
	str_insert(Buff, String).

str_fl_check([], 1).
str_fl_check([_H|T], N) :- 
/* The following check (enhanced) is now ensured by name/2
	term_type(H, Type), Type =:= 2,	% H is integer */
	str_fl_check(T,M), N is M + 1.

str_fl_put(Buff, N, []) :- buff_set_byte(Buff, N, 0).  % ending \0
str_fl_put(Buff, N, [A|T]) :-
	buff_set_byte(Buff, N, A),
	N1 is N + 1,
	str_fl_put(Buff, N1, T).

str_tolist(Buff, List) :- 
	( Buff == [] -> List = "[]" ; str_tolist(Buff, 0, List) ).

str_tolist(Buff, N, List) :-
	buff_byte(Buff, N, Char),
	( Char = 0 -> List = []
	; List = [Char|Rest],
	  N1 is N + 1,
	  str_tolist(Buff, N1, Rest)
	).

str_length(Buff, N) :- 
	str_length(Buff, 0, N).

str_length(Buff, Nin, Nout) :-
	buff_byte(Buff, Nin, Char),
	( Char = 0 -> Nin = Nout
	; Nmed is Nin + 1, str_length(Buff, Nmed, Nout)
	).

str_cat(S1, S2, S3) :-
	str_length(S1, L1), 
	str_length(S2, L2),
	L3 is L1 + L2,
	BufSize is L3 + 1,
	buff_alloc(BufSize, Buff, 0),		% alloc in heap 
	str_copy(S1, 0, Buff, 0),
	str_copy(S2, 0, Buff, L1),
	buff_set_byte(Buff, L3, 0),		% trailing 0
	str_insert(Buff, S3).

str_copy(Source, SDisp, Dest, DDisp) :-
	buff_byte(Source, SDisp, Char),
	( Char = 0 -> true
	; buff_set_byte(Dest, DDisp, Char),
	  SDisp1 is SDisp + 1,
	  DDisp1 is DDisp + 1,
	  str_copy(Source, SDisp1, Dest, DDisp1)
	).

str_substring(String, Loc, Len, SubString) :-
	BufSize is Len+1,
	buff_alloc(BufSize, Buff, 0),		% alloc in heap
	str_ncopy(String, Loc, Buff, 0, Len),
	buff_set_byte(Buff, Len, 0),		% trailing 0
	str_insert(Buff, SubString).

str_ncopy(_, _, _, _, 0).
str_ncopy(Source, SDisp, Dest, DDisp, Len) :- Len > 0,
	buff_byte(Source, SDisp, Char),
	buff_set_byte(Dest, DDisp, Char),
	SDisp1 is SDisp + 1,
	DDisp1 is DDisp + 1,
	Len1 is Len-1,
	str_ncopy(Source, SDisp1, Dest, DDisp1, Len1).


/* ---------------------- end of file string.P ------------------------ */
